// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

// Package piboot implements the bootloader protocol for the Raspberry Pi boot loader.
//
// This implementation assumes the following: The raspberry pi boot firmware is
// at least the release from 2020-10-28. The kernel image and command line files
// are stored in sub-directories of the firmware boot partition under slot-a/
// and slot-b/. Each kernel command line file points to the correct slot. The
// kernel image supports the RPI_FIRMWARE_SET_REBOOT_FLAGS feature.
//
// XXX: simplify slot-a/ and slot-b/ detection with the "prefix" variable.
//
// References:
//  - https://github.com/raspberrypi/linux/commit/757666748ebf69dc161a262faa3717a14d68e5aa
//  - https://github.com/raspberrypi/rpi-eeprom/blob/master/firmware/release-notes.md
//  - https://www.raspberrypi.org/documentation/hardware/raspberrypi/bootmodes/README.md
package piboot

import (
	"errors"
	"fmt"
	"os"
	"os/exec"
	"path/filepath"
	"strings"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/boot"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/picfg"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/picfg/condfilter"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/rauc/installhandler"
)

const (
	slotAPrefix = "slot-a/"
	slotBPrefix = "slot-b/"

	// configTxtName is the name of the firmware configuration file read during the boot process.
	configTxtName = "config.txt"

	// trybootTxtName is the name of the firmware configuration file read during the try-once boot process.
	// It is further described at https://github.com/raspberrypi/linux/commit/757666748ebf69dc161a262faa3717a14d68e5aa
	trybootTxtName = "tryboot.txt"

	kernelParameter    = "kernel"
	initramfsParameter = "initramfs"
	cmdlineParameter   = "cmdline"
)

// PiBoot implements boot.Protocol for the Raspberry Pi boot loader.
type PiBoot struct {
	// BootMountPoint determines the mount point of the firmware boot partition.
	// It is expected that the "config.txt" file is present there.
	BootMountPoint string

	// FilterPredicate decides if a config.txt section should be inspected.
	//
	// A non-zero filter predicate can be used to support a single system image
	// that is compatible with different device types, each with a distinct
	// kernel image by only considering those configuration sections that apply
	// to the current device.
	//
	// When the filter predicate is not used, section names are not parsed and
	// unrecognized sections are preserved perfectly. When the predicate is set
	// all sections must be recognized.
	FilterPredicate func([]condfilter.ConditionalFilter) bool
}

// New returns a new PiBoot with the given firmware boot partition mount point.
// The specified directory must contain the "config.txt" file.
func New(bootMountPoint string) (*PiBoot, error) {
	file, err := os.Open(filepath.Join(bootMountPoint, configTxtName))
	if err != nil {
		return nil, err
	}

	if err := file.Close(); err != nil {
		return nil, err
	}

	return &PiBoot{BootMountPoint: bootMountPoint}, nil
}

// UnknownActiveSlotError records problems with determining active slot.
type UnknownActiveSlotError struct {
	Err error
}

// Unwrap returns the error wrapped inside the unknown active slot error.
func (e *UnknownActiveSlotError) Unwrap() error {
	return e.Err
}

// Error implements the error interface.
func (e *UnknownActiveSlotError) Error() string {
	return fmt.Sprintf("cannot determine active slot: %v", e.Err)
}

// UnknownInactiveSlotError records problems with determining inactive slot.
type UnknownInactiveSlotError struct {
	Err error
}

// Unwrap returns the error wrapped inside the unknown inactive slot error.
func (e *UnknownInactiveSlotError) Unwrap() error {
	return e.Err
}

// Error implements the error interface.
func (e *UnknownInactiveSlotError) Error() string {
	return fmt.Sprintf("cannot determine inactive slot: %v", e.Err)
}

var (
	// ErrCorruptedConfig records boot configuration deviating from assumptions.
	ErrCorruptedConfig = errors.New("inconsistent or corrupted configuration")
	// ErrExpectedSlotAOrSlotB records existence of unexpected slot.
	ErrExpectedSlotAOrSlotB = errors.New("expected slot A or slot B")
)

// QueryActive returns the slot that is used for booting.
//
// Active slot is determined by inspecting the kernel and cmdline parameters in
// the last section that is applicable, as decided by the filter predicate. By
// arbitrary convention, both parameters must have the prefix "slot-a/" or
// "slot-b/" to be recognized.
//
// If there is an error, it will be of type *UnknownActiveSlotError.
func (pi *PiBoot) QueryActive() (slot boot.Slot, err error) {
	file, err := os.Open(filepath.Join(pi.BootMountPoint, configTxtName))
	if err != nil {
		return boot.InvalidSlot, &UnknownActiveSlotError{Err: err}
	}

	defer func() {
		e := file.Close()
		if err == nil {
			err = e
		}
	}()

	var configTxt picfg.ConfigTxt

	if err := picfg.NewDecoder(file).Decode(&configTxt); err != nil {
		return boot.InvalidSlot, &UnknownActiveSlotError{Err: err}
	}

	// Find the configured kernel and kernel cmdline
	var kernel, initramfs, cmdline string

	err = visitSections(&configTxt, pi.FilterPredicate, func(sect *picfg.Section) error {
		for _, param := range sect.Parameters {
			switch param.Name {
			case kernelParameter:
				kernel = param.Value
			case initramfsParameter:
				initramfs = param.Value
			case cmdlineParameter:
				cmdline = param.Value
			}
		}
		return nil
	})
	if err != nil {
		return boot.InvalidSlot, &UnknownActiveSlotError{Err: err}
	}
	// Assuming kernel and cmdline are stored in slot-specific sub-directories,
	// find the configured slot.
	switch {
	case strings.HasPrefix(kernel, slotAPrefix) && strings.HasPrefix(cmdline, slotAPrefix) && strings.HasPrefix(initramfs, slotAPrefix):
		return boot.SlotA, nil
	case strings.HasPrefix(kernel, slotBPrefix) && strings.HasPrefix(cmdline, slotBPrefix) && strings.HasPrefix(initramfs, slotBPrefix):
		return boot.SlotB, nil
	default:
		return boot.InvalidSlot, &UnknownActiveSlotError{Err: ErrCorruptedConfig}
	}
}

// QueryInactive returns the slot that is not used for booting.
//
// If there is an error, it will be of type *UnknownInactiveSlotError.
func (pi *PiBoot) QueryInactive() (boot.Slot, error) {
	// Find the active slot and return the other slot.
	slot, err := pi.QueryActive()
	if err != nil {
		// Unwrap UnknownActiveSlotError to avoid confusing error messages.
		var realErr = err

		var e *UnknownActiveSlotError

		if errors.As(err, &e) {
			realErr = e.Unwrap()
		}

		return boot.InvalidSlot, &UnknownInactiveSlotError{Err: realErr}
	}

	return boot.SynthesizeInactiveSlot(slot), nil
}

// TrySwitch configures the boot loader for a one-off boot using the given slot.
func (pi *PiBoot) TrySwitch(to boot.Slot) (err error) {
	// Load and decode config.txt
	configTxtFile, err := os.Open(filepath.Join(pi.BootMountPoint, configTxtName))
	if err != nil {
		return err
	}

	defer func() {
		e := configTxtFile.Close()
		if err == nil {
			err = e
		}
	}()

	var configTxt picfg.ConfigTxt
	if err := picfg.NewDecoder(configTxtFile).Decode(&configTxt); err != nil {
		return err
	}

	// Find the values to modify, swapping A and B around
	err = visitSections(&configTxt, pi.FilterPredicate, func(sect *picfg.Section) error {
		for paramIdx, param := range sect.Parameters {
			switch param.Name {
			case kernelParameter, cmdlineParameter, initramfsParameter:
				var oldPrefix, newPrefix string
				switch to {
				case boot.SlotA:
					oldPrefix = slotBPrefix
					newPrefix = slotAPrefix
				case boot.SlotB:
					oldPrefix = slotAPrefix
					newPrefix = slotBPrefix
				default:
					continue
				}
				sect.Parameters[paramIdx].Value = strings.Replace(param.Value, oldPrefix, newPrefix, -1)
			}
		}
		return nil
	})
	if err != nil {
		return err
	}

	// Encode and write modified config.txt as tryboot.txt
	trybootTxtFile, err := os.OpenFile(
		filepath.Join(pi.BootMountPoint, trybootTxtName),
		os.O_CREATE|os.O_TRUNC|os.O_WRONLY, 0644)
	if err != nil {
		return err
	}

	defer func() {
		e := trybootTxtFile.Close()
		if err == nil {
			err = e
		}
	}()

	if err := picfg.NewEncoder(trybootTxtFile).Encode(configTxt); err != nil {
		return err
	}

	return nil
}

// CommitSwitch re-configures a try-mode slot for continuous use.
func (pi *PiBoot) CommitSwitch() error {
	return os.Rename(
		filepath.Join(pi.BootMountPoint, trybootTxtName),
		filepath.Join(pi.BootMountPoint, configTxtName))
}

const rebootCommand = "reboot"

// Reboot gracefully reboots the system, passing any platform-specific information to the boot loader if necessary.
func (pi *PiBoot) Reboot(flags boot.RebootFlags) error {
	cmd := exec.Command(rebootCommand)
	if flags&boot.RebootTryBoot != 0 {
		cmd = exec.Command(rebootCommand, "0 tryboot")
	}

	return cmd.Run()
}

// CancelSwitch removes the try-mode boot configuration file.
func (pi *PiBoot) CancelSwitch() error {
	if err := os.Remove(filepath.Join(pi.BootMountPoint, trybootTxtName)); err != nil && !os.IsNotExist(err) {
		return err
	}

	return nil
}

// visitSection calls visitor function on each section in the config file.
//
// An optional predicate function may be used to visit only matching sections.
// The predicate is called with the effective fiters that are in effect at each
// phase of the traversal process. Sections can be mutated by the visitor.
func visitSections(configTxt *picfg.ConfigTxt, pred func([]condfilter.ConditionalFilter) bool, visitor func(sect *picfg.Section) error) error {
	var tracker condfilter.StateTracker

	for idx := range configTxt.Sections {
		sect := &configTxt.Sections[idx]

		if pred != nil && sect.Filter != "" {
			filter, err := condfilter.Unmarshal([]byte(sect.Filter))
			if err != nil {
				return err
			}

			tracker.Filter(filter)

			if !pred(tracker.EffectiveFilters()) {
				continue
			}
		}

		if err := visitor(sect); err != nil {
			return err
		}
	}

	return nil
}

// PostInstallHandler reboots the Raspberry Pi if necessary.
//
// Reboots are required after writing to RAUC slot of type "system".
// Reboot is graceful, shutting down services normally.
func (pi *PiBoot) PostInstallHandler(ctx *installhandler.HandlerContext) error {
	if shouldReboot(ctx) {
		return pi.Reboot(boot.RebootTryBoot)
	}

	return nil
}

// shouldReboot returns true if a slot with "system" class is being modified.
func shouldReboot(ctx *installhandler.HandlerContext) bool {
	for _, slotID := range ctx.TargetSlotIDs {
		if ctx.Slots[slotID].Class == "system" {
			return true
		}
	}

	return false
}
