// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package piboot_test

import (
	"io/ioutil"
	"os"
	"path/filepath"
	"testing"

	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/boot"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/boot/piboot"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/picfg/condfilter"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/rauc/installhandler"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/testutil"
)

func Test(t *testing.T) { TestingT(t) }

type bootTest struct {
	testutil.CommandSuite
	bootDir string
}

var _ = Suite(&bootTest{})

func (s *bootTest) SetUpTest(c *C) {
	s.bootDir = c.MkDir()
	s.MockConfigTxt(c, "")
}

func (s *bootTest) MockConfigTxt(c *C, text string) {
	configName := filepath.Join(s.bootDir, "config.txt")
	err := ioutil.WriteFile(configName, []byte(text), 0o644)
	c.Assert(err, IsNil)
}

func (s *bootTest) MockTrybootTxt(c *C, text string) {
	configName := filepath.Join(s.bootDir, "tryboot.txt")
	err := ioutil.WriteFile(configName, []byte(text), 0o644)
	c.Assert(err, IsNil)
}

func (s *bootTest) TestOpen(c *C) {
	pi, err := piboot.New(s.bootDir)
	c.Assert(err, IsNil)
	c.Check(pi, DeepEquals, &piboot.PiBoot{
		BootMountPoint: s.bootDir,
	})
}

func (s *bootTest) TestQueryActive(c *C) {
	pi, err := piboot.New(s.bootDir)
	c.Assert(err, IsNil)

	s.MockConfigTxt(c, ``)

	slot, err := pi.QueryActive()
	c.Assert(err, ErrorMatches, "cannot determine active slot: inconsistent or corrupted configuration")
	c.Check(slot, Equals, boot.InvalidSlot)

	s.MockConfigTxt(c, ""+
		"kernel=slot-a/vmlinuz\n"+
		"cmdline=slot-a/cmdline.txt\n"+
		"initramfs slot-a/initrd.img followkernel\n")

	slot, err = pi.QueryActive()
	c.Assert(err, IsNil)
	c.Check(slot, Equals, boot.SlotA)

	s.MockConfigTxt(c, ""+
		"kernel=slot-b/vmlinuz\n"+
		"cmdline=slot-b/cmdline.txt\n"+
		"initramfs slot-b/initrd.img followkernel\n")

	slot, err = pi.QueryActive()
	c.Assert(err, IsNil)
	c.Check(slot, Equals, boot.SlotB)

	s.MockConfigTxt(c, ""+
		"kernel=slot-a/vmlinuz\n"+
		"cmdline=slot-b/cmdline.txt\n"+
		"initramfs slot-c/initrd.img followkernel\n")

	slot, err = pi.QueryActive()
	c.Assert(err, ErrorMatches, "cannot determine active slot: inconsistent or corrupted configuration")
	c.Check(slot, Equals, boot.InvalidSlot)
}

func (s *bootTest) TestQueryActiveWithFilterPredicate(c *C) {
	pi, err := piboot.New(s.bootDir)
	c.Assert(err, IsNil)

	s.MockConfigTxt(c, ""+
		"[pi3]\n"+
		"kernel=slot-b/pi3-vmlinuz\n"+
		"cmdline=slot-b/pi3-cmdline.txt\n"+
		"initramfs slot-b/pi3-initrd.img followkernel\n"+
		"[pi4]\n"+
		"kernel=slot-a/pi4-vmlinuz\n"+
		"cmdline=slot-a/pi4-cmdline.txt\n"+
		"initramfs slot-a/pi4-initrd.img followkernel\n")

	pi.FilterPredicate = func(effective []condfilter.ConditionalFilter) bool {
		for _, filter := range effective {
			if model, ok := filter.(condfilter.PiModelFilter); ok {
				return model == condfilter.Pi4
			}
		}

		return true
	}
	slot, err := pi.QueryActive()
	c.Assert(err, IsNil)
	c.Check(slot, Equals, boot.SlotA)

	pi.FilterPredicate = func(effective []condfilter.ConditionalFilter) bool {
		for _, filter := range effective {
			if model, ok := filter.(condfilter.PiModelFilter); ok {
				return model == condfilter.Pi3
			}
		}

		return true
	}
	slot, err = pi.QueryActive()
	c.Assert(err, IsNil)
	c.Check(slot, Equals, boot.SlotB)
}

func (s *bootTest) TestQueryInactive(c *C) {
	pi, err := piboot.New(s.bootDir)
	c.Assert(err, IsNil)

	s.MockConfigTxt(c, ``)

	slot, err := pi.QueryInactive()
	c.Assert(err, ErrorMatches, "cannot determine inactive slot: inconsistent or corrupted configuration")
	c.Check(slot, Equals, boot.InvalidSlot)

	s.MockConfigTxt(c, ""+
		"kernel=slot-a/vmlinuz\n"+
		"cmdline=slot-a/cmdline.txt\n"+
		"initramfs slot-a/initrd.img followkernel\n")

	slot, err = pi.QueryInactive()
	c.Assert(err, IsNil)
	c.Check(slot, Equals, boot.SlotB)

	s.MockConfigTxt(c, ""+
		"kernel=slot-b/vmlinuz\n"+
		"cmdline=slot-b/cmdline.txt\n"+
		"initramfs slot-b/initrd.img followkernel\n")

	slot, err = pi.QueryInactive()
	c.Assert(err, IsNil)
	c.Check(slot, Equals, boot.SlotA)

	s.MockConfigTxt(c, ""+
		"kernel=slot-a/vmlinuz\n"+
		"cmdline=slot-b/cmdline.txt\n"+
		"initramfs slot-c/initrd.img followkernel\n")

	slot, err = pi.QueryInactive()
	c.Assert(err, ErrorMatches, "cannot determine inactive slot: inconsistent or corrupted configuration")
	c.Check(slot, Equals, boot.InvalidSlot)
}

func (s *bootTest) TestTrySwitchAtoB(c *C) {
	pi, err := piboot.New(s.bootDir)
	c.Assert(err, IsNil)

	s.MockConfigTxt(c, ""+
		"kernel=slot-a/vmlinuz\n"+
		"cmdline=slot-a/cmdline.txt\n"+
		"initramfs slot-a/initrd.img followkernel\n")

	err = pi.TrySwitch(boot.SlotB)
	c.Assert(err, IsNil)

	data, err := ioutil.ReadFile(filepath.Join(s.bootDir, "config.txt"))
	c.Assert(err, IsNil)
	c.Check(string(data), Equals, ""+
		"kernel=slot-a/vmlinuz\n"+
		"cmdline=slot-a/cmdline.txt\n"+
		"initramfs slot-a/initrd.img followkernel\n")

	data, err = ioutil.ReadFile(filepath.Join(s.bootDir, "tryboot.txt"))
	c.Assert(err, IsNil)
	c.Check(string(data), Equals, ""+
		"kernel=slot-b/vmlinuz\n"+
		"cmdline=slot-b/cmdline.txt\n"+
		"initramfs slot-b/initrd.img followkernel\n")
}

func (s *bootTest) TestTrySwitchBtoA(c *C) {
	pi, err := piboot.New(s.bootDir)
	c.Assert(err, IsNil)

	s.MockConfigTxt(c, ""+
		"kernel=slot-b/vmlinuz\n"+
		"cmdline=slot-b/cmdline.txt\n"+
		"initramfs slot-b/initrd.img followkernel\n")

	err = pi.TrySwitch(boot.SlotA)
	c.Assert(err, IsNil)

	data, err := ioutil.ReadFile(filepath.Join(s.bootDir, "config.txt"))
	c.Assert(err, IsNil)
	c.Check(string(data), Equals, ""+
		"kernel=slot-b/vmlinuz\n"+
		"cmdline=slot-b/cmdline.txt\n"+
		"initramfs slot-b/initrd.img followkernel\n")

	data, err = ioutil.ReadFile(filepath.Join(s.bootDir, "tryboot.txt"))
	c.Assert(err, IsNil)
	c.Check(string(data), Equals, ""+
		"kernel=slot-a/vmlinuz\n"+
		"cmdline=slot-a/cmdline.txt\n"+
		"initramfs slot-a/initrd.img followkernel\n")
}

func (s *bootTest) TestReboot(c *C) {
	pi, err := piboot.New(s.bootDir)
	c.Assert(err, IsNil)

	restore, rebootLog := s.MockCommand(c, "reboot")
	defer restore()

	err = pi.Reboot(0)
	c.Assert(err, IsNil)
	c.Check(rebootLog(c), DeepEquals, []string{`reboot`})

	err = pi.Reboot(boot.RebootTryBoot)
	c.Assert(err, IsNil)
	c.Check(rebootLog(c), DeepEquals, []string{`reboot 0\ tryboot`})
}

func (s *bootTest) TestCommitSwitch(c *C) {
	pi, err := piboot.New(s.bootDir)
	c.Assert(err, IsNil)

	s.MockConfigTxt(c, ""+
		"kernel=slot-a/vmlinuz\n"+
		"cmdline=slot-a/cmdline.txt\n"+
		"initramfs slot-a/initrd.img followkernel\n")
	s.MockTrybootTxt(c, ""+
		"kernel=slot-b/vmlinuz\n"+
		"cmdline=slot-b/cmdline.txt\n"+
		"initramfs slot-b/initrd.img followkernel\n")

	err = pi.CommitSwitch()
	c.Assert(err, IsNil)

	slot, err := pi.QueryActive()
	c.Assert(err, IsNil)
	c.Check(slot, Equals, boot.SlotB)

	_, err = os.Stat(filepath.Join(pi.BootMountPoint, "tryboot.txt"))
	c.Assert(os.IsNotExist(err), Equals, true)
}

func (s *bootTest) TestCancelSwitch(c *C) {
	pi, err := piboot.New(s.bootDir)
	c.Assert(err, IsNil)

	s.MockConfigTxt(c, ""+
		"kernel=slot-a/vmlinuz\n"+
		"cmdline=slot-a/cmdline.txt\n"+
		"initramfs slot-a/initrd.img followkernel\n")
	s.MockTrybootTxt(c, ""+
		"kernel=slot-b/vmlinuz\n"+
		"cmdline=slot-b/cmdline.txt\n"+
		"initramfs slot-b/initrd.img followkernel\n")

	err = pi.CancelSwitch()
	c.Assert(err, IsNil)

	slot, err := pi.QueryActive()
	c.Assert(err, IsNil)
	c.Check(slot, Equals, boot.SlotA)

	_, err = os.Stat(filepath.Join(pi.BootMountPoint, "tryboot.txt"))
	c.Assert(os.IsNotExist(err), Equals, true)

	// Idempotent
	err = pi.CancelSwitch()
	c.Assert(err, IsNil)
}

func (s *bootTest) TestVisitSectionsCondfilterErrors(c *C) {
	pi, err := piboot.New(s.bootDir)
	c.Assert(err, IsNil)

	pi.FilterPredicate = func([]condfilter.ConditionalFilter) bool { return true }

	s.MockConfigTxt(c, "[frozbonicator]")

	_, err = pi.QueryActive()
	c.Assert(err, ErrorMatches, `cannot determine active slot: conditional filter error "\[frozbonicator\]": unknown filter`)
}

func (s *bootTest) TestPostInstallHandler(c *C) {
	pi, err := piboot.New(s.bootDir)
	c.Assert(err, IsNil)

	restore, rebootLog := s.MockCommand(c, "reboot")
	defer restore()

	ctx, err := installhandler.NewHandlerContext([]string{
		"RAUC_SLOTS=1",
		"RAUC_TARGET_SLOTS=1",
		"RAUC_SLOT_CLASS_1=potato",
	})
	c.Check(err, IsNil)
	err = pi.PostInstallHandler(ctx)
	c.Check(err, IsNil)
	c.Check(rebootLog(c), DeepEquals, []string(nil))

	ctx, err = installhandler.NewHandlerContext([]string{
		"RAUC_SLOTS=1",
		"RAUC_TARGET_SLOTS=1",
		"RAUC_SLOT_CLASS_1=system",
	})
	c.Assert(err, IsNil)
	err = pi.PostInstallHandler(ctx)
	c.Check(err, IsNil)
	c.Check(rebootLog(c), DeepEquals, []string{`reboot 0\ tryboot`})
}
