// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package cmdrauccustomboothandler_test

import (
	"errors"
	"flag"
	"os"
	"testing"

	dbus "github.com/godbus/dbus/v5"
	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/boot"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/cmd/sysotad/cmdrauccustomboothandler"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/dbusutil"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/dbusutil/dbustest"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/ota"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/ota/raucinterface"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/rauc/rauctest"
)

func Test(t *testing.T) { TestingT(t) }

func mockArgs(args ...string) (restore func()) {
	old := os.Args

	os.Args = append([]string{cmdrauccustomboothandler.Name}, args...)

	return func() {
		os.Args = old
	}
}

type handlerSuite struct {
	dbustest.Suite
	conn        *dbus.Conn
	serviceHost *dbusutil.ServiceHost
	bootHandler rauctest.BootHandler
}

var _ = Suite(&handlerSuite{})

func (s *handlerSuite) TearDownTest(c *C) {
	if s.serviceHost != nil {
		err := s.serviceHost.Unexport()
		c.Check(err, IsNil)

		s.serviceHost = nil
	}

	if s.conn != nil {
		err := s.conn.Close()
		c.Check(err, IsNil)

		s.conn = nil
	}
}

func (s *handlerSuite) EnsureConn(c *C) *dbus.Conn {
	if s.conn == nil {
		var err error
		s.conn, err = dbusutil.SystemBus()
		c.Assert(err, IsNil)
	}

	return s.conn
}

func (s *handlerSuite) SetUpTest(c *C) {
	s.bootHandler.ResetCallbacks()
}

func (s *handlerSuite) EnsureService(c *C) {
	if s.serviceHost == nil {
		conn := s.EnsureConn(c)

		s.serviceHost = dbusutil.NewServiceHost(conn)
		s.serviceHost.AddHostedService(raucinterface.NewService(&s.bootHandler, &rauctest.InstallParticipant{}))

		err := s.serviceHost.Export()
		c.Assert(err, IsNil)

		reply, err := conn.RequestName(ota.ServiceBusName, dbus.NameFlagDoNotQueue)
		c.Assert(err, IsNil)
		c.Assert(reply, Equals, dbus.RequestNameReplyPrimaryOwner)
	}
}

func (s *handlerSuite) TestRunInsufficientArgs(c *C) {
	restore := mockArgs()
	defer restore()

	err := cmdrauccustomboothandler.Run()
	c.Assert(errors.Is(err, flag.ErrHelp), Equals, true)
}

func (s *handlerSuite) TestHelp(c *C) {
	restore := mockArgs("--help")
	defer restore()

	err := cmdrauccustomboothandler.Run()
	c.Assert(errors.Is(err, flag.ErrHelp), Equals, true)
}

func (s *handlerSuite) TestUnknownOptions(c *C) {
	restore := mockArgs("-potato")
	defer restore()

	err := cmdrauccustomboothandler.Run()
	c.Assert(err, ErrorMatches, "flag provided but not defined: -potato")
}

func (s *handlerSuite) TestRunSmoke(c *C) {
	s.EnsureService(c)

	// Just a smoke test, we don't repeat all the tests from boothandler package here.

	called := false
	s.bootHandler.SlotStateFn = func(slot boot.Slot) (boot.SlotState, error) {
		called = true

		c.Check(slot, Equals, boot.SlotA)

		return boot.GoodSlot, nil
	}

	restoreArgs := mockArgs("get-state", "A")
	defer restoreArgs()

	err := cmdrauccustomboothandler.Run()
	c.Assert(err, IsNil)
	c.Check(called, Equals, true)
}

type brokenDBusSuite struct{}

var _ = Suite(&brokenDBusSuite{})

func (s *brokenDBusSuite) TestRunButNoSystemBus(c *C) {
	restoreDBus, err := dbustest.SetDBusSystemBusAddress("unix:path=potato")
	c.Assert(err, IsNil)

	defer func() { c.Assert(restoreDBus(), IsNil) }()

	restoreArgs := mockArgs("get-primary")
	defer restoreArgs()

	err = cmdrauccustomboothandler.Run()
	c.Assert(err, ErrorMatches, `dial unix potato: connect: no such file or directory`)
}
