// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package cmdraucinstallhandler_test

import (
	"os"
	"testing"

	dbus "github.com/godbus/dbus/v5"
	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/cmd/sysotad/cmdraucinstallhandler"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/dbusutil"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/dbusutil/dbustest"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/ota"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/ota/raucinterface"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/rauc/installhandler"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/rauc/rauctest"
)

func Test(t *testing.T) { TestingT(t) }

type handlersSuite struct {
	dbustest.Suite
	conn               *dbus.Conn
	serviceHost        *dbusutil.ServiceHost
	installParticipant rauctest.InstallParticipant
}

var _ = Suite(&handlersSuite{})

func (s *handlersSuite) TearDownTest(c *C) {
	if s.serviceHost != nil {
		err := s.serviceHost.Unexport()
		c.Check(err, IsNil)

		s.serviceHost = nil
	}

	if s.conn != nil {
		err := s.conn.Close()
		c.Check(err, IsNil)

		s.conn = nil
	}
}

func (s *handlersSuite) EnsureConn(c *C) *dbus.Conn {
	if s.conn == nil {
		var err error
		s.conn, err = dbusutil.SystemBus()
		c.Assert(err, IsNil)
	}

	return s.conn
}

func (s *handlersSuite) EnsureService(c *C) {
	if s.serviceHost == nil {
		conn := s.EnsureConn(c)

		s.serviceHost = dbusutil.NewServiceHost(conn)
		s.serviceHost.AddHostedService(raucinterface.NewService(&rauctest.BootHandler{}, &s.installParticipant))

		err := s.serviceHost.Export()
		c.Assert(err, IsNil)

		reply, err := conn.RequestName(ota.ServiceBusName, dbus.NameFlagDoNotQueue)
		c.Assert(err, IsNil)
		c.Assert(reply, Equals, dbus.RequestNameReplyPrimaryOwner)
	}
}

func (s *handlersSuite) TestRunButBrokenEnvironment(c *C) {
	s.EnsureService(c)
	s.installParticipant.PreInstallHandlerFn = func(ctx *installhandler.HandlerContext) error {
		return nil
	}

	c.Assert(os.Setenv("RAUC_SLOTS", "potato"), IsNil)

	defer func() { c.Assert(os.Unsetenv("RAUC_SLOTS"), IsNil) }()

	err := cmdraucinstallhandler.RunPreInstallHandler()
	c.Assert(err, ErrorMatches, `cannot create handler context: variable RAUC_SLOTS refers to undefined slot potato`)
}

func (s *handlersSuite) TestRunPreInstallHandler(c *C) {
	s.EnsureService(c)

	called := false
	s.installParticipant.PreInstallHandlerFn = func(ctx *installhandler.HandlerContext) error {
		called = true

		return nil
	}
	err := cmdraucinstallhandler.RunPreInstallHandler()
	c.Assert(err, IsNil)
	c.Check(called, Equals, true)
}

func (s *handlersSuite) TestRunPostInstallHandler(c *C) {
	s.EnsureService(c)

	called := false
	s.installParticipant.PostInstallHandlerFn = func(ctx *installhandler.HandlerContext) error {
		called = true

		return nil
	}
	err := cmdraucinstallhandler.RunPostInstallHandler()
	c.Assert(err, IsNil)
	c.Check(called, Equals, true)
}

type brokenDBusSuite struct{}

var _ = Suite(&brokenDBusSuite{})

func (s *brokenDBusSuite) TestRunButNoSystemBus(c *C) {
	restore, err := dbustest.SetDBusSystemBusAddress("unix:path=potato")
	c.Assert(err, IsNil)

	defer func() { c.Assert(restore(), IsNil) }()

	err = cmdraucinstallhandler.RunPreInstallHandler()
	c.Assert(err, ErrorMatches, `dial unix potato: connect: no such file or directory`)
}
