// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

// Package cmdsysotad implements the sysotad system service.
package cmdsysotad

import (
	"errors"
	"fmt"
	"io"
	"os"
	"os/signal"
	"syscall"
	"time"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/dbusutil"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/ota"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/service"
)

var (
	// Stdout is os.Stdout but can changed for testing
	Stdout io.Writer = os.Stdout
	// Stderr is os.Stderr but can changed for testing
	Stderr io.Writer = os.Stderr
)

// Options describe input to sysotad necessary to find configuration and state.
type Options struct {
	// IdleDuration defines the frequency of inactivity checks.
	IdleDuration time.Duration
	ConfigFiles  []string
	StateFile    string
}

// DefaultOptions returns options suitable for normal operation.
func DefaultOptions() *Options {
	return &Options{
		IdleDuration: time.Minute,
		ConfigFiles:  ota.DefaultConfigFiles(),
		StateFile:    ota.StateFileName,
	}
}

// Run implements the sysotad binary.
func Run(opts *Options) (err error) {
	// Load configuration.
	conf, err := ota.LoadConfig(opts.ConfigFiles)
	if err != nil {
		return err
	}

	// Load state and defer state save on return.
	state, err := ota.LoadState(opts.StateFile)
	if err != nil {
		return err
	}

	defer func() {
		e := ota.SaveState(state, opts.StateFile)

		// Ignore save errors if the state directory does not exist.
		var pathErr *os.PathError
		if errors.As(e, &pathErr) && os.IsNotExist(pathErr) {
			return
		}

		// Did we save the state successfully?
		if e == nil {
			_, _ = fmt.Fprintf(Stdout, "System state saved to %s\n", opts.StateFile)
			return
		}

		// We failed to save the state. What shall we do with the error (e)?
		if err == nil {
			// Propagate state save error if possible.
			err = e
		} else {
			// Log state save error if it cannot be propagated (the service
			// will exit with a failure code in both cases).
			_, _ = fmt.Fprintf(Stderr, "%v\n", e)
		}
	}()

	// Connect to the system bus and defer disconnect on return.
	conn, err := dbusutil.SystemBus()
	if err != nil {
		return err
	}

	defer func() {
		e := conn.Close()
		if err == nil {
			err = e
		}
	}()

	// Create the SystemOTA service
	svc, err := service.New(conf, state, conn)
	if err != nil {
		return err
	}

	// Export the service to D-Bus having finished configuration.
	if err := svc.Export(); err != nil {
		return err
	}

	// Request bus name after exporting all objects, to signal readiness to systemd.
	if err := service.RequestBusName(conn); err != nil {
		return err
	}

	_, _ = fmt.Fprintf(Stdout, "Listening ...\n")

	// Set up a ticker to check for inactivity.
	idleTicker := time.NewTicker(opts.IdleDuration)
	sigCh := make(chan os.Signal, 1)

	// Set up signal handler to shut down gracefully on SIGINT and SIGTERM.
	signal.Notify(sigCh, os.Interrupt, syscall.SIGTERM)
	defer signal.Stop(sigCh)

	run := true
	for run {
		select {
		// Check if the service is idle and quit if so.
		case <-idleTicker.C:
			if svc.IsIdle() {
				idleTicker.Stop()

				_, _ = fmt.Fprintln(Stdout, "Exiting due to inactivity")
				run = false
			}
		case sigNum := <-sigCh:
			_, _ = fmt.Fprintf(Stdout, "Exiting due to signal %v\n", sigNum)

			run = false
		}
	}

	return nil
}
