// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package dbusutil

import (
	"sync"

	dbus "github.com/godbus/dbus/v5"
	"github.com/godbus/dbus/v5/introspect"
)

// ServiceRegistration is the interface offered by ServiceHost to HostedService.
//
// ServiceRegistration allows a hosted service to register any number of objects
// as well as to obtain the D-Bus connection object, if required.
type ServiceRegistration interface {
	// RegisterObject registers a single object path and interface name.
	//
	// The given object path, interface name, a map of methods to expose (name
	// mapping to a function) are exported immediately. Introspection data, if
	// available, is exported at a later stage, so that each object, regardless of
	// the number of interfaces it offers, is registered once.
	//
	// RegisterObject should be called from HostedService.JoinServiceHost.
	RegisterObject(path dbus.ObjectPath, iface string, methods map[string]interface{}, metadata introspect.Interface) error

	// DBusConn returns the bus connection object.
	//
	// The returned connection may be stored inside the hosted service to allow
	// it to send D-Bus signals.
	//
	// DBusConn should be called from HostedService.JoinServiceHost.
	DBusConn() *dbus.Conn
}

// HostedService is an interface between hosted service and ServiceHost.
//
// The interface defines only one function, JoinServiceHost, which allows an
// object to register itself with the service host. The method is called by
// ServiceHost.Export.
//
// The hosted service can make any number of calls to
// ServiceRegistration.RegisterObject to register distinct objects, distinct
// interfaces or both. In addition the hosted service may call
// ServiceRegistration.DBusConn to obtain remember the D-Bus connection necessary
// to send signals.
type HostedService interface {
	JoinServiceHost(reg ServiceRegistration) error
}

// ServiceHost provides the boilerplate to run a D-Bus service.
//
// ServiceHost handles exporting and un-exporting objects from D-Bus. It is
// suitable for exposing one or more objects and interfaces to the bus,
// arranging to clean up everything on shutdown. ServiceHost does not handle
// bus name registration. That is left as a responsibility of the application.
//
// ServiceHost is useful for creating test services, where the amount of code
// needed for a successful service is minimal compared to a direct approach.
type ServiceHost struct {
	m    sync.RWMutex
	conn *dbus.Conn

	services          []HostedService
	objectsInterfaces map[dbus.ObjectPath]map[string]bool
}

// NewServiceHost returns a new service host.
//
// The connection is not closed by the service host. It is the responsibility of
// the caller.
func NewServiceHost(conn *dbus.Conn) *ServiceHost {
	return &ServiceHost{conn: conn}
}

// AddHostedService adds a given hosted service to the service host.
//
// Any number of hosted services may be added before calling Export. After
// calling Export adding additional services panics. Hosted services cannot be
// removed.
func (sh *ServiceHost) AddHostedService(svc HostedService) {
	sh.m.Lock()
	defer sh.m.Unlock()

	if sh.objectsInterfaces != nil {
		panic("AddHostedService cannot be used after calling Export")
	}

	sh.services = append(sh.services, svc)
}

// registration implements ServiceRegistration.
type registration struct {
	*ServiceHost

	introspection map[dbus.ObjectPath][]introspect.Interface
}

// RegisterObject registers a single object path and interface name.
func (reg registration) RegisterObject(path dbus.ObjectPath, iface string, methods map[string]interface{}, metadata introspect.Interface) error {
	var sh *ServiceHost = reg.ServiceHost

	if err := sh.conn.ExportMethodTable(methods, path, iface); err != nil {
		return err
	}

	if sh.objectsInterfaces[path] == nil {
		sh.objectsInterfaces[path] = make(map[string]bool)
	}

	sh.objectsInterfaces[path][iface] = true

	reg.introspection[path] = append(reg.introspection[path], metadata)

	return nil
}

// DBusConn returns the bus connection object.
func (reg *registration) DBusConn() *dbus.Conn {
	var sh *ServiceHost = reg.ServiceHost

	return sh.conn
}

// Export exports all of the hosted services to the bus.
//
// Each hosted service added with AddHostedService is asked to provide the set
// of object paths, interface names, methods and introspection data by calling
// ServiceRegistration.RegisterObject on the object provided as an argument to
// HostedService.JoinServiceHost.
//
// Introspection data is aggregated by object path and exported as well, using
// the standard introspection interface.
//
// In case any of the operation fails, the objects exported thus far are
// unexported. Any error encountered during the unexport process is ignored.
//
// Export panics if called when there are any objects already exported.
func (sh *ServiceHost) Export() (err error) {
	sh.m.Lock()
	defer sh.m.Unlock()

	if sh.objectsInterfaces != nil {
		panic("Export already called")
	}

	sh.objectsInterfaces = make(map[dbus.ObjectPath]map[string]bool)

	defer func() {
		// introspection is only non-nil until the end of Export.
		if err != nil {
			_ = sh.unexportUnlocked()
		}
	}()

	reg := registration{
		ServiceHost:   sh,
		introspection: make(map[dbus.ObjectPath][]introspect.Interface),
	}

	for _, svc := range sh.services {
		if err := svc.JoinServiceHost(&reg); err != nil {
			return err
		}
	}

	// Export introspection data for each object in one go.
	for path, ifaces := range reg.introspection {
		node := &introspect.Node{Name: string(path), Interfaces: ifaces}
		if err := sh.conn.Export(introspect.NewIntrospectable(node), path, IntrospectableInterfaceName); err != nil {
			return err
		}

		sh.objectsInterfaces[path][IntrospectableInterfaceName] = true
	}

	return nil
}

// Unexport undoes the operations done by Export
//
// For each object path and interface name, the unexport process remembers
// progress, making sure not to unexport the same pair twice. The process stops
// on the first encountered error.
//
// Note that closing the connection with the bus automatically unexports all the
// objects and is a suitable approach to error recovery.
func (sh *ServiceHost) Unexport() error {
	sh.m.Lock()
	defer sh.m.Unlock()

	return sh.unexportUnlocked()
}

func (sh *ServiceHost) unexportUnlocked() error {
	for path, ifaces := range sh.objectsInterfaces {
		for iface, ok := range ifaces {
			if ok {
				if err := sh.conn.Export(nil, path, iface); err != nil {
					return err
				}

				sh.objectsInterfaces[path][iface] = false
			}
		}
	}

	sh.objectsInterfaces = nil

	return nil
}
