// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package dbusutil_test

import (
	"encoding/xml"

	dbus "github.com/godbus/dbus/v5"
	"github.com/godbus/dbus/v5/introspect"
	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/dbusutil"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/dbusutil/dbustest"
)

type serviceHostSuite struct {
	dbustest.Suite

	conn *dbus.Conn
}

var _ = Suite(&serviceHostSuite{})

func (s *serviceHostSuite) SetUpTest(c *C) {
	conn, err := dbusutil.SystemBus()
	c.Assert(err, IsNil)

	s.conn = conn
}

func (s *serviceHostSuite) TearDownTest(c *C) {
	c.Assert(s.conn.Close(), IsNil)
}

type TestService struct {
	conn        *dbus.Conn
	helloCalled bool
}

func (s *serviceHostSuite) TestSmoke(c *C) {
	sh := dbusutil.NewServiceHost(s.conn)
	ts := TestService{}

	sh.AddHostedService(&ts)
	// Bus connection is not relayed yet.
	c.Check(ts.conn, IsNil)

	// Services are exported correctly.
	c.Assert(sh.Export(), IsNil)

	// Register a bus name to make testing easier.
	reply, err := s.conn.RequestName("org.example.Test", dbus.NameFlagDoNotQueue)
	c.Assert(err, IsNil)
	c.Assert(reply, Equals, dbus.RequestNameReplyPrimaryOwner)

	// Bus connection was relayed correctly.
	c.Check(ts.conn, Equals, s.conn)

	// We can call functions correctly.
	obj := s.conn.Object("org.example.Test", "/test")
	err = obj.Call("org.example.TestService.Hello", 0).Store()
	c.Assert(err, IsNil)
	c.Check(ts.helloCalled, Equals, true)

	// Introspection data is correctly published.
	node, err := introspect.Call(obj)
	c.Assert(err, IsNil)
	c.Check(node, DeepEquals, &introspect.Node{
		XMLName: xml.Name{Local: "node"},
		Name:    "/test",
		Interfaces: []introspect.Interface{
			{
				Name: "org.example.TestService",
				Methods: []introspect.Method{
					{Name: "Hello"},
				},
			},
			{
				Name: "org.freedesktop.DBus.Introspectable",
				Methods: []introspect.Method{
					{
						Name: "Introspect", Args: []introspect.Arg{
							{Name: "out", Type: "s", Direction: "out"},
						},
					},
				},
			},
		},
	})

	// We can unexport everthing correctly.
	c.Assert(sh.Unexport(), IsNil)

	// After that we cannot call things anymore.
	err = s.conn.Object("org.example.Test", "/test").Call("org.example.TestService.Hello", 0).Store()
	c.Assert(err, ErrorMatches, "Object does not implement the interface")
}

func (s *serviceHostSuite) TestMisuse(c *C) {
	sh := dbusutil.NewServiceHost(s.conn)

	c.Assert(sh.Export(), IsNil)
	c.Assert(func() { _ = sh.Export() }, PanicMatches, `Export already called`)
	c.Assert(func() { sh.AddHostedService(&TestService{}) }, PanicMatches, `AddHostedService cannot be used after calling Export`)
}

func (ts *TestService) Hello() *dbus.Error {
	ts.helloCalled = true
	return nil
}

func (ts *TestService) JoinServiceHost(reg dbusutil.ServiceRegistration) error {
	ts.conn = reg.DBusConn()

	return reg.RegisterObject("/test", "org.example.TestService", map[string]interface{}{
		"Hello": ts.Hello,
	}, introspect.Interface{
		Name: "org.example.TestService",
		Methods: []introspect.Method{
			{
				Name: "Hello",
			},
		},
	})
}
