// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package ini

import (
	"bufio"
	"bytes"
	"encoding"
	"errors"
	"fmt"
	"io"
	"reflect"
	"strconv"
	"strings"
)

var (
	// ErrEmptyFieldName records assignments with empty field name.
	ErrEmptyFieldName = errors.New("empty field name")
	// ErrUnterminatedContinuationAtEOF records unterminated multi-line entry.
	ErrUnterminatedContinuationAtEOF = errors.New("unterminated continuation at end of input")
)

// DecoderError records problem encountered during parsing.
type DecoderError struct {
	LineNo int
	Err    error
}

// Unwrap returns the error wrapped inside the decoder error.
func (decErr *DecoderError) Unwrap() error {
	return decErr.Err
}

// Error implements the error interface.
func (decErr *DecoderError) Error() string {
	return fmt.Sprintf("cannot decode line %d: %s", decErr.LineNo, decErr.Err)
}

// UnexpectedInputError records input not expected by the parser.
type UnexpectedInputError []byte

// Error implements the error interface.
func (e UnexpectedInputError) Error() string {
	return fmt.Sprintf("unexpected input: %q", string(e))
}

// UnknownSectionError records an unknown section.
//
// Unknown sections are ignored unless Decoder.DisallowUnknownSections is
// called on a new decoder instance.
type UnknownSectionError string

// SectionName returns the name of the section with '[' and ']' stripped out.
func (e UnknownSectionError) SectionName() string {
	return string(e)
}

// Error implements the error interface.
func (e UnknownSectionError) Error() string {
	return fmt.Sprintf("unknown section: %s", e.SectionName())
}

// UnknownFieldError records an unknown field.
//
// Unknown fields are ignored unless Decoder.DisallowUnknownFields is
// called on a new decoder instance.
type UnknownFieldError string

// FieldName returns the name of the field.
func (e UnknownFieldError) FieldName() string {
	return string(e)
}

// Error implements the error interface.
func (e UnknownFieldError) Error() string {
	return fmt.Sprintf("unknown field: %s", e.FieldName())
}

// FieldDecodeError records failed attempt to decode a structure field.
type FieldDecodeError struct {
	StructName string
	FieldName  string
	Err        error
}

// Error implements error.
func (e *FieldDecodeError) Error() string {
	return fmt.Sprintf("cannot decode %s.%s: %v", e.StructName, e.FieldName, e.Err)
}

// Unwrap returns the error wrapped by field decode error.
func (e *FieldDecodeError) Unwrap() error {
	return e.Err
}

// Decoder reads INI values from an input stream.
type Decoder struct {
	r                       io.Reader
	disallowUnknownFields   bool
	disallowUnknownSections bool
	caseSensitive           bool
	commentRune             rune
}

// NewDecoder returns a new decoder that reads from r.
func NewDecoder(r io.Reader) *Decoder {
	return &Decoder{r: r}
}

// DisallowUnknownSections makes the decoder reject unknown section names.
func (dec *Decoder) DisallowUnknownSections() {
	dec.disallowUnknownSections = true
}

// DisallowUnknownFields makes the decoder reject unknown field names.
func (dec *Decoder) DisallowUnknownFields() {
	dec.disallowUnknownFields = true
}

// UseUnixComments makes the decoder recognize comments introduced by '#'.
func (dec *Decoder) UseUnixComments() {
	dec.commentRune = '#'
}

// UseWindowsComments makes the decoder recognize comments introduced by ';'.
func (dec *Decoder) UseWindowsComments() {
	dec.commentRune = ';'
}

// CaseSensitive makes the decoder case-sensitive for both sections and field names.
func (dec *Decoder) CaseSensitive() {
	dec.caseSensitive = true
}

// Decode decodes a structure from the INI representation.
//
// The argument must be a pointer to a structure.
//
// Comments introduced with '#' or ';' are ignored. If a preference for certain
// style has been set, then only such comments are recognized.
//
// See the documentation of Unmarshal more details.
func (dec *Decoder) Decode(v interface{}) error {
	if reflect.TypeOf(v).Kind() != reflect.Ptr || reflect.TypeOf(v).Elem().Kind() != reflect.Struct {
		return fmt.Errorf("cannot decode values of type %T", v)
	}

	structValue := reflect.Indirect(reflect.ValueOf(v))
	structType := structValue.Type()

	iniFieldSlice, iniFieldIndexMap, err := iniMapOfType(structType, dec.caseSensitive)
	if err != nil {
		return err
	}

	scanner := bufio.NewScanner(dec.r)

	var sectionRealName, sectionLookupName string

	var accumulationBuf bytes.Buffer

	var accumulating bool

	var lineno int

	for lineno = 1; scanner.Scan(); lineno++ {
		rawLine := scanner.Bytes()

		// Trim whitespace. In general whitespace is irrelevant, including
		// around the '=' rune in key=value pairs below.
		rawLine = bytes.TrimSpace(rawLine)

		// Skip lines that are entirely empty.
		if len(rawLine) == 0 {
			continue
		}

		// Skip comments. Those take priority over line continuations, allowing comments
		// in the middle of input. If a preference was set, recognize only specific type.
		// If no preference was set recognize both '#' and ';' comments, assuming they
		// are the first rune of a line.
		commentFound := false

		switch dec.commentRune {
		case '#':
			commentFound = bytes.HasPrefix(rawLine, []byte("#"))
		case ';':
			commentFound = bytes.HasPrefix(rawLine, []byte(";"))
		default:
			commentFound = bytes.HasPrefix(rawLine, []byte("#")) || bytes.HasPrefix(rawLine, []byte(";"))
		}

		if commentFound {
			continue
		}

		// Handle continuations from previous line. Long lines can be split into
		// multiple sections, each but the last one terminated with a '\' rune.
		// If the continuation buffer is not empty, then the first line without
		// the continuation rune terminates the sequence.
		if bytes.HasSuffix(rawLine, []byte("\\")) {
			accumulationBuf.Write(rawLine[:len(rawLine)-1])
			accumulationBuf.WriteRune(' ')

			accumulating = true

			continue
		} else if accumulating {
			accumulationBuf.Write(rawLine)
			rawLine = accumulationBuf.Bytes()
			accumulationBuf.Reset()
			accumulating = false
		}

		// Handle [sections].
		if bytes.HasPrefix(rawLine, []byte(`[`)) && bytes.HasSuffix(rawLine, []byte(`]`)) {
			sectionRealName = string(rawLine[1 : len(rawLine)-1])
			sectionLookupName = sectionRealName

			if !dec.caseSensitive {
				sectionLookupName = strings.ToLower(sectionLookupName)
			}

			if _, ok := iniFieldIndexMap[sectionLookupName]; !ok && dec.disallowUnknownSections {
				return &DecoderError{LineNo: lineno, Err: UnknownSectionError(sectionRealName)}
			}

			continue
		}

		// Handle key=value pairs.
		if idx := bytes.IndexRune(rawLine, '='); idx != -1 {
			fieldRealName := string(bytes.TrimSpace(rawLine[:idx]))
			fieldLookupName := fieldRealName

			if len(fieldRealName) == 0 {
				return &DecoderError{LineNo: lineno, Err: ErrEmptyFieldName}
			}

			fieldRawValue := bytes.TrimSpace(rawLine[idx+1:])

			if !dec.caseSensitive {
				fieldLookupName = strings.ToLower(fieldLookupName)
			}

			if iniFieldIndex, ok := iniFieldIndexMap[sectionLookupName][fieldLookupName]; ok {
				if err := dec.decodeField(fieldRawValue, structType, &structValue, &iniFieldSlice[iniFieldIndex]); err != nil {
					return err
				}
			} else if dec.disallowUnknownFields {
				return &DecoderError{LineNo: lineno, Err: UnknownFieldError(fieldRealName)}
			}

			continue
		}

		// Report everything else as syntax error.
		return &DecoderError{LineNo: lineno, Err: UnexpectedInputError(rawLine)}
	}

	if accumulating {
		return &DecoderError{LineNo: lineno, Err: ErrUnterminatedContinuationAtEOF}
	}

	if err := scanner.Err(); err != nil {
		return err
	}

	return nil
}

// StringUnquoteError records problem with removing quoting marks from a string.
type StringUnquoteError struct {
	Err error
}

// Error implements the error interface.
func (e *StringUnquoteError) Error() string {
	return fmt.Sprintf("cannot unquote string: %v", e.Err)
}

// Unwrap returns the error wrapped inside the string unquote error.
func (e *StringUnquoteError) Unwrap() error {
	return e.Err
}

func (dec *Decoder) decodeField(valueData []byte, structType reflect.Type, structValue *reflect.Value, iniFieldInfo *iniFieldInfo) (err error) {
	structField := structType.FieldByIndex(iniFieldInfo.FieldIndex)
	indirectFieldValue := reflect.Indirect(structValue.FieldByIndex(iniFieldInfo.FieldIndex))

	if reflect.PtrTo(structField.Type).Implements(textUnmarshalerType) {
		err = structValue.FieldByIndex(iniFieldInfo.FieldIndex).Addr().Interface().(encoding.TextUnmarshaler).UnmarshalText(valueData)
	} else {
		valueDataText := string(valueData)
		switch kind := structField.Type.Kind(); kind {
		case reflect.Bool:
			var boolValue bool
			boolValue, err = strconv.ParseBool(valueDataText)
			if err == nil {
				indirectFieldValue.SetBool(boolValue)
			}
		case reflect.Int, reflect.Int8, reflect.Int16, reflect.Int32, reflect.Int64:
			var intValue int64
			bitSize := 0
			switch kind {
			case reflect.Int8:
				bitSize = 8
			case reflect.Int16:
				bitSize = 16
			case reflect.Int32:
				bitSize = 32
			case reflect.Int64:
				bitSize = 64
			}
			intValue, err = strconv.ParseInt(valueDataText, 10, bitSize)
			if err == nil {
				indirectFieldValue.SetInt(int64(intValue))
			}
		case reflect.Uint, reflect.Uint8, reflect.Uint16, reflect.Uint32, reflect.Uint64:
			var uintValue uint64
			bitSize := 0
			switch kind {
			case reflect.Uint8:
				bitSize = 8
			case reflect.Uint16:
				bitSize = 16
			case reflect.Uint32:
				bitSize = 32
			case reflect.Uint64:
				bitSize = 64
			}
			uintValue, err = strconv.ParseUint(valueDataText, 10, bitSize)
			if err == nil {
				indirectFieldValue.SetUint(uint64(uintValue))
			}
		case reflect.Float32, reflect.Float64:
			var floatValue float64
			bitSize := 32
			if kind == reflect.Float64 {
				bitSize = 64
			}
			floatValue, err = strconv.ParseFloat(valueDataText, bitSize)
			if err == nil {
				indirectFieldValue.SetFloat(floatValue)
			}
		case reflect.String:
			strValue := valueDataText
			if iniFieldInfo.Quote {
				var unquotedStrValue string
				unquotedStrValue, err = strconv.Unquote(strValue)
				if err != nil {
					err = &StringUnquoteError{Err: err}
				} else {
					strValue = unquotedStrValue
				}
			}
			if err == nil {
				structValue.FieldByIndex(iniFieldInfo.FieldIndex).SetString(strValue)
			}
		default:
			// This can never happen if iniSliceOfType is in sync.
			panic(fmt.Sprintf("cannot decode value of unsupported kind %v", kind))
		}
	}

	if err != nil {
		return &FieldDecodeError{StructName: structType.Name(), FieldName: structField.Name, Err: err}
	}

	return nil
}
