// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package ini_test

import (
	"bytes"
	"errors"
	"math"
	"reflect"
	"strconv"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/encoding/ini"
	. "gopkg.in/check.v1"
)

type decoderSuite struct{}

var _ = Suite(&decoderSuite{})

func (s *decoderSuite) TestDecodeNonStructType(c *C) {
	var value int

	dec := ini.NewDecoder(bytes.NewBufferString(""))
	err := dec.Decode(&value)
	c.Assert(err, ErrorMatches, "cannot decode values of type .*")
}

func (s *decoderSuite) TestDecodeNonPointerType(c *C) {
	dec := ini.NewDecoder(bytes.NewBufferString(""))
	err := dec.Decode(struct{}{})
	c.Assert(err, ErrorMatches, "cannot decode values of type .*")
}

func (s *decoderSuite) TestUseUnixComments(c *C) {
	dec := ini.NewDecoder(bytes.NewBufferString("# This is an UNIX comment"))
	dec.UseUnixComments()
	err := dec.Decode(&struct{}{})
	c.Assert(err, IsNil)

	dec = ini.NewDecoder(bytes.NewBufferString("; This is a Windows comment"))
	dec.UseUnixComments()
	err = dec.Decode(&struct{}{})
	c.Assert(err, ErrorMatches, `cannot decode line 1: unexpected input: "; This is a Windows comment"`)
}

func (s *decoderSuite) TestUseWindowsComments(c *C) {
	dec := ini.NewDecoder(bytes.NewBufferString("; This is a Windows comment"))
	dec.UseWindowsComments()
	err := dec.Decode(&struct{}{})
	c.Assert(err, IsNil)

	dec = ini.NewDecoder(bytes.NewBufferString("# This is an UNIX comment"))
	dec.UseWindowsComments()
	err = dec.Decode(&struct{}{})
	c.Assert(err, ErrorMatches, `cannot decode line 1: unexpected input: "# This is an UNIX comment"`)
}

func (s *decoderSuite) TestDecodeCommentTypes(c *C) {
	for _, input := range []string{
		"# First comment\n# Second comment\n",
		"; First comment\n; Second comment\n",
		"# First comment\n; Second comment\n",
		"; First comment\n# Second comment\n",
	} {
		dec := ini.NewDecoder(bytes.NewBufferString(input))
		err := dec.Decode(&struct{}{})
		c.Assert(err, IsNil, Commentf("input: %q", input))
	}
}

func (s *decoderSuite) TestDecodeContinuations(c *C) {
	type testType struct {
		Field string `ini:"F"`
	}

	for _, input := range []string{
		"F=v a l u e",
		"F=v a l\\\nu e",
		"F=v a l\\\n  u e",
		"F=v a l\\\n  u e  ",
		"F=\\\nv a l u e",
		"F=\\\nv\\\na\\\nl\\\nu\\\ne",
		"F=\\\n# comment\nv\\\na\\\nl\\\n;comment\nu\\\ne",
	} {
		comment := Commentf("input: %q", input)
		dec := ini.NewDecoder(bytes.NewBufferString(input))

		var value testType

		err := dec.Decode(&value)
		c.Assert(err, IsNil, comment)
		c.Check(value.Field, Equals, "v a l u e", comment)
	}
}

func (s *decoderSuite) TestDecodeContinuationsDangling(c *C) {
	dec := ini.NewDecoder(bytes.NewBufferString("\\"))
	err := dec.Decode(&struct{}{})
	c.Assert(err, ErrorMatches, `cannot decode line 2: unterminated continuation at end of input`)
	c.Assert(errors.Unwrap(err), Equals, ini.ErrUnterminatedContinuationAtEOF)
}

func (s *decoderSuite) TestDecodeEmptyLines(c *C) {
	for _, input := range []string{
		"",
		"\n",
		"\n  \n",
		"\r\n",
		"\t",
		"\v",
		" ",
		"\r\n\t\v ",
	} {
		comment := Commentf("input: %q", input)
		dec := ini.NewDecoder(bytes.NewBufferString(input))
		err := dec.Decode(&struct{}{})
		c.Assert(err, IsNil, comment)
	}
}

func (s *decoderSuite) TestDecodeFieldWithoutValue(c *C) {
	dec := ini.NewDecoder(bytes.NewBufferString("FieldWithoutValue"))
	err := dec.Decode(&struct{}{})
	c.Assert(err, ErrorMatches, `cannot decode line 1: unexpected input: "FieldWithoutValue"`)
}

func (s *decoderSuite) TestDecodeValueWithoutField(c *C) {
	dec := ini.NewDecoder(bytes.NewBufferString("=ValueWithoutField"))
	err := dec.Decode(&struct{}{})
	c.Assert(err, ErrorMatches, `cannot decode line 1: empty field name`)
	c.Assert(errors.Unwrap(err), Equals, ini.ErrEmptyFieldName)
}

func (s *decoderSuite) TestDisallowUnknownSections(c *C) {
	dec := ini.NewDecoder(bytes.NewBufferString("[Section]\n"))
	dec.DisallowUnknownSections()
	err := dec.Decode(&struct{}{})
	c.Assert(err, ErrorMatches, `cannot decode line 1: unknown section: Section`)
	c.Assert(errors.Unwrap(err), Equals, ini.UnknownSectionError("Section"))
}

func (s *decoderSuite) TestDisallowUnknownFields(c *C) {
	dec := ini.NewDecoder(bytes.NewBufferString("Field=Value\n"))
	dec.DisallowUnknownFields()
	err := dec.Decode(&struct{}{})
	c.Assert(err, ErrorMatches, `cannot decode line 1: unknown field: Field`)
	c.Assert(errors.Unwrap(err), Equals, ini.UnknownFieldError("Field"))
}

func (s *decoderSuite) TestSectionTracking(c *C) {
	dec := ini.NewDecoder(bytes.NewBufferString("" +
		"[Integers]\n" +
		"IntField1=1\n" +
		"[Surprise]\n" +
		"IntField3=3\n"))

	type testType struct {
		IntField1 int `ini:"IntField1,[Integers]"`
		IntField2 int `ini:"IntField2,[Surprise],omitempty"`
		IntField3 int `ini:"IntField3"`
	}

	var value testType

	err := dec.Decode(&value)
	c.Assert(err, IsNil)
	c.Check(value.IntField1, Equals, 1)
	c.Check(value.IntField2, Equals, 0)
	c.Check(value.IntField3, Equals, 3)
}

func (s *decoderSuite) TestCaseInsensitivity(c *C) {
	dec := ini.NewDecoder(bytes.NewBufferString("" +
		"[SECTION]\n" +
		"FIELD=ok\n"))
	dec.DisallowUnknownFields()
	dec.DisallowUnknownSections()

	type testType struct {
		Field string `ini:"Field,[Section]"`
	}

	var value testType

	err := dec.Decode(&value)
	c.Assert(err, IsNil)
	c.Check(value, Equals, testType{Field: "ok"})
}

func (s *decoderSuite) TestCaseSensitivity(c *C) {
	dec := ini.NewDecoder(bytes.NewBufferString("" +
		"[SECTION]\n" +
		"FIELD=ok\n"))
	dec.CaseSensitive()
	dec.DisallowUnknownFields()
	dec.DisallowUnknownSections()

	type testType struct {
		Field string `ini:"Field,[Section]"`
	}

	var value testType

	err := dec.Decode(&value)
	c.Assert(err, ErrorMatches, `cannot decode line 1: unknown section: SECTION`)

	dec = ini.NewDecoder(bytes.NewBufferString("" +
		"[Section]\n" +
		"FIELD=ok\n"))
	dec.CaseSensitive()
	dec.DisallowUnknownFields()
	dec.DisallowUnknownSections()
	err = dec.Decode(&value)
	c.Assert(err, ErrorMatches, `cannot decode line 2: unknown field: FIELD`)

	dec = ini.NewDecoder(bytes.NewBufferString("" +
		"[Section]\n" +
		"Field=ok\n"))
	dec.CaseSensitive()
	dec.DisallowUnknownFields()
	dec.DisallowUnknownSections()
	err = dec.Decode(&value)
	c.Assert(err, IsNil)
	c.Check(value, Equals, testType{Field: "ok"})
}

func (s *decoderSuite) TestDecodeBoolean(c *C) {
	type testType struct {
		Field bool `ini:"F"`
	}

	var value testType

	dec := ini.NewDecoder(bytes.NewBufferString("F=true"))
	c.Assert(dec.Decode(&value), IsNil)
	c.Check(value.Field, Equals, true)

	dec = ini.NewDecoder(bytes.NewBufferString("F=false"))
	c.Assert(dec.Decode(&value), IsNil)
	c.Check(value.Field, Equals, false)

	dec = ini.NewDecoder(bytes.NewBufferString("F=potato"))
	err := dec.Decode(&value)
	c.Assert(err, ErrorMatches, `cannot decode testType.Field: strconv.ParseBool: parsing "potato": invalid syntax`)
	c.Assert(errors.Unwrap(err), DeepEquals, &strconv.NumError{
		Func: "ParseBool",
		Num:  "potato",
		Err:  strconv.ErrSyntax,
	})
}

func (s *decoderSuite) TestDecodeInt(c *C) {
	type testCase struct {
		value interface{}
		min   int64
		max   int64
	}

	for _, tc := range []testCase{
		{
			value: &struct {
				Field int `ini:"F"`
			}{},
			min: MinInt,
			max: MaxInt,
		},
		{
			value: &struct {
				Field int8 `ini:"F"`
			}{},
			min: math.MinInt8,
			max: math.MaxInt8,
		},
		{
			value: &struct {
				Field int16 `ini:"F"`
			}{},
			min: math.MinInt16,
			max: math.MaxInt16,
		},
		{
			value: &struct {
				Field int32 `ini:"F"`
			}{},
			min: math.MinInt32,
			max: math.MaxInt32,
		},
		{
			value: &struct {
				Field int64 `ini:"F"`
			}{},
			min: math.MinInt64,
			max: math.MaxInt64,
		},
	} {
		comment := Commentf("test case: %#v", tc)

		dec := ini.NewDecoder(bytes.NewBufferString("F=" + strconv.FormatInt(tc.min, 10)))
		c.Assert(dec.Decode(tc.value), IsNil, comment)
		c.Check(reflect.ValueOf(tc.value).Elem().FieldByName("Field").Int(), Equals, tc.min, comment)

		dec = ini.NewDecoder(bytes.NewBufferString("F=" + strconv.FormatInt(tc.max, 10)))
		c.Assert(dec.Decode(tc.value), IsNil, comment)
		c.Check(reflect.ValueOf(tc.value).Elem().FieldByName("Field").Int(), Equals, tc.max, comment)

		// We cannot compute math.MinInt64-1 without underflowing.
		if tc.min > math.MinInt64 {
			dec = ini.NewDecoder(bytes.NewBufferString("F=" + strconv.FormatInt(tc.min-1, 10)))
			err := dec.Decode(tc.value)
			c.Assert(err, ErrorMatches, `cannot decode .Field: strconv.ParseInt: parsing "-[0-9]+": value out of range`, comment)
			c.Assert(errors.Unwrap(err), DeepEquals, &strconv.NumError{
				Func: "ParseInt",
				Num:  strconv.FormatInt(tc.min-1, 10),
				Err:  strconv.ErrRange,
			}, comment)
		}

		// We cannot compute math.MaxInt64+1 without overflowing.
		if tc.max < math.MaxInt64 {
			dec = ini.NewDecoder(bytes.NewBufferString("F=" + strconv.FormatInt(tc.max+1, 10)))
			err := dec.Decode(tc.value)
			c.Assert(err, ErrorMatches, `cannot decode .Field: strconv.ParseInt: parsing "[0-9]+": value out of range`, comment)
			c.Assert(errors.Unwrap(err), DeepEquals, &strconv.NumError{
				Func: "ParseInt",
				Num:  strconv.FormatInt(tc.max+1, 10),
				Err:  strconv.ErrRange,
			}, comment)
		}

		dec = ini.NewDecoder(bytes.NewBufferString("F=potato"))
		err := dec.Decode(tc.value)
		c.Assert(err, ErrorMatches, `cannot decode .Field: strconv.ParseInt: parsing "potato": invalid syntax`, comment)
		c.Assert(errors.Unwrap(err), DeepEquals, &strconv.NumError{
			Func: "ParseInt",
			Num:  "potato",
			Err:  strconv.ErrSyntax,
		}, comment)
	}
}

func (s *decoderSuite) TestDecodeUint(c *C) {
	type testCase struct {
		value interface{}
		max   uint64
	}

	for _, tc := range []testCase{
		{
			value: &struct {
				Field uint `ini:"F"`
			}{},
			max: MaxUint,
		},
		{
			value: &struct {
				Field uint8 `ini:"F"`
			}{},
			max: math.MaxUint8,
		},
		{
			value: &struct {
				Field uint16 `ini:"F"`
			}{},
			max: math.MaxUint16,
		},
		{
			value: &struct {
				Field uint32 `ini:"F"`
			}{},
			max: math.MaxUint32,
		},
		{
			value: &struct {
				Field uint64 `ini:"F"`
			}{},
			max: math.MaxUint64,
		},
	} {
		comment := Commentf("test case: %#v", tc)

		dec := ini.NewDecoder(bytes.NewBufferString("F=0"))
		c.Assert(dec.Decode(tc.value), IsNil, comment)
		c.Check(reflect.ValueOf(tc.value).Elem().FieldByName("Field").Uint(), Equals, uint64(0), comment)

		dec = ini.NewDecoder(bytes.NewBufferString("F=" + strconv.FormatUint(tc.max, 10)))
		c.Assert(dec.Decode(tc.value), IsNil, comment)
		c.Check(reflect.ValueOf(tc.value).Elem().FieldByName("Field").Uint(), Equals, tc.max, comment)

		// We cannot compute math.MaxUint64+1 without overflowing.
		if tc.max < math.MaxUint64 {
			dec = ini.NewDecoder(bytes.NewBufferString("F=" + strconv.FormatUint(tc.max+1, 10)))
			err := dec.Decode(tc.value)
			c.Assert(err, ErrorMatches, `cannot decode .Field: strconv.ParseUint: parsing "[0-9]+": value out of range`, comment)
			c.Assert(errors.Unwrap(err), DeepEquals, &strconv.NumError{
				Func: "ParseUint",
				Num:  strconv.FormatUint(tc.max+1, 10),
				Err:  strconv.ErrRange,
			}, comment)
		}

		dec = ini.NewDecoder(bytes.NewBufferString("F=potato"))
		err := dec.Decode(tc.value)
		c.Assert(err, ErrorMatches, `cannot decode .Field: strconv.ParseUint: parsing "potato": invalid syntax`, comment)
		c.Assert(errors.Unwrap(err), DeepEquals, &strconv.NumError{
			Func: "ParseUint",
			Num:  "potato",
			Err:  strconv.ErrSyntax,
		}, comment)
	}
}

func (s *decoderSuite) TestDecodeFloat(c *C) {
	type testCase struct {
		value   interface{}
		min     float64
		max     float64
		bitSize int
	}

	for _, tc := range []testCase{
		{
			value: &struct {
				Field float32 `ini:"F"`
			}{},
			min:     math.SmallestNonzeroFloat32,
			max:     math.MaxFloat32,
			bitSize: 32,
		},
		{
			value: &struct {
				Field float64 `ini:"F"`
			}{},
			min:     math.SmallestNonzeroFloat64,
			max:     math.MaxFloat64,
			bitSize: 64,
		},
	} {
		comment := Commentf("test case: %#v", tc)

		dec := ini.NewDecoder(bytes.NewBufferString("F=" + strconv.FormatFloat(tc.min, 'x', -1, tc.bitSize)))
		c.Assert(dec.Decode(tc.value), IsNil, comment)
		c.Check(reflect.ValueOf(tc.value).Elem().FieldByName("Field").Float(), Equals, tc.min, comment)

		dec = ini.NewDecoder(bytes.NewBufferString("F=" + strconv.FormatFloat(tc.max, 'x', -1, tc.bitSize)))
		c.Assert(dec.Decode(tc.value), IsNil, comment)
		c.Check(reflect.ValueOf(tc.value).Elem().FieldByName("Field").Float(), Equals, tc.max, comment)

		dec = ini.NewDecoder(bytes.NewBufferString("F=potato"))
		err := dec.Decode(tc.value)
		c.Assert(err, ErrorMatches, `cannot decode .Field: strconv.ParseFloat: parsing "potato": invalid syntax`, comment)
		c.Assert(errors.Unwrap(err), DeepEquals, &strconv.NumError{
			Func: "ParseFloat",
			Num:  "potato",
			Err:  strconv.ErrSyntax,
		}, comment)
	}
}

func (s *decoderSuite) TestDecodeString(c *C) {
	type testCase struct {
		value      interface{}
		textIn     string
		textOut    string
		errPattern string
	}

	for _, tc := range []testCase{
		{
			value: &struct {
				Field string `ini:"F"`
			}{},
			textIn:  "hello world",
			textOut: "hello world",
		},
		{
			value: &struct {
				Field string `ini:"F,quote"`
			}{},
			textIn:  `"hello world"`,
			textOut: "hello world",
		},
		{
			value: &struct {
				Field string `ini:"F,quote"`
			}{},
			textIn:     `"hello world`,
			errPattern: `cannot decode .Field: cannot unquote string: invalid syntax`,
		},
	} {
		comment := Commentf("test case: %#v", tc)

		dec := ini.NewDecoder(bytes.NewBufferString("F=" + tc.textIn))
		if tc.errPattern == "" {
			c.Assert(dec.Decode(tc.value), IsNil, comment)
			c.Check(reflect.ValueOf(tc.value).Elem().FieldByName("Field").String(), Equals, tc.textOut, comment)
		} else {
			err := dec.Decode(tc.value)
			c.Assert(err, ErrorMatches, tc.errPattern, comment)
			c.Check(errors.Is(err, strconv.ErrSyntax), Equals, true)
		}
	}

	dec := ini.NewDecoder(bytes.NewBufferString(`F="broken`))
	err := dec.Decode(&struct {
		Field string `ini:"F,quote"`
	}{})
	c.Assert(err, NotNil)
	c.Assert(errors.Unwrap(err), ErrorMatches, `cannot unquote string: invalid syntax`)
}

func (s *decoderSuite) TestDecodeTextUnmarshaler(c *C) {
	type testType struct {
		Field testEnum `ini:"F"`
	}

	var value testType

	dec := ini.NewDecoder(bytes.NewBufferString("F=foo"))
	c.Assert(dec.Decode(&value), IsNil)
	c.Check(value.Field, Equals, testFoo)

	dec = ini.NewDecoder(bytes.NewBufferString("F=bar"))
	c.Assert(dec.Decode(&value), IsNil)
	c.Check(value.Field, Equals, testBar)

	dec = ini.NewDecoder(bytes.NewBufferString("F=froz"))
	c.Assert(dec.Decode(&value), ErrorMatches, `cannot decode testType.Field: cannot unmarshal testEnum: unknown value: "froz"`)
}
