// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package ini

import (
	"encoding"
	"errors"
	"fmt"
	"io"
	"reflect"
	"strconv"
	"strings"
)

// Encoder writes INI value to an output stream.
type Encoder struct {
	w io.Writer
}

// NewEncoder returns a new encoder that writes to w.
func NewEncoder(w io.Writer) *Encoder {
	return &Encoder{w: w}
}

var (
	// ErrEncodeNil reports attempt to encode a nil value
	ErrEncodeNil = errors.New("cannot encode nil values")
)

// Encode encodes a structure to the INI representation.
func (enc *Encoder) Encode(v interface{}) error {
	if v == nil {
		return ErrEncodeNil
	}

	structValue := reflect.ValueOf(v)
	if structValue.Kind() == reflect.Ptr {
		structValue = reflect.Indirect(structValue)
	}

	if structValue.Type().Kind() != reflect.Struct {
		return fmt.Errorf("cannot encode values of type: %T", v)
	}

	structType := structValue.Type()

	iniFieldSlice, err := iniSliceOfType(structType)
	if err != nil {
		return err
	}

	lastSection := ""

	for _, iniFieldInfo := range iniFieldSlice {
		structField := structType.FieldByIndex(iniFieldInfo.FieldIndex)

		fieldValue := structValue.FieldByIndex(iniFieldInfo.FieldIndex)
		if fieldValue.IsZero() && iniFieldInfo.OmitEmpty {
			continue
		}

		var valueDataText string

		if textMarshaler, ok := fieldValue.Interface().(encoding.TextMarshaler); ok {
			var valueData []byte

			valueData, err = textMarshaler.MarshalText()
			if err == nil {
				valueDataText = string(valueData)
			}
		} else {
			switch kind := structField.Type.Kind(); kind {
			case reflect.Bool:
				valueDataText = strconv.FormatBool(fieldValue.Bool())
			case reflect.Int, reflect.Int8, reflect.Int16, reflect.Int32, reflect.Int64:
				valueDataText = strconv.FormatInt(fieldValue.Int(), 10)
			case reflect.Uint, reflect.Uint8, reflect.Uint16, reflect.Uint32, reflect.Uint64:
				valueDataText = strconv.FormatUint(fieldValue.Uint(), 10)
			case reflect.Float32, reflect.Float64:
				bitSize := 32
				if kind == reflect.Float64 {
					bitSize = 64
				}
				valueDataText = strconv.FormatFloat(fieldValue.Float(), 'f', -1, bitSize)
			case reflect.String:
				valueDataText = fieldValue.String()
				if iniFieldInfo.Quote {
					valueDataText = strconv.Quote(valueDataText)
				} else {
					prefix := strings.Repeat(" ", len(iniFieldInfo.Name)+1)
					valueDataText = strings.ReplaceAll(valueDataText, "\n", "\\\n"+prefix)
				}
			default:
				// This can never happen if iniSliceOfType is in sync.
				panic(fmt.Sprintf("cannot encode value of unsupported kind %v", kind))
			}
		}

		if err == nil {
			if iniFieldInfo.Section != lastSection {
				_, err = fmt.Fprintf(enc.w, "[%s]\n", iniFieldInfo.Section)
				lastSection = iniFieldInfo.Section
			}

			if err == nil {
				_, err = fmt.Fprintf(enc.w, "%s=%s\n", iniFieldInfo.Name, valueDataText)
			}
		}

		if err != nil {
			return fmt.Errorf("cannot encode %s.%s: %w", structType.Name(), structField.Name, err)
		}
	}

	return nil
}
