// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package ini_test

import (
	"bytes"
	"errors"
	"math"
	"reflect"
	"strconv"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/encoding/ini"
	. "gopkg.in/check.v1"
)

type encoderSuite struct {
	buf bytes.Buffer
}

var _ = Suite(&encoderSuite{})

func (s *encoderSuite) SetUpTest(c *C) {
	s.buf.Reset()
}

func (s *encoderSuite) TestEncodeNilValue(c *C) {
	enc := ini.NewEncoder(&s.buf)
	err := enc.Encode(nil)
	c.Assert(err, ErrorMatches, `cannot encode nil values`)
	c.Assert(errors.Is(err, ini.ErrEncodeNil), Equals, true)
}

func (s *encoderSuite) TestEncodeNonStructurePointer(c *C) {
	var i int

	enc := ini.NewEncoder(&s.buf)
	c.Assert(enc.Encode(&i), ErrorMatches, `cannot encode values of type: \*int`)
}

func (s *encoderSuite) TestEncodeSkipsUntaggedFields(c *C) {
	type testType struct {
		Field int
	}

	enc := ini.NewEncoder(&s.buf)
	c.Assert(enc.Encode(&testType{Field: 42}), IsNil)
	c.Check(s.buf.String(), Equals, "")
}

func (s *encoderSuite) TestEncodeSkipsUnexportedFields(c *C) {
	type testType struct {
		field int `ini:"f"`
	}

	enc := ini.NewEncoder(&s.buf)
	c.Assert(enc.Encode(&testType{field: 42}), IsNil)
	c.Check(s.buf.String(), Equals, "")
}

func (s *encoderSuite) TestEncodeVariousTypes(c *C) {
	type testType struct {
		IntField    int    `ini:"IntField"`
		BoolField   bool   `ini:"BoolField"`
		StringField string `ini:"StringField"`
	}

	enc := ini.NewEncoder(&s.buf)
	value := &testType{
		IntField:    42,
		BoolField:   true,
		StringField: "hello world",
	}
	c.Assert(enc.Encode(value), IsNil)
	c.Check(s.buf.String(), Equals, ""+
		"IntField=42\n"+
		"BoolField=true\n"+
		"StringField=hello world\n")
}

func (s *encoderSuite) TestEncodeRespectsOmitEmptyTag(c *C) {
	type testType struct {
		IntField    int    `ini:"IntField,omitempty"`
		BoolField   bool   `ini:"BoolField,omitempty"`
		StringField string `ini:"StringField,omitempty"`
	}

	enc := ini.NewEncoder(&s.buf)
	c.Assert(enc.Encode(&testType{}), IsNil)
	c.Check(s.buf.String(), Equals, "")
}

func (s *encoderSuite) TestEncodeRespectsIniSections(c *C) {
	type testType struct {
		IntField    int    `ini:"IntField,[Integers]"`
		BoolField   bool   `ini:"BoolField,[Booleans]"`
		StringField string `ini:"StringField,[Strings]"`
	}

	enc := ini.NewEncoder(&s.buf)
	value := &testType{
		IntField:    42,
		BoolField:   true,
		StringField: "hello world",
	}
	c.Assert(enc.Encode(value), IsNil)
	c.Check(s.buf.String(), Equals, ""+
		"[Integers]\n"+
		"IntField=42\n"+
		"[Booleans]\n"+
		"BoolField=true\n"+
		"[Strings]\n"+
		"StringField=hello world\n")
}

func (s *encoderSuite) TestEncodeKeepsTrackOfSections(c *C) {
	type testType struct {
		IntField1 int  `ini:"IntField1,[Integers]"`
		IntField2 int  `ini:"IntField2"`
		IntField3 int  `ini:"IntField3,[Integers]"`
		BoolField bool `ini:"BoolField,[Booleans]"`
	}

	enc := ini.NewEncoder(&s.buf)
	value := &testType{
		IntField1: 1,
		IntField2: 2,
		IntField3: 3,
		BoolField: true,
	}
	c.Assert(enc.Encode(value), IsNil)
	c.Check(s.buf.String(), Equals, ""+
		"[Integers]\n"+
		"IntField1=1\n"+
		"IntField2=2\n"+
		"IntField3=3\n"+
		"[Booleans]\n"+
		"BoolField=true\n")
}

func (s *encoderSuite) TestEncodeSectinsAndOmittedEmptyFields(c *C) {
	type testType struct {
		IntField1 int `ini:"IntField1,[Integers]"`
		// Field is omitted when value is zero, but the section name sticks.
		IntField2 int `ini:"IntField2,[Surprise],omitempty"`
		IntField3 int `ini:"IntField3"`
	}

	enc := ini.NewEncoder(&s.buf)
	value := &testType{
		IntField1: 1,
		IntField2: 0,
		IntField3: 3,
	}
	c.Assert(enc.Encode(value), IsNil)
	c.Check(s.buf.String(), Equals, ""+
		"[Integers]\n"+
		"IntField1=1\n"+
		"[Surprise]\n"+
		"IntField3=3\n")
}

func (s *encoderSuite) TestEncodeBool(c *C) {
	type testType struct {
		Field bool `ini:"F"`
	}

	enc := ini.NewEncoder(&s.buf)
	value := &testType{Field: false}
	c.Assert(enc.Encode(value), IsNil)
	c.Check(s.buf.String(), Equals, "F=false\n")
	s.buf.Reset()

	value = &testType{Field: true}
	c.Assert(enc.Encode(value), IsNil)
	c.Check(s.buf.String(), Equals, "F=true\n")
	s.buf.Reset()
}

func (s *encoderSuite) TestEncodeInt(c *C) {
	type testCase struct {
		value interface{}
		min   int64
		max   int64
	}

	enc := ini.NewEncoder(&s.buf)

	for _, tc := range []testCase{
		{
			value: &struct {
				Field int `ini:"F"`
			}{},
			min: MinInt,
			max: MaxInt,
		},
		{
			value: &struct {
				Field int8 `ini:"F"`
			}{},
			min: math.MinInt8,
			max: math.MaxInt8,
		},
		{
			value: &struct {
				Field int16 `ini:"F"`
			}{},
			min: math.MinInt16,
			max: math.MaxInt16,
		},
		{
			value: &struct {
				Field int32 `ini:"F"`
			}{},
			min: math.MinInt32,
			max: math.MaxInt32,
		},
		{
			value: &struct {
				Field int64 `ini:"F"`
			}{},
			min: math.MinInt64,
			max: math.MaxInt64,
		},
	} {
		comment := Commentf("test case: %#v", tc)

		reflect.ValueOf(tc.value).Elem().FieldByName("Field").SetInt(tc.min)
		c.Assert(enc.Encode(tc.value), IsNil)
		c.Check(s.buf.String(), Equals, "F="+strconv.FormatInt(tc.min, 10)+"\n", comment)
		s.buf.Reset()

		reflect.ValueOf(tc.value).Elem().FieldByName("Field").SetInt(tc.max)
		c.Assert(enc.Encode(tc.value), IsNil)
		c.Check(s.buf.String(), Equals, "F="+strconv.FormatInt(tc.max, 10)+"\n", comment)
		s.buf.Reset()
	}
}

func (s *encoderSuite) TestEncodeUint(c *C) {
	enc := ini.NewEncoder(&s.buf)

	type testCase struct {
		value interface{}
		max   uint64
	}

	for _, tc := range []testCase{
		{
			value: &struct {
				Field uint `ini:"F"`
			}{},
			max: MaxUint,
		},
		{
			value: &struct {
				Field uint8 `ini:"F"`
			}{},
			max: math.MaxUint8,
		},
		{
			value: &struct {
				Field uint16 `ini:"F"`
			}{},
			max: math.MaxUint16,
		},
		{
			value: &struct {
				Field uint32 `ini:"F"`
			}{},
			max: math.MaxUint32,
		},
		{
			value: &struct {
				Field uint64 `ini:"F"`
			}{},
			max: math.MaxUint64,
		},
	} {
		comment := Commentf("test case: %#v", tc)

		reflect.ValueOf(tc.value).Elem().FieldByName("Field").SetUint(tc.max)
		c.Assert(enc.Encode(tc.value), IsNil)
		c.Check(s.buf.String(), Equals, "F="+strconv.FormatUint(tc.max, 10)+"\n", comment)
		s.buf.Reset()

		reflect.ValueOf(tc.value).Elem().FieldByName("Field").SetUint(0)
		c.Assert(enc.Encode(tc.value), IsNil)
		c.Check(s.buf.String(), Equals, "F=0\n", comment)
		s.buf.Reset()
	}
}

func (s *encoderSuite) TestEncodeFloat(c *C) {
	enc := ini.NewEncoder(&s.buf)

	type testCase struct {
		value   interface{}
		min     float64
		max     float64
		bitSize int
	}

	for _, tc := range []testCase{
		{
			value: &struct {
				Field float32 `ini:"F"`
			}{},
			min:     math.SmallestNonzeroFloat32,
			max:     math.MaxFloat32,
			bitSize: 32,
		},
		{
			value: &struct {
				Field float64 `ini:"F"`
			}{},
			min:     math.SmallestNonzeroFloat64,
			max:     math.MaxFloat64,
			bitSize: 64,
		},
	} {
		comment := Commentf("test case: %#v", tc)

		reflect.ValueOf(tc.value).Elem().FieldByName("Field").SetFloat(tc.min)
		c.Assert(enc.Encode(tc.value), IsNil)
		c.Check(s.buf.String(), Equals, "F="+strconv.FormatFloat(tc.min, 'f', -1, tc.bitSize)+"\n", comment)
		s.buf.Reset()

		reflect.ValueOf(tc.value).Elem().FieldByName("Field").SetFloat(tc.max)
		c.Assert(enc.Encode(tc.value), IsNil)
		c.Check(s.buf.String(), Equals, "F="+strconv.FormatFloat(tc.max, 'f', -1, tc.bitSize)+"\n", comment)
		s.buf.Reset()
	}
}

func (s *encoderSuite) TestEncodeString(c *C) {
	enc := ini.NewEncoder(&s.buf)

	type testCase struct {
		value   interface{}
		textIn  string
		textOut string
	}

	for _, tc := range []testCase{
		{
			value: &struct {
				Field string `ini:"F"`
			}{},
			textIn:  "hello",
			textOut: "hello",
		},
		{
			value: &struct {
				Field string `ini:"F"`
			}{},
			textIn: "hello\nnewline\nfriendly",
			// The two spaces on each following line are to align with F= on the
			// first line, not pictured here.
			textOut: "hello\\\n  newline\\\n  friendly",
		},
		{
			value: &struct {
				Field string `ini:"F,quote"`
			}{},
			textIn:  "hello",
			textOut: `"hello"`,
		},
		{
			value: &struct {
				Field string `ini:"F,quote"`
			}{},
			textIn:  "hello\nnewline\nfriendly",
			textOut: `"hello\nnewline\nfriendly"`,
		},
	} {
		comment := Commentf("test case: %#v", tc)

		reflect.ValueOf(tc.value).Elem().FieldByName("Field").SetString(tc.textIn)
		c.Assert(enc.Encode(tc.value), IsNil, comment)
		c.Check(s.buf.String(), Equals, "F="+tc.textOut+"\n", comment)
		s.buf.Reset()
	}
}

func (s *encoderSuite) TestEncodeTextMarshaler(c *C) {
	enc := ini.NewEncoder(&s.buf)

	type testType struct {
		Field testEnum `ini:"F"`
	}

	value := &testType{Field: testFoo}
	c.Assert(enc.Encode(value), IsNil)
	c.Check(s.buf.String(), Equals, "F=foo\n")

	value = &testType{Field: testEnum(42)}
	c.Assert(enc.Encode(value), ErrorMatches, `cannot encode testType.Field: cannot marshal testEnum: 42`)
}
