// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package ini

import (
	"encoding"
	"fmt"
	"reflect"
	"strings"
)

type iniFieldInfo struct {
	Section   string
	Name      string
	OmitEmpty bool
	Quote     bool

	FieldIndex []int
}

var (
	textMarshalerType   = reflect.TypeOf((*encoding.TextMarshaler)(nil)).Elem()
	textUnmarshalerType = reflect.TypeOf((*encoding.TextUnmarshaler)(nil)).Elem()
)

func iniSliceOfType(structType reflect.Type) ([]iniFieldInfo, error) {
	result := make([]iniFieldInfo, 0, structType.NumField())
	section := "" // keeps track of current section name

	for i := 0; i < structType.NumField(); i++ {
		structField := structType.Field(i)
		// TODO: switch to structField.IsExported available since go 1.17
		if structField.PkgPath != "" {
			continue
		}

		switch structField.Type.Kind() {
		case reflect.Bool:
		case reflect.Int, reflect.Int8, reflect.Int16, reflect.Int32, reflect.Int64:
		case reflect.Uint, reflect.Uint8, reflect.Uint16, reflect.Uint32, reflect.Uint64:
		case reflect.Float32, reflect.Float64:
		case reflect.String:
		default:
			// Check if the field can be an encoding.TextMarshaler and encoding.TextUnmarshaler.
			if !structField.Type.Implements(textMarshalerType) {
				return nil, fmt.Errorf("ini error: field %s non-basic type does not implement encoding.TextMarshaler", structField.Name)
			}
			// Since TextUnmarshaler requires practical implementations to be called by pointer,
			// perform the check on the pointer to the type of the field.
			if !reflect.PtrTo(structField.Type).Implements(textUnmarshalerType) {
				return nil, fmt.Errorf("ini error: field %s non-basic type does not implement encoding.TextUnmarshaler", structField.Name)
			}
		}

		tagValue, ok := structField.Tag.Lookup("ini")
		if !ok {
			continue
		}

		var name, newSection string

		var omitempty, quote bool

		iniTags := strings.Split(tagValue, ",")
		for i, tag := range iniTags {
			if i == 0 {
				name = tag
				continue
			}

			switch tag {
			case "omitempty":
				omitempty = true
			case "quote":
				quote = true
			default:
				if strings.HasPrefix(tag, "[") && strings.HasSuffix(tag, "]") {
					newSection = tag[1 : len(tag)-1]
				} else {
					return nil, fmt.Errorf("ini error: field %s contains unsupported struct-tag: %q", structField.Name, tag)
				}
			}
		}

		if name == "" {
			return nil, fmt.Errorf("ini error: field %s lacks struct-tag with field name", structField.Name)
		}

		if newSection != "" && section != newSection {
			section = newSection
		}

		result = append(result, iniFieldInfo{
			Name:       name,
			Section:    section,
			FieldIndex: structField.Index,
			OmitEmpty:  omitempty,
			Quote:      quote,
		})
	}

	return result, nil
}

func iniMapOfType(structType reflect.Type, caseSensitive bool) (slice []iniFieldInfo, indexMap map[string]map[string]int, err error) {
	slice, err = iniSliceOfType(structType)
	indexMap = make(map[string]map[string]int, len(slice))

	for idx, info := range slice {
		section := info.Section
		name := info.Name

		if !caseSensitive {
			section = strings.ToLower(section)
			name = strings.ToLower(name)
		}

		if indexMap[section] == nil {
			indexMap[section] = make(map[string]int)
		}

		indexMap[section][name] = idx
	}

	return slice, indexMap, err
}
