// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package ini_test

import (
	"reflect"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/encoding/ini"
	. "gopkg.in/check.v1"
)

type infoSuite struct{}

var _ = Suite(&infoSuite{})

func (s *infoSuite) TestIniSliceOfTypeUnsupportedStructTags(c *C) {
	var value struct {
		Field int `ini:"field,magic"`
	}

	_, err := ini.IniSliceOfType(reflect.TypeOf(value))
	c.Assert(err, ErrorMatches, `ini error: field Field contains unsupported struct-tag: "magic"`)
}

func (s *infoSuite) TestIniSliceOfTypeMissingFieldName(c *C) {
	var value struct {
		Field int `ini:""`
	}

	_, err := ini.IniSliceOfType(reflect.TypeOf(value))
	c.Assert(err, ErrorMatches, `ini error: field Field lacks struct-tag with field name`)
}

func (s *infoSuite) TestIniSliceOfTypeTagParsing(c *C) {
	var value struct {
		internal  int    `ini:"internal,[section]"`
		Exported  int    `ini:"Exported"`
		OmitEmpty int    `ini:"OmitEmpty,omitempty"`
		Quoted    string `ini:"Quoted,quote"`
	}

	slice, err := ini.IniSliceOfType(reflect.TypeOf(value))
	c.Assert(err, IsNil)
	c.Assert(slice, HasLen, 3)
	c.Check(slice[0].Name, Equals, "Exported")
	c.Check(slice[0].Section, Equals, "")
	c.Check(slice[0].Quote, Equals, false)
	c.Check(slice[1].Name, Equals, "OmitEmpty")
	c.Check(slice[1].Section, Equals, "")
	c.Check(slice[1].Quote, Equals, false)
	c.Check(slice[2].Name, Equals, "Quoted")
	c.Check(slice[2].Section, Equals, "")
	c.Check(slice[2].Quote, Equals, true)
}

func (s *infoSuite) TestIniSliceOfTypeSectionTracking(c *C) {
	var value struct {
		internal  int    `ini:"internal,[ignored]"`
		Exported  int    `ini:"Exported,[section]"`
		OmitEmpty int    `ini:"OmitEmpty,omitempty"`
		Quoted    string `ini:"Quoted,quote,[another]"`
	}

	slice, err := ini.IniSliceOfType(reflect.TypeOf(value))
	c.Assert(err, IsNil)
	c.Assert(slice, HasLen, 3)
	c.Check(slice[0].Section, Equals, "section")
	c.Check(slice[1].Section, Equals, "section")
	c.Check(slice[2].Section, Equals, "another")
}

func (s *infoSuite) TestIniMapOfTypeCaseSensitivity(c *C) {
	var value struct {
		Field int `ini:"FIELD,[SECTION]"`
	}

	const caseSensitive = false

	slice, idxMap, err := ini.IniMapOfType(reflect.TypeOf(value), caseSensitive)
	c.Assert(err, IsNil)
	c.Assert(slice, HasLen, 1)
	c.Check(slice[0].Section, Equals, "SECTION")
	c.Check(slice[0].Name, Equals, "FIELD")
	c.Assert(idxMap, HasLen, 1)
	c.Assert(idxMap["section"], NotNil)
	c.Assert(idxMap["section"], HasLen, 1)
	c.Assert(idxMap["section"]["field"], Equals, 0)

	slice, idxMap, err = ini.IniMapOfType(reflect.TypeOf(value), !caseSensitive)
	c.Assert(err, IsNil)
	c.Assert(slice, HasLen, 1)
	c.Check(slice[0].Section, Equals, "SECTION")
	c.Check(slice[0].Name, Equals, "FIELD")
	c.Assert(idxMap, HasLen, 1)
	c.Assert(idxMap["SECTION"], NotNil)
	c.Assert(idxMap["SECTION"], HasLen, 1)
	c.Assert(idxMap["SECTION"]["FIELD"], Equals, 0)
}

type neitherMarshalNorUnmarshal struct{}

func (s *infoSuite) TestIniSliceOfTypeCustomTypeNeitherMarshalNorUnmarshal(c *C) {
	var value struct {
		NeitherMarshalNorUnmarshal neitherMarshalNorUnmarshal `ini:"NeitherMarshalNorUnmarshal"`
	}

	_, err := ini.IniSliceOfType(reflect.TypeOf(value))
	// XXX: error message doesn't mention TextUnmarshaler even though it is also missing.
	c.Assert(err, ErrorMatches, `ini error: field NeitherMarshalNorUnmarshal non-basic type does not implement encoding.TextMarshaler`)
}

type onlyMarshal struct{}

func (onlyMarshal) MarshalText() ([]byte, error) {
	return []byte(""), nil
}

func (s *infoSuite) TestIniSliceOfTypeCustomTypeOnlyMarshal(c *C) {
	var value struct {
		OnlyMarshal onlyMarshal `ini:"OnlyMarshal"`
	}

	_, err := ini.IniSliceOfType(reflect.TypeOf(value))
	c.Assert(err, ErrorMatches, `ini error: field OnlyMarshal non-basic type does not implement encoding.TextUnmarshaler`)
}

type onlyUnmarshal struct{}

func (o *onlyUnmarshal) UnmarshalText(data []byte) error {
	return nil
}

func (s *infoSuite) TestIniSliceOfTypeCustomTypeOnlyUnmarshal(c *C) {
	var value struct {
		OnlyUnmarshal onlyUnmarshal `ini:"OnlyUnmarshal"`
	}

	_, err := ini.IniSliceOfType(reflect.TypeOf(value))
	c.Assert(err, ErrorMatches, `ini error: field OnlyUnmarshal non-basic type does not implement encoding.TextMarshaler`)
}

type bothMarshalAndUnmarshal struct{}

func (bothMarshalAndUnmarshal) MarshalText() ([]byte, error) {
	return []byte(""), nil
}

func (o *bothMarshalAndUnmarshal) UnmarshalText(data []byte) error {
	return nil
}

func (s *infoSuite) TestIniSliceOfTypeCustomTypeBothMarshalAndUnmarshal(c *C) {
	var value struct {
		BothMarshalAndUnmarshal bothMarshalAndUnmarshal `ini:"BothMarshalAndUnmarshal"`
	}

	_, err := ini.IniSliceOfType(reflect.TypeOf(value))
	c.Assert(err, IsNil)
}
