// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

// Package ota implements system over-the-air update for Linux gateways.
package ota

import (
	"bytes"
	"fmt"
	"os"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/encoding/ini"
)

// TODO: add marshalable server response types here.

// BootLoaderType designates the type of the boot loader.
type BootLoaderType int

const (
	// InertBootLoader indicates that the OTA system will not interact with the boot loader.
	InertBootLoader BootLoaderType = iota
	// TestBoot can only be used in unit tests, it cannot be marshled.
	// To use it, call service.SetTestBootProtocol and pass &ota.Config{BootLoaderType: TestBoot} to service.New.
	TestBoot
	// PiBoot indicates that the OTA system will configure the Raspberry Pi boot loader.
	PiBoot
)

const (
	inert    = "inert"
	testBoot = "test"
	piBoot   = "pi-boot"
)

// String returns the type of the boot loader as text.
func (t BootLoaderType) String() string {
	switch t {
	case InertBootLoader:
		return inert
	case TestBoot:
		return testBoot
	case PiBoot:
		return piBoot
	default:
		return fmt.Sprintf("BootLoaderType(%d)", int(t))
	}
}

// MarshalText implements TextMarshaler.
func (t BootLoaderType) MarshalText() ([]byte, error) {
	switch t {
	case InertBootLoader:
		return []byte(inert), nil
	case PiBoot:
		return []byte(piBoot), nil
	case TestBoot:
		return nil, fmt.Errorf("test boot loader type cannot be marshaled")
	default:
		return nil, fmt.Errorf("invalid boot loader type: %d", int(t))
	}
}

// UnmarshalText implements TextUnmarshaler.
func (t *BootLoaderType) UnmarshalText(data []byte) error {
	if bytes.Equal(data, []byte(inert)) {
		*t = InertBootLoader
		return nil
	} else if bytes.Equal(data, []byte(piBoot)) {
		*t = PiBoot
		return nil
	} else {
		return fmt.Errorf("cannot unmarshal boot loader type from %q", string(data))
	}
}

// Config describes the configuration of the OTA service.
type Config struct {
	// BootLoaderType indicates the type of the boot loader used by the system,
	// that the OTA servie is to interact with.
	BootLoaderType BootLoaderType `ini:"BootLoaderType,[OTA],omitempty"`

	// BootPartitionMountDir indicates where the boot is mounted. This may be
	// "/boot" or a sub-directory, such as "/boot/firmware", depending on the
	// platform.
	BootPartitionMountDir string `ini:"BootPartitionMountDir,omitempty"`
}

// DefaultConfigFiles returns sysotad configuration files loaded on startup.
//
// There are three configuration files, each with distinct purpose:
//
// The first loaded file is /usr/lib/sysota/sysotad.conf. This file should come
// with the system image and should contain defaults suitable for the device.
//
// The second loaded file is /etc/sysota/sysotad.conf. This file should not be
// present on a device unless it was created during first boot and can be
// re-created automatically.
//
// The third file is /run/sysota/sysotad.conf. This file can be generated by
// another program that runs before sysotad.service in the start up chain. It
// can be used generate the configuration entirely dynamically, and since it has
// highest priority, decide on the final form of the configuration.
func DefaultConfigFiles() []string {
	return []string{
		"/usr/lib/sysota/sysotad.conf",
		"/etc/sysota/sysotad.conf",
		// FIXME: this file is removed when sysotad.service is stopped because
		// defines RuntimeDirectory=sysota and according to systemd.exec(5)
		// that directory is removed when the service terminates.
		"/run/sysota/sysotad.conf",
	}
}

// LoadConfig loads and combines sysota configuration files.
//
// As files are loaded, settings are stored into the configuration object are
// overwritten by newer definitions. Care must be taken to define a consistent
// set.
//
// If there is an error, it will be of type *ConfigError.
func LoadConfig(confFiles []string) (cfg *Config, err error) {
	cfg = &Config{}

	var errs []error

	for _, name := range confFiles {
		f, err := os.Open(name)
		if err != nil {
			if os.IsNotExist(err) {
				continue
			}

			errs = append(errs, err)
		}

		defer func() {
			e := f.Close()
			if err == nil {
				err = e
			}
		}()

		dec := ini.NewDecoder(f)
		if err := dec.Decode(cfg); err != nil {
			errs = append(errs, &ConfigError{FileName: name, Err: err})
		}
	}
	// TODO: log remaining errors
	if len(errs) > 0 {
		return nil, fmt.Errorf("cannot load configuration: %w", errs[0])
	}

	return cfg, nil
}

// ConfigError reports errors with configuration files.
type ConfigError struct {
	FileName string
	Err      error
}

// Unwrap returns the error wrapped inside the config error.
func (cfgErr *ConfigError) Unwrap() error {
	return cfgErr.Err
}

// Error implements the error interface.
func (cfgErr *ConfigError) Error() string {
	return fmt.Sprintf("%s: %s", cfgErr.FileName, cfgErr.Err)
}

const (
	// ServiceBusName is the D-Bus bus name used by the SystemOTA service.
	ServiceBusName = "dev.ostc.sysota1"
	// ServiceObjectPath is the D-Bus object path used by the SystemOTA service.
	ServiceObjectPath = "/dev/ostc/sysota1/Service"
)
