// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package ota_test

import (
	"errors"
	"io/ioutil"
	"os"
	"path/filepath"
	"testing"

	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/encoding/ini"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/ota"
)

func Test(t *testing.T) { TestingT(t) }

type confSuite struct{}

var _ = Suite(&confSuite{})

func (s *confSuite) TestEncodeConfPiBootLoader(c *C) {
	conf := ota.Config{
		BootLoaderType:        ota.PiBoot,
		BootPartitionMountDir: "/boot/firmware",
	}
	data, err := ini.Marshal(conf)
	c.Assert(err, IsNil)
	c.Check(string(data), Equals, ""+
		"[OTA]\n"+
		"BootLoaderType=pi-boot\n"+
		"BootPartitionMountDir=/boot/firmware\n")
}

func (s *confSuite) TestLoadConfigMissingFiles(c *C) {
	d := c.MkDir()

	conf, err := ota.LoadConfig([]string{filepath.Join(d, "one.conf"), filepath.Join(d, "two.conf")})
	c.Assert(err, IsNil)
	c.Check(conf, DeepEquals, &ota.Config{})
}

func (s *confSuite) TestDefaultConfigFiles(c *C) {
	c.Check(ota.DefaultConfigFiles(), DeepEquals, []string{
		"/usr/lib/sysota/sysotad.conf",
		"/etc/sysota/sysotad.conf",
		"/run/sysota/sysotad.conf",
	})
}

const otaConfForPiBoot = `
[OTA]
BootLoaderType=pi-boot
BootPartitionMountDir=/boot/firmware
`

func (s *confSuite) TestLoadConfig(c *C) {
	d := c.MkDir()
	name := filepath.Join(d, "sysotad.conf")

	c.Assert(ioutil.WriteFile(name, []byte(otaConfForPiBoot), 0600), IsNil)

	conf, err := ota.LoadConfig([]string{name})
	c.Assert(err, IsNil)
	c.Check(conf, DeepEquals, &ota.Config{
		BootLoaderType:        ota.PiBoot,
		BootPartitionMountDir: "/boot/firmware",
	})
}

func (s *confSuite) TestLoadConfigConfigPriority(c *C) {
	d := c.MkDir()
	lower := filepath.Join(d, "lower.conf")
	higher := filepath.Join(d, "higher.conf")

	c.Assert(ioutil.WriteFile(lower, []byte("[OTA]\nBootLoaderType=pi-boot\n"), 0600), IsNil)
	c.Assert(ioutil.WriteFile(higher, []byte("[OTA]\nBootLoaderType=inert\n"), 0600), IsNil)

	conf, err := ota.LoadConfig([]string{lower, higher})
	c.Assert(err, IsNil)
	c.Check(conf, DeepEquals, &ota.Config{BootLoaderType: ota.InertBootLoader})
}

func (s *confSuite) TestLoadConfigCorruptedFile(c *C) {
	d := c.MkDir()
	broken := filepath.Join(d, "broken.conf")
	c.Assert(ioutil.WriteFile(broken, []byte("[BootL...!@#!$%"), 0600), IsNil)

	_, err := ota.LoadConfig([]string{broken})
	c.Assert(err, ErrorMatches, `cannot load configuration: .*/broken.conf: cannot decode line 1: unexpected input: .*`)
	// If necessary, wrapped errors can be extracted and analyzed
	var cfgErr *ota.ConfigError

	c.Assert(errors.As(err, &cfgErr), Equals, true)
	c.Assert(cfgErr.FileName, Equals, broken)

	var decErr *ini.DecoderError

	c.Assert(errors.As(err, &decErr), Equals, true)
	c.Assert(decErr, ErrorMatches, `cannot decode line 1: unexpected input: .*`)
}

func (s *confSuite) TestLoadConfigInaccessibleFile(c *C) {
	if os.Geteuid() == 0 {
		c.Skip("cannot run as root, test depends on chmod")
	}

	d := c.MkDir()
	inaccessible := filepath.Join(d, "inaccessible.conf")
	c.Assert(ioutil.WriteFile(inaccessible, []byte("[BootLoader]\nType=foo\n"), 0600), IsNil)
	c.Assert(os.Chmod(inaccessible, 0o200), IsNil)

	_, err := ota.LoadConfig([]string{inaccessible})
	c.Assert(err, ErrorMatches, `cannot load configuration: open .*: permission denied`)
	c.Assert(errors.Unwrap(err), ErrorMatches, `open .*: permission denied`)
}

type bootLoaderTypeSuite struct{}

var _ = Suite(&bootLoaderTypeSuite{})

func (*bootLoaderTypeSuite) TestString(c *C) {
	c.Check(ota.InertBootLoader.String(), Equals, "inert")
	c.Check(ota.TestBoot.String(), Equals, "test")
	c.Check(ota.PiBoot.String(), Equals, "pi-boot")
	c.Check(ota.BootLoaderType(42).String(), Equals, "BootLoaderType(42)")
}

func (*bootLoaderTypeSuite) TestMarshalText(c *C) {
	data, err := ota.InertBootLoader.MarshalText()
	c.Assert(err, IsNil)
	c.Check(data, DeepEquals, []byte("inert"))

	_, err = ota.TestBoot.MarshalText()
	c.Assert(err, ErrorMatches, `test boot loader type cannot be marshaled`)

	data, err = ota.PiBoot.MarshalText()
	c.Assert(err, IsNil)
	c.Check(data, DeepEquals, []byte("pi-boot"))

	_, err = ota.BootLoaderType(42).MarshalText()
	c.Assert(err, ErrorMatches, `invalid boot loader type: 42`)
}

func (*bootLoaderTypeSuite) TestUnmarshalText(c *C) {
	var typ ota.BootLoaderType

	err := typ.UnmarshalText([]byte(`inert`))
	c.Assert(err, IsNil)
	c.Check(typ, Equals, ota.InertBootLoader)

	err = typ.UnmarshalText([]byte(`test`))
	c.Assert(err, ErrorMatches, `cannot unmarshal boot loader type from "test"`)

	err = typ.UnmarshalText([]byte(`pi-boot`))
	c.Assert(err, IsNil)
	c.Check(typ, Equals, ota.PiBoot)

	err = typ.UnmarshalText([]byte(`potato`))
	c.Assert(err, ErrorMatches, `cannot unmarshal boot loader type from "potato"`)
}
