// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package raucinterface_test

import (
	"encoding/xml"
	"errors"
	"testing"

	"github.com/godbus/dbus/v5"
	"github.com/godbus/dbus/v5/introspect"
	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/boot"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/dbusutil"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/dbusutil/dbustest"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/ota"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/ota/raucinterface"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/rauc/installhandler"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/rauc/rauctest"
)

func Test(t *testing.T) { TestingT(t) }

var errBoom = errors.New("boom")

type clientSuite struct {
	dbustest.Suite
	conn               *dbus.Conn
	serviceHost        *dbusutil.ServiceHost
	serviceBackend     rauctest.BootHandler
	installParticipant rauctest.InstallParticipant
}

var _ = Suite(&clientSuite{})

func (s *clientSuite) SetUpTest(c *C) {
	s.serviceBackend.ResetCallbacks()
	s.installParticipant.ResetCallbacks()
}

func (s *clientSuite) EnsureConn(c *C) *dbus.Conn {
	if s.conn == nil {
		var err error
		s.conn, err = dbusutil.SystemBus()
		c.Assert(err, IsNil)
	}

	return s.conn
}

func (s *clientSuite) EnsureService(c *C) {
	if s.serviceHost == nil {
		conn := s.EnsureConn(c)

		s.serviceHost = dbusutil.NewServiceHost(conn)
		s.serviceHost.AddHostedService(raucinterface.NewService(&s.serviceBackend, &s.installParticipant))

		err := s.serviceHost.Export()
		c.Assert(err, IsNil)

		reply, err := conn.RequestName(ota.ServiceBusName, dbus.NameFlagDoNotQueue)
		c.Assert(err, IsNil)
		c.Assert(reply, Equals, dbus.RequestNameReplyPrimaryOwner)
	}
}

func (s *clientSuite) TearDownTest(c *C) {
	if s.serviceHost != nil {
		err := s.serviceHost.Unexport()
		c.Check(err, IsNil)

		s.serviceHost = nil
	}

	if s.conn != nil {
		err := s.conn.Close()
		c.Check(err, IsNil)

		s.conn = nil
	}
}

func (s *clientSuite) TestServiceMissing(c *C) {
	client := raucinterface.NewClient(s.EnsureConn(c))

	_, err := client.PrimarySlot()
	c.Assert(err, ErrorMatches, `The name dev.ostc.sysota1 was not provided by any .service files`)
}

func (s *clientSuite) TestPrimarySlot(c *C) {
	s.EnsureService(c)
	client := raucinterface.NewClient(s.EnsureConn(c))

	called := false
	s.serviceBackend.PrimarySlotFn = func() (boot.Slot, error) {
		called = true
		return boot.SlotA, nil
	}

	// Both client and service are happy.
	slot, err := client.PrimarySlot()
	c.Assert(err, IsNil)
	c.Check(slot, Equals, boot.SlotA)
	c.Check(called, Equals, true)

	// Service returns invalid slot, client is unhappy.
	called = false
	s.serviceBackend.PrimarySlotFn = func() (boot.Slot, error) {
		called = true

		return boot.InvalidSlot, nil
	}

	_, err = client.PrimarySlot()
	c.Assert(err, ErrorMatches, `invalid boot slot`)
	c.Check(called, Equals, true)

	// Service fails, client reports the error.
	called = false
	s.serviceBackend.PrimarySlotFn = func() (boot.Slot, error) {
		called = true

		return boot.InvalidSlot, errBoom
	}

	_, err = client.PrimarySlot()
	c.Assert(err, ErrorMatches, `boom`)
	c.Check(called, Equals, true)
}

func (s *clientSuite) TestSetPrimarySlot(c *C) {
	s.EnsureService(c)
	client := raucinterface.NewClient(s.EnsureConn(c))

	// Both client and service are happy.
	called := false
	s.serviceBackend.SetPrimarySlotFn = func(slot boot.Slot) error {
		c.Check(slot, Equals, boot.SlotA)

		called = true

		return nil
	}

	err := client.SetPrimarySlot(boot.SlotA)
	c.Assert(err, IsNil)
	c.Check(called, Equals, true)

	// Service fails, client reports the error.
	called = false
	s.serviceBackend.SetPrimarySlotFn = func(slot boot.Slot) error {
		c.Assert(slot, Equals, boot.SlotA)

		called = true

		return errBoom
	}

	err = client.SetPrimarySlot(boot.SlotA)
	c.Assert(err, ErrorMatches, `boom`)
	c.Check(called, Equals, true)

	// Client tries to set invalid slot, service doesn't see the request.
	s.serviceBackend.SetPrimarySlotFn = func(slot boot.Slot) error {
		panic("service should have not been called")
	}

	err = client.SetPrimarySlot(boot.InvalidSlot)
	c.Assert(err, ErrorMatches, `invalid boot slot`)
}

func (s *clientSuite) TestSlotState(c *C) {
	s.EnsureService(c)
	client := raucinterface.NewClient(s.EnsureConn(c))

	// Both client and service are happy.
	called := false
	s.serviceBackend.SlotStateFn = func(slot boot.Slot) (boot.SlotState, error) {
		c.Check(slot, Equals, boot.SlotA)

		called = true

		return boot.GoodSlot, nil
	}

	state, err := client.SlotState(boot.SlotA)
	c.Assert(err, IsNil)
	c.Check(state, Equals, boot.GoodSlot)
	c.Check(called, Equals, true)

	// Service fails, client reports the error.
	called = false
	s.serviceBackend.SlotStateFn = func(slot boot.Slot) (boot.SlotState, error) {
		c.Check(slot, Equals, boot.SlotB)

		called = true

		return boot.InvalidSlotState, errBoom
	}

	_, err = client.SlotState(boot.SlotB)
	c.Assert(err, ErrorMatches, `boom`)

	// Client asks about invalid slot, service does not see the request.
	s.serviceBackend.SlotStateFn = func(slot boot.Slot) (boot.SlotState, error) {
		panic("service should have not been called")
	}

	state, err = client.SlotState(boot.InvalidSlot)
	c.Assert(err, ErrorMatches, `invalid boot slot`)
	c.Check(state, Equals, boot.InvalidSlotState)

	// Service returns invalid state, client is unhappy.
	called = false
	s.serviceBackend.SlotStateFn = func(slot boot.Slot) (boot.SlotState, error) {
		c.Check(slot, Equals, boot.SlotB)

		called = true

		return boot.InvalidSlotState, nil
	}

	_, err = client.SlotState(boot.SlotB)
	c.Assert(err, ErrorMatches, `invalid slot state`)
	c.Check(called, Equals, true)
}

func (s *clientSuite) TestSetSlotState(c *C) {
	s.EnsureService(c)
	client := raucinterface.NewClient(s.EnsureConn(c))

	// Both client and service are happy.
	called := false
	s.serviceBackend.SetSlotStateFn = func(slot boot.Slot, state boot.SlotState) error {
		c.Check(slot, Equals, boot.SlotA)
		c.Check(state, Equals, boot.GoodSlot)

		called = true

		return nil
	}

	err := client.SetSlotState(boot.SlotA, boot.GoodSlot)
	c.Assert(err, IsNil)
	c.Check(called, Equals, true)

	// Service fails, client reports the error.
	called = false
	s.serviceBackend.SetSlotStateFn = func(slot boot.Slot, state boot.SlotState) error {
		c.Check(slot, Equals, boot.SlotA)
		c.Check(state, Equals, boot.GoodSlot)

		called = true

		return errBoom
	}

	err = client.SetSlotState(boot.SlotA, boot.GoodSlot)
	c.Assert(err, ErrorMatches, `boom`)
	c.Check(called, Equals, true)

	// Client sets the state of an invalid slot, service does not see the request.
	s.serviceBackend.SetSlotStateFn = func(slot boot.Slot, state boot.SlotState) error {
		panic("service should have not been called")
	}

	err = client.SetSlotState(boot.InvalidSlot, boot.GoodSlot)
	c.Assert(err, ErrorMatches, `invalid boot slot`)

	// Client sets an invalid slot state, service does not see the request.
	s.serviceBackend.SetSlotStateFn = func(slot boot.Slot, state boot.SlotState) error {
		panic("service should have not been called")
	}

	err = client.SetSlotState(boot.SlotA, boot.InvalidSlotState)
	c.Assert(err, ErrorMatches, `invalid slot state`)
}

func (s *clientSuite) TestPreInstallSuccess(c *C) {
	// Both client and service are happy.
	s.EnsureService(c)
	client := raucinterface.NewClient(s.EnsureConn(c))
	ctx := &installhandler.HandlerContext{
		SystemConfig: "foo",
	}
	called := false
	s.installParticipant.PreInstallHandlerFn = func(ctx2 *installhandler.HandlerContext) error {
		c.Check(ctx, DeepEquals, ctx2)

		called = true

		return nil
	}

	err := client.PreInstall(ctx)
	c.Assert(err, IsNil)
	c.Check(called, Equals, true)
}

func (s *clientSuite) TestPreInstallBadInput(c *C) {
	// Client provides faulty data, service rejects it.
	s.EnsureService(c)
	client := raucinterface.NewClient(s.EnsureConn(c))
	err := client.PreInstall(&installhandler.HandlerContext{SlotIDs: []string{"potato"}})
	c.Assert(err, ErrorMatches, `cannot create handler context: variable RAUC_SLOTS refers to undefined slot potato`)
}

func (s *clientSuite) TestPreInstallServiceFails(c *C) {
	// Service fails, client reports the error.
	s.EnsureService(c)
	client := raucinterface.NewClient(s.EnsureConn(c))
	s.installParticipant.PreInstallHandlerFn = func(ctx *installhandler.HandlerContext) error {
		return errBoom
	}
	err := client.PreInstall(&installhandler.HandlerContext{})
	c.Assert(err, ErrorMatches, `boom`)
}

func (s *clientSuite) TestPostInstallSuccess(c *C) {
	// Both client and service are happy.
	s.EnsureService(c)
	client := raucinterface.NewClient(s.EnsureConn(c))
	ctx := &installhandler.HandlerContext{
		SystemConfig: "foo",
	}
	called := false
	s.installParticipant.PostInstallHandlerFn = func(ctx2 *installhandler.HandlerContext) error {
		c.Check(ctx, DeepEquals, ctx2)

		called = true

		return nil
	}

	err := client.PostInstall(ctx)
	c.Assert(err, IsNil)
	c.Check(called, Equals, true)
}

func (s *clientSuite) TestPostInstallBadInput(c *C) {
	// Client provides faulty data, service rejects it.
	s.EnsureService(c)
	client := raucinterface.NewClient(s.EnsureConn(c))
	err := client.PostInstall(&installhandler.HandlerContext{SlotIDs: []string{"potato"}})
	c.Assert(err, ErrorMatches, `cannot create handler context: variable RAUC_SLOTS refers to undefined slot potato`)
}

func (s *clientSuite) TestPostInstallServiceFails(c *C) {
	// Service fails, client reports the error.
	s.EnsureService(c)
	client := raucinterface.NewClient(s.EnsureConn(c))
	s.installParticipant.PostInstallHandlerFn = func(ctx *installhandler.HandlerContext) error {
		return errBoom
	}
	err := client.PostInstall(&installhandler.HandlerContext{})
	c.Assert(err, ErrorMatches, `boom`)
}

func (s *clientSuite) TestIntrospection(c *C) {
	s.EnsureService(c)

	// Introspection data is published and looks correct.
	obj := s.conn.Object(ota.ServiceBusName, ota.ServiceObjectPath)
	node, err := introspect.Call(obj)
	c.Assert(err, IsNil)
	c.Check(node, DeepEquals, &introspect.Node{
		XMLName: xml.Name{Local: "node"},
		Name:    "/dev/ostc/sysota1/Service",
		Interfaces: []introspect.Interface{{
			Name: "dev.ostc.sysota1.RAUC",
			Methods: []introspect.Method{{
				Name: "GetPrimary",
				Args: []introspect.Arg{
					{Name: "slot", Type: "s", Direction: "out"},
				},
			}, {
				Name: "SetPrimary",
				Args: []introspect.Arg{
					{Name: "slot", Type: "s", Direction: "in"},
				},
			}, {
				Name: "GetState",
				Args: []introspect.Arg{
					{Name: "slot", Type: "s", Direction: "in"},
					{Name: "state", Type: "s", Direction: "out"}},
			}, {
				Name: "SetState",
				Args: []introspect.Arg{
					{Name: "slot", Type: "s", Direction: "in"},
					{Name: "state", Type: "s", Direction: "in"},
				},
			}, {
				Name: "PreInstall",
				Args: []introspect.Arg{
					{Name: "context", Type: "as", Direction: "in"},
				},
			}, {
				Name: "PostInstall",
				Args: []introspect.Arg{
					{Name: "context", Type: "as", Direction: "in"},
				},
			}},
		}, {
			Name: "org.freedesktop.DBus.Introspectable",
			Methods: []introspect.Method{{
				Name: "Introspect",
				Args: []introspect.Arg{
					{Name: "out", Type: "s", Direction: "out"}},
			}},
		}},
	})
}
