// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package raucinterface

import (
	"github.com/godbus/dbus/v5"
	"github.com/godbus/dbus/v5/introspect"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/boot"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/dbusutil"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/ota"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/rauc/boothandler"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/rauc/installhandler"
)

const (
	// Methods of the boot handler

	getPrimaryDBusName = "GetPrimary"
	setPrimaryDBusName = "SetPrimary"
	getStateDBusName   = "GetState"
	setStateDBusName   = "SetState"

	// Methods of the install handler

	preInstallDBusName  = "PreInstall"
	postInstallDBusName = "PostInstall"

	// Argument names

	slotArgName    = "slot"
	stateArgName   = "state"
	contextArgName = "context"
)

// Service implements dbusutil.HostedService exposing boothandler.Handler and installhandler.Participant over D-Bus.
type Service struct {
	bootHandler    boothandler.Handler
	installHandler installhandler.Participant
}

// NewService returns a service with the given boot backend.
func NewService(bootHandler boothandler.Handler, installHandler installhandler.Participant) *Service {
	return &Service{bootHandler: bootHandler, installHandler: installHandler}
}

// JoinServiceHost integrates with dbusutil.ServiceHost.
func (svc *Service) JoinServiceHost(reg dbusutil.ServiceRegistration) error {
	methods := map[string]interface{}{
		getPrimaryDBusName: svc.GetPrimary,
		setPrimaryDBusName: svc.SetPrimary,
		getStateDBusName:   svc.GetState,
		setStateDBusName:   svc.SetState,

		preInstallDBusName:  svc.PreInstall,
		postInstallDBusName: svc.PostInstall,
	}

	metadata := introspect.Interface{
		Name: InterfaceName,
		Methods: []introspect.Method{
			{
				Name: getPrimaryDBusName,
				Args: []introspect.Arg{
					{
						Name:      slotArgName,
						Type:      "s",
						Direction: dbusutil.Out,
					},
				},
			},
			{
				Name: setPrimaryDBusName,
				Args: []introspect.Arg{
					{
						Name:      slotArgName,
						Type:      "s",
						Direction: dbusutil.In,
					},
				},
			},
			{
				Name: getStateDBusName,
				Args: []introspect.Arg{
					{
						Name:      slotArgName,
						Type:      "s",
						Direction: dbusutil.In,
					},
					{
						Name:      stateArgName,
						Type:      "s",
						Direction: dbusutil.Out,
					},
				},
			},
			{
				Name: setStateDBusName,
				Args: []introspect.Arg{
					{
						Name:      slotArgName,
						Type:      "s",
						Direction: dbusutil.In,
					},
					{
						Name:      stateArgName,
						Type:      "s",
						Direction: dbusutil.In,
					},
				},
			},
			{
				Name: preInstallDBusName,
				Args: []introspect.Arg{
					{
						Name: contextArgName,
						// Pass environment block as input.
						// This allows the context to evolve without changing the API.
						Type:      "as",
						Direction: dbusutil.In,
					},
				},
			},
			{
				Name: postInstallDBusName,
				Args: []introspect.Arg{
					{
						Name:      contextArgName,
						Type:      "as",
						Direction: dbusutil.In,
					},
				},
			},
		},
	}

	return reg.RegisterObject(ota.ServiceObjectPath, InterfaceName, methods, metadata)
}

// GetPrimary maps boothandler.Handler.PrimarySlot to D-Bus.
func (svc *Service) GetPrimary() (slotName string, dbusErr *dbus.Error) {
	slot, err := svc.bootHandler.PrimarySlot()
	if err != nil {
		return "", dbus.MakeFailedError(err)
	}

	return slot.String(), nil
}

// SetPrimary maps boothandler.Handler.SetPrimarySlot to D-Bus.
func (svc *Service) SetPrimary(slotName string) (dbusErr *dbus.Error) {
	var slot boot.Slot

	if err := slot.UnmarshalText([]byte(slotName)); err != nil {
		return dbus.MakeFailedError(err)
	}

	if err := svc.bootHandler.SetPrimarySlot(slot); err != nil {
		return dbus.MakeFailedError(err)
	}

	return nil
}

// GetState maps boothandler.Handler.SlotState to D-Bus.
func (svc *Service) GetState(slotName string) (bootState string, dbusErr *dbus.Error) {
	var slot boot.Slot

	if err := slot.UnmarshalText([]byte(slotName)); err != nil {
		return "", dbus.MakeFailedError(err)
	}

	state, err := svc.bootHandler.SlotState(slot)
	if err != nil {
		return "", dbus.MakeFailedError(err)
	}

	return state.String(), nil
}

// SetState maps boothandler.Handler.SetSlotState to D-Bus.
func (svc *Service) SetState(slotName string, stateName string) (dbusErr *dbus.Error) {
	var slot boot.Slot

	if err := slot.UnmarshalText([]byte(slotName)); err != nil {
		return dbus.MakeFailedError(err)
	}

	var state boot.SlotState

	if err := state.UnmarshalText([]byte(stateName)); err != nil {
		return dbus.MakeFailedError(err)
	}

	if err := svc.bootHandler.SetSlotState(slot, state); err != nil {
		return dbus.MakeFailedError(err)
	}

	return nil
}

// PreInstall maps installhandler.Participant.PreInstallHandler to D-Bus.
func (svc *Service) PreInstall(env []string) (dbusErr *dbus.Error) {
	ctx, err := installhandler.NewHandlerContext(env)
	if err != nil {
		return dbus.MakeFailedError(err)
	}

	if err := svc.installHandler.PreInstallHandler(ctx); err != nil {
		return dbus.MakeFailedError(err)
	}

	return nil
}

// PostInstall maps installhandler.Participant.PostInstallHandler to D-Bus.
func (svc *Service) PostInstall(env []string) (dbusErr *dbus.Error) {
	ctx, err := installhandler.NewHandlerContext(env)
	if err != nil {
		return dbus.MakeFailedError(err)
	}

	if err := svc.installHandler.PostInstallHandler(ctx); err != nil {
		return dbus.MakeFailedError(err)
	}

	return nil
}
