// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package ota

import (
	"errors"
	"fmt"
	"os"
	"path/filepath"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/boot"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/encoding/ini"
)

// SystemState contains all the state required by the OTA service.
//
// SystemOTA maintains information about the state of the system necessary
// during the update transaction. State can be marshaled to an INI file and must
// be stored persistently in a location that is immune to data loss during
// rollback, when the update transaction is cancelled.
//
// SystemState does not duplicate information for which the authoritative source
// is the boot.Protocol.
type SystemState struct {
	BootMode boot.Mode `ini:"BootMode,[System],omitempty"`
}

const (
	// StateFileName is the name of the SystemOTA state file.
	//
	// This location cannot be changed easily, at least not without special
	// coordination. Old versions of sysotad will use it on roll-back. Without
	// an epoch-like system, where certain upgrade path is guaranteed, making
	// sure that old systems transition through an intermediate version that
	// performs the migration and locks out rollback, changing this path will
	// lead to broken systems.
	StateFileName = "/var/lib/sysota/state.ini"
)

// LoadState loads the SystemOTA state from the given file.
//
// The file may be missing without causing an error. In such case the zero value
// state is returned.
func LoadState(name string) (state *SystemState, err error) {
	defer func() {
		if err != nil {
			err = &LoadStateError{FileName: name, Err: err}
		}
	}()

	state = &SystemState{}

	f, err := os.Open(name)
	if err != nil {
		if os.IsNotExist(err) {
			return state, nil
		}

		return nil, err
	}

	defer func() {
		e := f.Close()
		if err == nil {
			err = e
		}
	}()

	dec := ini.NewDecoder(f)
	if err := dec.Decode(state); err != nil {
		return nil, err
	}

	return state, nil
}

// LoadStateError records an error with loading the state.
type LoadStateError struct {
	FileName string
	Err      error
}

// Unwrap returns the error wrapped inside the load state error.
func (e *LoadStateError) Unwrap() error {
	return e.Err
}

// Error implements the error interface.
func (e *LoadStateError) Error() string {
	var pathErr *os.PathError

	if errors.As(e.Err, &pathErr) {
		// Omit the path since PathError repeats it.
		return fmt.Sprintf("cannot load state: %v", e.Err)
	}

	return fmt.Sprintf("cannot load state from %s: %v", e.FileName, e.Err)
}

// tempFilePrefix is the prefix of temporary state files.
//
// The prefix can be used with os.CreateTemp to create temporary files or with
// path/filepath.Glob to enumerate them.
const tempFilePrefix = "tmp-*"

// SaveState saves given SystemOTA state to the given file.
//
// The file is not written to directly, instead a temporary file is created in
// the same directory, written to and then renamed.
//
// The entire state directory should be created on the same filesystem to
// support race-free, atomic operation. The state file should not be
// bind-mounted.
func SaveState(state *SystemState, name string) (err error) {
	defer func() {
		if err != nil {
			err = &SaveStateError{Err: err}
		}
	}()

	dirName := filepath.Dir(name)

	// Use a fixed prefix for all temporary files. We can enumerate and remove
	// them reliably in case we are power-cycled after this call but before we
	// manage to clean up in the deferred function below.
	f, err := os.CreateTemp(dirName, tempFilePrefix)
	if err != nil {
		return err
	}
	// TODO(zyga): write TidyState which takes a directory name and handles cleanup.

	var renamed bool

	defer func() {
		e := f.Close()

		if err == nil && e != nil {
			err = e
		}

		// Remove the temporary if we have not renamed it yet.
		if !renamed {
			_ = os.Remove(f.Name())
		}
	}()

	dir, err := os.Open(dirName)
	if err != nil {
		return err
	}

	defer func() {
		e := dir.Close()
		if err == nil {
			err = e
		}
	}()

	enc := ini.NewEncoder(f)
	if err := enc.Encode(state); err != nil {
		return err
	}

	// As the first step of the best-effort attempt to atomically replace a
	// file, synchronize the content of the temporary file.
	if err := f.Sync(); err != nil {
		return err
	}

	// As the second step of the best-effort attempt to atomically replace a
	// file, rename the temporary file over to the desired name.
	if err := os.Rename(f.Name(), name); err != nil {
		return err
	}
	// Once we've renamed the file make sure the deferred cleanup function does
	// not attempt to remove the old name.
	renamed = true

	// As the third and last step of best-effort attempt to atomically replace a
	// file, synchronize the directory it resides in.
	if err := dir.Sync(); err != nil {
		return err
	}

	return nil
}

// SaveStateError records an error with saving the state.
type SaveStateError struct {
	Err error
}

// Unwrap returns the error wrapped inside the save state error.
func (e *SaveStateError) Unwrap() error {
	return e.Err
}

// Error implements the error interface.
func (e *SaveStateError) Error() string {
	return fmt.Sprintf("cannot save state: %v", e.Err)
}
