// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package ota_test

import (
	"errors"
	"io/ioutil"
	"os"
	"path/filepath"

	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/boot"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/ota"
)

type stateSuite struct {
	name string
}

var _ = Suite(&stateSuite{})

func (s *stateSuite) SetUpTest(c *C) {
	d := c.MkDir()
	s.name = filepath.Join(d, "state.ini")
}

func (s *stateSuite) TestLoadStateWithoutStateFile(c *C) {
	state, err := ota.LoadState(s.name)
	c.Assert(err, IsNil)
	c.Check(state.BootMode, Equals, boot.Normal)
}

func (s *stateSuite) TestLoadStateWithStateFile(c *C) {
	err := ioutil.WriteFile(s.name, []byte(`
[System]
BootMode=try
`), 0600)
	c.Assert(err, IsNil)

	state, err := ota.LoadState(s.name)
	c.Assert(err, IsNil)
	c.Check(state.BootMode, Equals, boot.Try)
}

func (s *stateSuite) TestLoadStateWithCorruptedStateFile(c *C) {
	err := ioutil.WriteFile(s.name, []byte(`[System`), 0600)
	c.Assert(err, IsNil)

	_, err = ota.LoadState(s.name)
	c.Assert(err, ErrorMatches, `cannot load state from .*/state.ini: cannot decode line 1: .*`)
}

func (s *stateSuite) TestLoadStateWithBrokenStateFile(c *C) {
	err := os.Symlink(s.name, s.name)
	c.Assert(err, IsNil)

	_, err = ota.LoadState(s.name)
	c.Assert(err, ErrorMatches, `cannot load state: open .*/state.ini: too many levels of symbolic links`)

	c.Check(errors.Unwrap(err), ErrorMatches, `open .*: too many levels of symbolic links`)
}

func (s *stateSuite) TestSaveStateEmpty(c *C) {
	err := ota.SaveState(&ota.SystemState{}, s.name)
	c.Assert(err, IsNil)

	data, err := ioutil.ReadFile(s.name)
	c.Assert(err, IsNil)
	c.Check(string(data), Equals, "")
}

func (s *stateSuite) TestSaveStateNonEmpty(c *C) {
	err := ota.SaveState(&ota.SystemState{BootMode: boot.Try}, s.name)
	c.Assert(err, IsNil)

	data, err := ioutil.ReadFile(s.name)
	c.Assert(err, IsNil)
	c.Check(string(data), Equals, "[System]\nBootMode=try\n")
}

func (s *stateSuite) TestSaveStateFinalPermission(c *C) {
	err := ota.SaveState(&ota.SystemState{}, s.name)
	c.Assert(err, IsNil)

	fi, err := os.Stat(s.name)
	c.Assert(err, IsNil)
	c.Check(fi.Mode(), Equals, os.FileMode(0o600))
}

func (s *stateSuite) TestSaveStateOverwrites(c *C) {
	c.Assert(ioutil.WriteFile(s.name, []byte("[System]\nBootMode=potato\n"), 0600), IsNil)

	err := ota.SaveState(&ota.SystemState{BootMode: boot.Try}, s.name)
	c.Assert(err, IsNil)

	data, err := ioutil.ReadFile(s.name)
	c.Assert(err, IsNil)
	c.Check(string(data), Equals, "[System]\nBootMode=try\n")
}

func (s *stateSuite) TestSaveStateCannotRename(c *C) {
	// Create a directory where the state file would otherwise be.
	c.Assert(os.Mkdir(s.name, 0600), IsNil)
	err := ota.SaveState(&ota.SystemState{BootMode: boot.Try}, s.name)
	c.Assert(err, ErrorMatches, `cannot save state: rename .* .*/state.ini: .*`)
	c.Assert(errors.Unwrap(err), ErrorMatches, `rename .*`)

	// Only the "blocker" directory is present.
	entries, err := os.ReadDir(filepath.Dir(s.name))
	c.Assert(err, IsNil)
	c.Assert(entries, HasLen, 1)
	c.Check(entries[0].Name(), Equals, filepath.Base(s.name))
	c.Check(entries[0].IsDir(), Equals, true)
}

func (s *stateSuite) TestSaveStateCannotMarshal(c *C) {
	err := ota.SaveState(&ota.SystemState{BootMode: boot.Mode(42)}, s.name)
	c.Assert(err, ErrorMatches, `cannot save state: cannot encode SystemState.BootMode: invalid boot mode`)

	// No temporary files present
	entries, err := os.ReadDir(filepath.Dir(s.name))
	c.Assert(err, IsNil)
	c.Assert(entries, HasLen, 0)
}

func (s *stateSuite) TestSaveStateCannotMarshalDoesNotClobberOldState(c *C) {
	c.Assert(ioutil.WriteFile(s.name, []byte("[System]\nBootMode=potato\n"), 0600), IsNil)

	err := ota.SaveState(&ota.SystemState{BootMode: boot.Mode(42)}, s.name)
	c.Assert(err, ErrorMatches, `cannot save state: cannot encode SystemState.BootMode: invalid boot mode`)

	data, err := ioutil.ReadFile(s.name)
	c.Assert(err, IsNil)
	c.Check(string(data), Equals, "[System]\nBootMode=potato\n")
}
