// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package condfilter_test

import (
	"errors"
	"testing"

	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/picfg/condfilter"
)

func Test(t *testing.T) { TestingT(t) }

type unmarshalSuite struct{}

var _ = Suite(&unmarshalSuite{})

func (s *unmarshalSuite) TestUnmarshal(c *C) {
	type testCase struct {
		text   string
		filter condfilter.ConditionalFilter
	}

	for _, t := range []testCase{
		{"[all]", condfilter.All},
		{"[none]", condfilter.None},
		{"[pi0]", condfilter.Pi0},
		{"[pi0w]", condfilter.Pi0W},
		{"[pi1]", condfilter.Pi1},
		{"[pi2]", condfilter.Pi2},
		{"[pi3]", condfilter.Pi3},
		{"[pi3+]", condfilter.Pi3Plus},
		{"[pi4]", condfilter.Pi4},
		{"[pi400]", condfilter.Pi400},
		{"[cm4]", condfilter.Cm4},
		{"[EDID=foo]", condfilter.EDIDFilter("foo")},
		{"[gpio2=1]", condfilter.GPIOFilter{PinNumber: 2, High: true}},
		{"[hdmi:1]", condfilter.HDMIFilter{PortNumber: 1}},
		{"[0x1234]", condfilter.SerialFilter(0x1234)},
	} {
		filter, err := condfilter.Unmarshal([]byte(t.text))
		c.Assert(err, IsNil)
		c.Check(filter, Equals, t.filter)
	}
}

func (s *unmarshalSuite) TestMustUnmarshal(c *C) {
	c.Check(condfilter.MustUnmarshal([]byte(`[all]`)), Equals, condfilter.All)
	c.Check(func() { _ = condfilter.MustUnmarshal([]byte(`[potato]`)) }, PanicMatches, `conditional filter error "\[potato\]": unknown filter`)
}

func (s *unmarshalSuite) TestUnmarshalUnknownFilter(c *C) {
	_, err := condfilter.Unmarshal([]byte("[potato]"))
	c.Assert(err, ErrorMatches, `conditional filter error "\[potato\]": unknown filter`)
}

func (s *unmarshalSuite) TestUnmarshalSyntaxError(c *C) {
	_, err := condfilter.Unmarshal([]byte("[potato]"))
	c.Assert(err, ErrorMatches, `conditional filter error "\[potato\]": unknown filter`)

	for _, potato := range []string{"[potato", "potato]", "potato"} {
		_, err = condfilter.Unmarshal([]byte(potato))
		c.Assert(err, ErrorMatches, `conditional filter error ".*": syntax error`)
	}
}

type filterErrorSuite struct{}

var _ = Suite(&filterErrorSuite{})

var errJustTesting = errors.New("just testing")

func (*filterErrorSuite) TestError(c *C) {
	// Typeless form
	err := &condfilter.FilterError{
		Filter: "[foo]",
		Err:    errJustTesting,
	}
	c.Check(err, ErrorMatches, `conditional filter error "\[foo\]": just testing`)

	// Type-specific form
	err.Type = "FooType"
	c.Check(err, ErrorMatches, `FooType "\[foo\]" error: just testing`)
}

func (*filterErrorSuite) TestUnwrap(c *C) {
	err := &condfilter.FilterError{Err: errJustTesting}
	c.Check(errors.Unwrap(err), ErrorMatches, `just testing`)
}
