// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package condfilter_test

import (
	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/picfg"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/picfg/condfilter"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/picfg/pimodel"
)

type simSuite struct{}

var _ = Suite(&simSuite{})

// A number of revisions that have specific properties. Here we mostly care
// about the board type and in a specific case, about the processor used.

const (
	Cm1RevCode         = 0x0014
	Cm3PlusRevCode     = 0xa02100
	Cm3RevCode         = 0xa020a0
	Cm4RevCode         = 0xa03140
	Pi0RevCode         = 0x900092
	Pi0WRevCode        = 0x9000c1
	Pi1APlusRevCode    = 0x0012
	Pi1ARevCode        = 0x0007
	Pi1BPlusRevCode    = 0x0010
	Pi1BRevCode        = 0x0002
	Pi2BRevCodeBcm2836 = 0xa01040
	Pi2BRevCodeBcm2837 = 0xa02042
	Pi3APlusRevCode    = 0x9020e0
	Pi3BPlusRevCode    = 0xa020d3
	Pi3BRevCode        = 0xa02082
	Pi400RevCode       = 0xc03130
	Pi4BRevCode        = 0xa03111
)

const sampleSerial = 0x10000000aa586ea0

func (s *simSuite) TestNewSimulatorSupportedFilters(c *C) {
	cfg := &picfg.ConfigTxt{Sections: []picfg.Section{
		{Filter: "[all]", Location: picfg.Location{FileName: "config.txt", LineNo: 1}}},
	}
	_, err := condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, IsNil)

	cfg = &picfg.ConfigTxt{Sections: []picfg.Section{
		{Filter: "[none]", Location: picfg.Location{FileName: "config.txt", LineNo: 1}}},
	}
	_, err = condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, IsNil)

	cfg = &picfg.ConfigTxt{Sections: []picfg.Section{
		{Filter: "[pi4]", Location: picfg.Location{FileName: "config.txt", LineNo: 1}}},
	}
	_, err = condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, IsNil)

	cfg = &picfg.ConfigTxt{Sections: []picfg.Section{
		{Filter: "[0x1234]", Location: picfg.Location{FileName: "config.txt", LineNo: 1}}},
	}
	_, err = condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, IsNil)
}

func (s *simSuite) TestNewSimulatorUnsupportedFilters(c *C) {
	// HDMI filter is not supported.
	//
	// It could plausibly be supported as long as the section do not contain any
	// settings relevant to the boot process.
	cfg := &picfg.ConfigTxt{Sections: []picfg.Section{
		{Filter: "[hdmi:1]", Location: picfg.Location{FileName: "config.txt", LineNo: 1}}},
	}
	_, err := condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, ErrorMatches, `conditional filter \[hdmi:1\] \(from config.txt:1\) is not supported`)

	// EDID filter is not supported.
	//
	// It could plausibly be supported as long as the section do not contain any
	// settings relevant to the boot process.
	cfg = &picfg.ConfigTxt{Sections: []picfg.Section{
		{Filter: "[EDID=VSC-TD2220]", Location: picfg.Location{FileName: "config.txt", LineNo: 1}}},
	}
	_, err = condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, ErrorMatches, `conditional filter \[EDID=VSC-TD2220\] \(from config.txt:1\) is not supported`)

	// GPIO filter is not supported.
	//
	// It could plausibly be supported as long as the section do not contain any
	// settings relevant to the boot process.
	cfg = &picfg.ConfigTxt{Sections: []picfg.Section{
		{Filter: "[gpio12=1]", Location: picfg.Location{FileName: "config.txt", LineNo: 1}}},
	}
	_, err = condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, ErrorMatches, `conditional filter \[gpio12=1\] \(from config.txt:1\) is not supported`)
}

func (s *simSuite) TestNewSimulatorInvalidFilter(c *C) {
	cfg := &picfg.ConfigTxt{Sections: []picfg.Section{
		{Filter: "[potato]", Location: picfg.Location{FileName: "config.txt", LineNo: 1}}},
	}
	_, err := condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, ErrorMatches, `conditional filter error "\[potato\]": unknown filter`)
}

func (s *simSuite) TestNewSimulatorInvalidRevision(c *C) {
	cfg := &picfg.ConfigTxt{}
	_, err := condfilter.NewSimulator(cfg, 0, sampleSerial)
	c.Assert(err, ErrorMatches, `cannot decode revision code 0000, unknown board type`)
}

func (s *simSuite) TestConditionalFilters(c *C) {
	cfg := &picfg.ConfigTxt{
		Sections: []picfg.Section{{
			Filter:     "[all]",
			Parameters: []picfg.Parameter{{Name: "cmdline", Value: "all.txt"}},
		}, {
			Filter:     "[pi3]",
			Parameters: []picfg.Parameter{{Name: "cmdline", Value: "pi3.txt"}},
		}, {
			Filter:     "[pi4]",
			Parameters: []picfg.Parameter{{Name: "cmdline", Value: "pi4.txt"}},
		}, {
			Filter:     "[none]",
			Parameters: []picfg.Parameter{{Name: "cmdline", Value: "none.txt"}},
		}},
	}

	for _, t := range []struct {
		code  pimodel.RevisionCode
		value string
	}{
		{code: Pi1BRevCode, value: "all.txt"},
		{code: Pi2BRevCodeBcm2836, value: "all.txt"},
		{code: Pi4BRevCode, value: "pi4.txt"},
		{code: Pi3BRevCode, value: "pi3.txt"},
	} {
		comment := Commentf("t: %#v", t)

		sim, err := condfilter.NewSimulator(cfg, t.code, sampleSerial)
		c.Assert(err, IsNil, comment)

		param := sim.Parameter("cmdline")
		c.Assert(param, NotNil, comment)
		c.Check(param.Value, Equals, t.value, comment)
	}
}

func (s *simSuite) TestCoherentParameterLastDefinition(c *C) {
	// CoherentParameter picks up the last definition.
	cfg := &picfg.ConfigTxt{
		Sections: []picfg.Section{{
			Parameters: []picfg.Parameter{{
				Name:     "os_prefix",
				Value:    "slot-a/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 1},
			}, {
				Name:     "os_prefix",
				Value:    "slot-a/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 2},
			}},
		}},
	}

	sim, err := condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, IsNil)

	param, err := sim.CoherentParameter("os_prefix")
	c.Assert(err, IsNil)
	c.Check(param.Value, Equals, "slot-a/")
	// Note that we pick up the second definition.
	c.Check(param.Location, DeepEquals, picfg.Location{FileName: "config.txt", LineNo: 2})
}

func (s *simSuite) TestCoherentParameterSkipsNoneSection(c *C) {
	// CoherentParameter skips the [none] section and does look at anything
	// inside.
	cfg := &picfg.ConfigTxt{
		Sections: []picfg.Section{{
			Parameters: []picfg.Parameter{{
				Name:     "os_prefix",
				Value:    "slot-a/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 1},
			}},
		}, {
			Filter: "[none]",
			Parameters: []picfg.Parameter{{
				Name:     "os_prefix",
				Value:    "potato/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 3},
			}},
		}},
	}

	sim, err := condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, IsNil)

	param, err := sim.CoherentParameter("os_prefix")
	c.Assert(err, IsNil)
	c.Assert(param, NotNil)
	c.Check(param.Location.LineNo, Equals, 1)
}

func (s *simSuite) TestSetCoherentParameterSkipsNoneSection(c *C) {
	// SetCoherentParameter skips the [none] section and does not update
	// anything inside.
	cfg := &picfg.ConfigTxt{
		Sections: []picfg.Section{{
			Parameters: []picfg.Parameter{{
				Name:     "os_prefix",
				Value:    "slot-a/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 1},
			}},
		}, {
			Filter: "[none]",
			Parameters: []picfg.Parameter{{
				Name:     "os_prefix",
				Value:    "potato/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 3},
			}},
		}},
	}

	sim, err := condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, IsNil)

	err = sim.SetCoherentParameter("os_prefix", "slot-b/")
	c.Assert(err, IsNil)

	c.Check(cfg, DeepEquals, &picfg.ConfigTxt{
		Sections: []picfg.Section{{
			Parameters: []picfg.Parameter{{
				Name:     "os_prefix",
				Value:    "slot-b/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 1},
			}},
		}, {
			Filter: "[none]",
			Parameters: []picfg.Parameter{{
				Name:     "os_prefix",
				Value:    "potato/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 3},
			}},
		}},
	})
}

func (s *simSuite) TestSetCoherentParameterAddsAllSection(c *C) {
	// SetCoherentParameter adds a new [all] section is one is not available.
	cfg := &picfg.ConfigTxt{
		Sections: []picfg.Section{{
			Filter: "[pi4]",
		}},
	}

	sim, err := condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, IsNil)

	err = sim.SetCoherentParameter("os_prefix", "slot-b/")
	c.Assert(err, IsNil)

	c.Check(cfg, DeepEquals, &picfg.ConfigTxt{
		Sections: []picfg.Section{
			{
				Filter: "[pi4]",
			},
			{
				Filter: "[all]",
				Parameters: []picfg.Parameter{{
					Name:  "os_prefix",
					Value: "slot-b/",
				}},
			},
		},
	})
}

func (s *simSuite) TestSetCoherentParameterRefusesToClobberNonCoherent(c *C) {
	// SetCoherentParameter refuses to modify a non-coherent parameter.
	cfg := &picfg.ConfigTxt{
		Sections: []picfg.Section{{
			Filter: "[pi4]",
			Parameters: []picfg.Parameter{{
				Name:     "os_prefix",
				Value:    "potato/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 2},
			}},
		}, {
			Filter: "[all]",
			Parameters: []picfg.Parameter{{
				Name:     "os_prefix",
				Value:    "tomato/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 4},
			}},
		}},
	}

	sim, err := condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, IsNil)

	err = sim.SetCoherentParameter("os_prefix", "slot-b/")
	c.Assert(err, ErrorMatches, `conflicting definition of os_prefix, first defined in config.txt:2 as potato/, then in config.txt:4 as tomato/`)
}

func (s *simSuite) TestSetCoherentParameterUpdatesAllInstances(c *C) {
	// SetCoherentParameter updates all the definitions.
	cfg := &picfg.ConfigTxt{
		Sections: []picfg.Section{{
			Parameters: []picfg.Parameter{{
				Name:     "os_prefix",
				Value:    "slot-a/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 1},
			}, {
				Name:     "os_prefix",
				Value:    "slot-a/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 2},
			}},
		}},
	}

	sim, err := condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, IsNil)

	err = sim.SetCoherentParameter("os_prefix", "slot-b/")
	c.Assert(err, IsNil)

	c.Check(cfg, DeepEquals, &picfg.ConfigTxt{
		Sections: []picfg.Section{{
			Parameters: []picfg.Parameter{{
				Name:     "os_prefix",
				Value:    "slot-b/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 1},
			}, {
				Name:     "os_prefix",
				Value:    "slot-b/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 2},
			}},
		}},
	})
}

func (s *simSuite) TestSetCoherentParameterAddsParamaterToLastAllSection(c *C) {
	// SetCoherentParameter updates all the definitions.
	cfg := &picfg.ConfigTxt{
		Sections: []picfg.Section{
			{},                // Implicit [all] section.
			{Filter: "[all]"}, // First explicit [all] section.
			{Filter: "[all]"}, // Second explicit [all] section
		},
	}

	sim, err := condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, IsNil)

	err = sim.SetCoherentParameter("os_prefix", "slot-a/")
	c.Assert(err, IsNil)

	c.Check(cfg, DeepEquals, &picfg.ConfigTxt{
		Sections: []picfg.Section{
			{},
			{Filter: "[all]"},
			{Filter: "[all]", Parameters: []picfg.Parameter{{Name: "os_prefix", Value: "slot-a/"}}},
		},
	})
}

func (s *simSuite) TestNonCoherentNonMatchingOSPrefix(c *C) {
	// The os_prefix= parameter should be coherent among all the sections, even
	// those not matching the current model. Here the Pi4B model matches [pi4]
	// but not the [pi3] above it.
	cfg := &picfg.ConfigTxt{
		Sections: []picfg.Section{{
			Filter: "[pi3]",
			Parameters: []picfg.Parameter{{
				Name:     "os_prefix",
				Value:    "potato/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 2},
			}},
		}, {
			Filter: "[pi4]",
			Parameters: []picfg.Parameter{{
				Name:     "os_prefix",
				Value:    "tomato/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 4},
			}},
		}},
	}

	sim, err := condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, IsNil)

	_, err = sim.CoherentParameter("os_prefix")
	c.Assert(err, ErrorMatches, `conflicting definition of os_prefix, `+
		`first defined in config.txt:2 as potato/, `+
		`then in config.txt:4 as tomato/`)
}

func (s *simSuite) TestNonCoherentMatchingOSPrefix(c *C) {
	// The os_prefix= parameter should be coherent among all the sections
	// matching the current model. Here the Pi4B model matches both [pi4] and
	// [all].
	cfg := &picfg.ConfigTxt{
		Sections: []picfg.Section{{
			Filter: "[pi4]",
			Parameters: []picfg.Parameter{{
				Name:     "os_prefix",
				Value:    "potato/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 2},
			}},
		}, {
			Filter: "[all]",
			Parameters: []picfg.Parameter{{
				Name:     "os_prefix",
				Value:    "tomato/",
				Location: picfg.Location{FileName: "config.txt", LineNo: 4},
			}},
		}},
	}

	sim, err := condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, IsNil)

	_, err = sim.CoherentParameter("os_prefix")
	c.Assert(err, ErrorMatches, `conflicting definition of os_prefix, `+
		`first defined in config.txt:2 as potato/, `+
		`then in config.txt:4 as tomato/`)
}

func (s *simSuite) TestParameterLookup(c *C) {
	cfg := &picfg.ConfigTxt{
		Sections: []picfg.Section{{
			Parameters: []picfg.Parameter{
				{Name: "param", Value: "value-1"},
				{Name: "other", Value: "value-2"},
			},
		}},
	}

	sim, err := condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, IsNil)

	param := sim.Parameter("param")
	c.Check(param.Name, Equals, "param")
	c.Check(param.Value, Equals, "value-1")

	param = sim.Parameter("other")
	c.Check(param.Name, Equals, "other")
	c.Check(param.Value, Equals, "value-2")

	param = sim.Parameter("missing")
	c.Check(param, IsNil)
}

func (s *simSuite) TestCoherentParameterLookup(c *C) {
	cfg := &picfg.ConfigTxt{
		Sections: []picfg.Section{{
			Parameters: []picfg.Parameter{
				{Name: "param", Value: "value-1"},
				{Name: "other", Value: "value-2"},
			},
		}},
	}

	sim, err := condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, IsNil)

	param, err := sim.CoherentParameter("param")
	c.Assert(err, IsNil)
	c.Check(param.Name, Equals, "param")
	c.Check(param.Value, Equals, "value-1")

	param, err = sim.CoherentParameter("other")
	c.Assert(err, IsNil)
	c.Check(param.Name, Equals, "other")
	c.Check(param.Value, Equals, "value-2")

	param, err = sim.CoherentParameter("missing")
	c.Assert(err, IsNil)
	c.Check(param, IsNil)
}

func (s *simSuite) TestParameterFindsModelMatch(c *C) {
	cfg := &picfg.ConfigTxt{
		Sections: []picfg.Section{{
			Filter:     "[pi4]",
			Parameters: []picfg.Parameter{{Name: "param", Value: "value"}},
		}},
	}

	sim, err := condfilter.NewSimulator(cfg, Pi4BRevCode, sampleSerial)
	c.Assert(err, IsNil)

	param := sim.Parameter("param")
	c.Check(param.Name, Equals, "param")
	c.Check(param.Value, Equals, "value")

	sim, err = condfilter.NewSimulator(cfg, Pi3BRevCode, sampleSerial) // non-matching case.
	c.Assert(err, IsNil)

	param = sim.Parameter("param")
	c.Check(param, IsNil)
}

func (s *simSuite) TestParameterFindsSerialMatch(c *C) {
	cfg := &picfg.ConfigTxt{
		Sections: []picfg.Section{{
			Filter:     "[0x1234]",
			Parameters: []picfg.Parameter{{Name: "param", Value: "value"}},
		}},
	}

	sim, err := condfilter.NewSimulator(cfg, Pi4BRevCode, 0x1234)
	c.Assert(err, IsNil)

	param := sim.Parameter("param")
	c.Check(param.Name, Equals, "param")
	c.Check(param.Value, Equals, "value")

	sim, err = condfilter.NewSimulator(cfg, Pi4BRevCode, 0x5678) // non-matching case.
	c.Assert(err, IsNil)

	param = sim.Parameter("param")
	c.Check(param, IsNil)
}
