// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package condfilter_test

import (
	"errors"
	"strconv"

	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/picfg/condfilter"
)

type gpioFilterSuite struct {
	filter condfilter.GPIOFilter
}

var _ = Suite(&gpioFilterSuite{})

func (s *gpioFilterSuite) TestUnmarshalText(c *C) {
	var paramErr *condfilter.ParameterError

	err := s.filter.UnmarshalText([]byte(`[gpio123=0]`))
	c.Assert(err, IsNil)
	c.Check(s.filter, Equals, condfilter.GPIOFilter{PinNumber: 123, High: false})

	err = s.filter.UnmarshalText([]byte(`[gpio123=1]`))
	c.Assert(err, IsNil)
	c.Check(s.filter, Equals, condfilter.GPIOFilter{PinNumber: 123, High: true})

	err = s.filter.UnmarshalText([]byte(`[gpio` + `foo=1]`))
	c.Assert(err, ErrorMatches, `GPIOFilter "\[gpio`+`foo=1\]" error: invalid GPIO pin number: invalid syntax`)
	c.Check(errors.Is(err, strconv.ErrSyntax), Equals, true)
	c.Assert(errors.As(err, &paramErr), Equals, true)
	c.Check(paramErr.What, Equals, condfilter.GPIOPinNumber)

	err = s.filter.UnmarshalText([]byte(`[gpio=1]`))
	c.Check(errors.Is(err, condfilter.ErrDifferent), Equals, true)

	err = s.filter.UnmarshalText([]byte(`[gpio123=2]`))
	c.Assert(err, ErrorMatches, `GPIOFilter "\[gpio123=2\]" error: invalid GPIO pin state: value out of range`)
	c.Check(errors.Is(err, strconv.ErrRange), Equals, true)

	err = s.filter.UnmarshalText([]byte(`[gpio123=bar]`))
	c.Assert(err, ErrorMatches, `GPIOFilter "\[gpio123=bar\]" error: invalid GPIO pin state: invalid syntax`)
	c.Check(errors.Is(err, strconv.ErrSyntax), Equals, true)
	c.Assert(errors.As(err, &paramErr), Equals, true)
	c.Check(paramErr.What, Equals, condfilter.GPIOPinState)

	err = s.filter.UnmarshalText([]byte(`[gpio123=]`))
	c.Assert(err, ErrorMatches, `GPIOFilter "\[gpio123=\]" error: invalid GPIO pin state: parameter value required`)
	c.Check(errors.Is(err, condfilter.ErrParameterRequired), Equals, true)
	c.Assert(errors.As(err, &paramErr), Equals, true)
	c.Check(paramErr.What, Equals, condfilter.GPIOPinState)

	err = s.filter.UnmarshalText([]byte(`[gpio123]`))
	c.Assert(err, ErrorMatches, `GPIOFilter "\[gpio123\]" error: different filter`)
	c.Check(errors.Is(err, condfilter.ErrDifferent), Equals, true)

	err = s.filter.UnmarshalText([]byte(`[foo`))
	c.Assert(err, ErrorMatches, `GPIOFilter "\[foo" error: syntax error`)
	c.Check(errors.Is(err, condfilter.ErrSyntax), Equals, true)

	err = s.filter.UnmarshalText([]byte(`foo]`))
	c.Assert(err, ErrorMatches, `GPIOFilter "foo\]" error: syntax error`)
	c.Check(errors.Is(err, condfilter.ErrSyntax), Equals, true)

	err = s.filter.UnmarshalText([]byte(`[foo]`))
	c.Assert(err, ErrorMatches, `GPIOFilter "\[foo\]" error: different filter`)
	c.Check(errors.Is(err, condfilter.ErrDifferent), Equals, true)
}

func (s *gpioFilterSuite) TestMarshalText(c *C) {
	s.filter.PinNumber = 123
	s.filter.High = false
	b, err := s.filter.MarshalText()
	c.Assert(err, IsNil)
	c.Check(string(b), Equals, `[gpio123=0]`)

	s.filter.PinNumber = 123
	s.filter.High = true
	b, err = s.filter.MarshalText()
	c.Assert(err, IsNil)
	c.Check(string(b), Equals, `[gpio123=1]`)
}

func (s *gpioFilterSuite) TestConditionalFilter(c *C) {
	s.filter.PinNumber = 123
	s.filter.High = false
	c.Check(s.filter.ConditionalFilter(), Equals, `[gpio123=0]`)

	s.filter.PinNumber = 123
	s.filter.High = true
	c.Check(s.filter.ConditionalFilter(), Equals, `[gpio123=1]`)
}
