// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package condfilter_test

import (
	"errors"
	"strconv"

	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/picfg/condfilter"
)

type hdmiFilterSuite struct {
	filter condfilter.HDMIFilter
}

var _ = Suite(&hdmiFilterSuite{})

func (s *hdmiFilterSuite) TestUnmarshalText(c *C) {
	var paramErr *condfilter.ParameterError

	err := s.filter.UnmarshalText([]byte(`[hdmi:1]`))
	c.Assert(err, IsNil)
	c.Check(s.filter.PortNumber, Equals, uint(1))

	err = s.filter.UnmarshalText([]byte(`[hdmi:-1]`))
	c.Assert(err, ErrorMatches, `HDMIFilter "\[hdmi:-1\]" error: invalid HDMI port number: invalid syntax`)
	c.Check(errors.Is(err, strconv.ErrSyntax), Equals, true)
	c.Assert(errors.As(err, &paramErr), Equals, true)
	c.Check(paramErr.What, Equals, condfilter.HDMIPortNumber)

	err = s.filter.UnmarshalText([]byte(`[hdmi:foo]`))
	c.Assert(err, ErrorMatches, `HDMIFilter "\[hdmi:foo\]" error: invalid HDMI port number: invalid syntax`)
	c.Check(errors.Is(err, strconv.ErrSyntax), Equals, true)
	c.Assert(errors.As(err, &paramErr), Equals, true)
	c.Check(paramErr.What, Equals, condfilter.HDMIPortNumber)

	err = s.filter.UnmarshalText([]byte(`[hdmi:]`))
	c.Assert(err, ErrorMatches, `HDMIFilter "\[hdmi:\]" error: invalid HDMI port number: parameter value required`)
	c.Check(errors.Is(err, condfilter.ErrParameterRequired), Equals, true)
	c.Assert(errors.As(err, &paramErr), Equals, true)
	c.Check(paramErr.What, Equals, condfilter.HDMIPortNumber)

	err = s.filter.UnmarshalText([]byte(`[hdmi]`))
	c.Assert(err, ErrorMatches, `HDMIFilter "\[hdmi\]" error: different filter`)
	c.Check(errors.Is(err, condfilter.ErrDifferent), Equals, true)

	err = s.filter.UnmarshalText([]byte(`[foo`))
	c.Assert(err, ErrorMatches, `HDMIFilter "\[foo" error: syntax error`)
	c.Check(errors.Is(err, condfilter.ErrSyntax), Equals, true)

	err = s.filter.UnmarshalText([]byte(`foo]`))
	c.Assert(err, ErrorMatches, `HDMIFilter "foo\]" error: syntax error`)
	c.Check(errors.Is(err, condfilter.ErrSyntax), Equals, true)

	err = s.filter.UnmarshalText([]byte(`[foo]`))
	c.Assert(err, ErrorMatches, `HDMIFilter "\[foo\]" error: different filter`)
	c.Check(errors.Is(err, condfilter.ErrDifferent), Equals, true)

	err = s.filter.UnmarshalText([]byte(`[foo=]`))
	c.Assert(err, ErrorMatches, `HDMIFilter "\[foo=\]" error: different filter`)
	c.Check(errors.Is(err, condfilter.ErrDifferent), Equals, true)
}

func (s *hdmiFilterSuite) TestMarshalText(c *C) {
	s.filter.PortNumber = 123
	b, err := s.filter.MarshalText()
	c.Assert(err, IsNil)
	c.Check(string(b), Equals, `[hdmi:123]`)
}

func (s *hdmiFilterSuite) TestConditionalFilter(c *C) {
	s.filter.PortNumber = 123
	c.Check(s.filter.ConditionalFilter(), Equals, `[hdmi:123]`)
}
