// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package condfilter_test

import (
	"errors"

	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/picfg/condfilter"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/picfg/pimodel"
)

type piModelFilterSuite struct {
	filter condfilter.PiModelFilter
}

var _ = Suite(&piModelFilterSuite{})

var allModelFilters = map[string]condfilter.PiModelFilter{
	"[pi0]":   condfilter.Pi0,
	"[pi0w]":  condfilter.Pi0W,
	"[pi1]":   condfilter.Pi1,
	"[pi2]":   condfilter.Pi2,
	"[pi3]":   condfilter.Pi3,
	"[pi3+]":  condfilter.Pi3Plus,
	"[pi4]":   condfilter.Pi4,
	"[pi400]": condfilter.Pi400,
	"[cm4]":   condfilter.Cm4,
}

func (s *piModelFilterSuite) TestUnmarshalText(c *C) {
	for text, filter := range allModelFilters {
		err := s.filter.UnmarshalText([]byte(text))
		c.Assert(err, IsNil)
		c.Check(s.filter, Equals, filter)
	}

	err := s.filter.UnmarshalText([]byte(`[foo`))
	c.Assert(err, ErrorMatches, `PiModelFilter "\[foo" error: syntax error`)
	c.Check(errors.Is(err, condfilter.ErrSyntax), Equals, true)

	err = s.filter.UnmarshalText([]byte(`foo]`))
	c.Assert(err, ErrorMatches, `PiModelFilter "foo\]" error: syntax error`)
	c.Check(errors.Is(err, condfilter.ErrSyntax), Equals, true)

	err = s.filter.UnmarshalText([]byte(`[foo]`))
	c.Assert(err, ErrorMatches, `PiModelFilter "\[foo\]" error: different filter`)
	c.Check(errors.Is(err, condfilter.ErrDifferent), Equals, true)
}

func (s *piModelFilterSuite) TestMarshalText(c *C) {
	for text, filter := range allModelFilters {
		b, err := filter.MarshalText()
		c.Assert(err, IsNil)
		c.Check(string(b), Equals, text)
	}

	s.filter = condfilter.PiModelFilter(1000000)
	b, err := s.filter.MarshalText()
	c.Assert(err, IsNil)
	// XXX: Perhaps unexpected?
	c.Check(string(b), Equals, "")
}

func (s *piModelFilterSuite) TestMatchesModel(c *C) {
	testMatrix := map[condfilter.PiModelFilter]map[pimodel.Model]bool{
		// [pi0] matches Pi0 and Pi0W
		condfilter.Pi0: {
			pimodel.Pi0:  true,
			pimodel.Pi0W: true,
		},
		// [pi0w] matches only Pi0W
		condfilter.Pi0W: {
			pimodel.Pi0W: true,
		},
		// [pi1] matches Pi1A, Pi1B and Cm1
		condfilter.Pi1: {
			pimodel.Pi1A: true,
			pimodel.Pi1B: true,
			pimodel.Cm1:  true,
		},
		// [pi2] matches Pi2B
		condfilter.Pi2: {
			pimodel.Pi2B: true,
		},
		// [pi3] matches Pi3B, Pi3APlus, Pi3BPlus and Cm3
		condfilter.Pi3: {
			pimodel.Pi3B:     true,
			pimodel.Pi3APlus: true,
			pimodel.Pi3BPlus: true,
			pimodel.Cm3:      true,
		},
		// [pi3+] matches Pi3APlus and Pi3BPlus
		condfilter.Pi3Plus: {
			pimodel.Pi3APlus: true,
			pimodel.Pi3BPlus: true,
		},
		// [pi4] matches Pi4B, Pi400 and Cm4
		condfilter.Pi4: {
			pimodel.Pi4B:  true,
			pimodel.Pi400: true,
			pimodel.Cm4:   true,
		},
		// [pi400] matches only Pi400
		condfilter.Pi400: {
			pimodel.Pi400: true,
		},
		// [cm4] matches only Cm4
		condfilter.Cm4: {
			pimodel.Cm4: true,
		},
	}

	// Perform exhaustive check
	for filter, cases := range testMatrix {
		for model, shouldMatch := range cases {
			c.Check(filter.MatchesModel(model), Equals, shouldMatch, Commentf("filter %s, model: %s", filter, model))
		}
	}
}
