// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package condfilter_test

import (
	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/picfg/condfilter"
)

type stateTrackerSuite struct {
	st condfilter.StateTracker
}

var _ = Suite(&stateTrackerSuite{})

func (s *stateTrackerSuite) SetUpTest(c *C) {
	s.st = condfilter.StateTracker{}
}

func (s *stateTrackerSuite) TestEffectiveFilters(c *C) {
	// When there are no conditional filters, it's equivalent to the [all] filter.
	c.Check(s.st.EffectiveFilters(), DeepEquals, []condfilter.ConditionalFilter{condfilter.All})

	// Conditional filters of the same type override each other
	s.st.Filter(condfilter.Pi0)
	c.Check(s.st.EffectiveFilters(), DeepEquals, []condfilter.ConditionalFilter{condfilter.Pi0})
	s.st.Filter(condfilter.Pi1)
	c.Check(s.st.EffectiveFilters(), DeepEquals, []condfilter.ConditionalFilter{condfilter.Pi1})
	s.st.Filter(condfilter.Pi2)
	c.Check(s.st.EffectiveFilters(), DeepEquals, []condfilter.ConditionalFilter{condfilter.Pi2})

	// Unrelated filters stack. There are separate types for: model, edid, hdmi, serial and gpio
	s.st.Filter(condfilter.Pi4)
	s.st.Filter(condfilter.EDIDFilter("foo"))
	s.st.Filter(condfilter.HDMIFilter{PortNumber: 1})
	s.st.Filter(condfilter.SerialFilter(0x1234))
	s.st.Filter(condfilter.GPIOFilter{PinNumber: 4, High: true})

	// Order of effective filters is fixed, regardless of order in which filters are presented.
	c.Check(s.st.EffectiveFilters(), DeepEquals, []condfilter.ConditionalFilter{
		condfilter.Pi4,
		condfilter.EDIDFilter("foo"),
		condfilter.GPIOFilter{PinNumber: 4, High: true},
		condfilter.SerialFilter(0x1234),
		condfilter.HDMIFilter{PortNumber: 1},
	})
	c.Check(s.st.IsNone(), Equals, false)
	c.Check(s.st.IsAll(), Equals, false)

	// The [none] filter overrides all other filters
	s.st.Filter(condfilter.None)
	c.Check(s.st.EffectiveFilters(), DeepEquals, []condfilter.ConditionalFilter{condfilter.None})

	c.Check(s.st.IsNone(), Equals, true)
	c.Check(s.st.IsAll(), Equals, false)

	// The [all] filter resets all previous filters
	s.st.Filter(condfilter.All)
	c.Check(s.st.EffectiveFilters(), DeepEquals, []condfilter.ConditionalFilter{condfilter.All})

	c.Check(s.st.IsNone(), Equals, false)
	c.Check(s.st.IsAll(), Equals, true)
}

func (s *stateTrackerSuite) TestAllFilter(c *C) {
	f := condfilter.All
	s.st.Filter(&f)
	s.st.Filter(f)

	c.Check(s.st.IsNone(), Equals, false)
	c.Check(s.st.IsAll(), Equals, true)
}

func (s *stateTrackerSuite) TestNoneFilter(c *C) {
	f := condfilter.None
	s.st.Filter(&f)
	s.st.Filter(f)

	c.Check(s.st.IsNone(), Equals, true)
	c.Check(s.st.IsAll(), Equals, false)
}

func (s *stateTrackerSuite) TestPiModelFilter(c *C) {
	f := condfilter.Pi0
	s.st.Filter(&f)
	s.st.Filter(f)

	c.Check(s.st.IsNone(), Equals, false)
	c.Check(s.st.IsAll(), Equals, false)
}

func (s *stateTrackerSuite) TestSerialFilter(c *C) {
	f := condfilter.SerialFilter(0x1234)
	s.st.Filter(&f)
	s.st.Filter(f)

	c.Check(s.st.IsNone(), Equals, false)
	c.Check(s.st.IsAll(), Equals, false)
}

func (s *stateTrackerSuite) TestGPIOFilter(c *C) {
	f := condfilter.GPIOFilter{PinNumber: 1, High: true}
	s.st.Filter(&f)
	s.st.Filter(f)

	c.Check(s.st.IsNone(), Equals, false)
	c.Check(s.st.IsAll(), Equals, false)
}

func (s *stateTrackerSuite) TestEDIDFilter(c *C) {
	f := condfilter.EDIDFilter("foo")
	s.st.Filter(&f)
	s.st.Filter(f)

	c.Check(s.st.IsNone(), Equals, false)
	c.Check(s.st.IsAll(), Equals, false)
}

func (s *stateTrackerSuite) TestHDMIFilter(c *C) {
	f := condfilter.HDMIFilter{PortNumber: 1}
	s.st.Filter(&f)
	s.st.Filter(f)

	c.Check(s.st.IsNone(), Equals, false)
	c.Check(s.st.IsAll(), Equals, false)
}

type mysteryFilter struct{}

func (mysteryFilter) ConditionalFilter() string {
	return "[???]"
}

func (s *stateTrackerSuite) TestUnknownFilter(c *C) {
	c.Check(func() { s.st.Filter(mysteryFilter{}) }, PanicMatches, "unexpected filter type: condfilter_test.mysteryFilter")
}
