// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package picfg

import (
	"bufio"
	"bytes"
	"errors"
	"fmt"
	"io"
	"os"
)

var errTooLong = errors.New("cannot parse line longer than 80 bytes")

// Decoder reads and decodes config.txt from an input stream.
type Decoder struct {
	r io.Reader
}

// NewDecoder returns a decoder that reads from r.
func NewDecoder(r io.Reader) *Decoder {
	return &Decoder{r: r}
}

func (dec *Decoder) fileName() string {
	if f, ok := dec.r.(*os.File); ok {
		return f.Name()
	}

	return ""
}

// Decode reads the complete config.txt file and stores it in cfg.
func (dec *Decoder) Decode(cfg *ConfigTxt) error {
	var newCfg ConfigTxt

	var seqNum uint

	fileName := dec.fileName()
	sect := Section{Location: Location{FileName: fileName}}

	scanner := bufio.NewScanner(dec.r)
	for lineno := 1; scanner.Scan(); lineno++ {
		line := scanner.Bytes()
		if len(line) > 80 {
			return &ParseError{LineNo: lineno, Err: errTooLong}
		}
		// Ignore comments if they span the whole line.
		// TODO: verify exact firmware behavior.
		if bytes.IndexRune(line, '#') == 0 {
			continue
		}
		// Strip _trailing_ spaces
		line = bytes.TrimRight(line, "\n\r\t ")
		// Skip empty lines
		if len(line) == 0 {
			continue
		}

		if bytes.HasPrefix(line, []byte(`[`)) && bytes.HasSuffix(line, []byte(`]`)) {
			if sect.Location.LineNo == 0 {
				seqNum++

				sect.Location.LineNo = lineno
				sect.Location.SeqNum = seqNum
			}

			if sect.Filter != "" {
				seqNum++

				newCfg.Sections = append(newCfg.Sections, sect)
				sect = Section{Location: Location{FileName: fileName, LineNo: lineno, SeqNum: seqNum}}
			}

			sect.Filter = string(line)
		} else {
			var param Parameter
			if err := param.UnmarshalText(line); err != nil {
				return &ParseError{LineNo: lineno, Err: err}
			}

			seqNum++

			param.Location = Location{FileName: fileName, LineNo: lineno, SeqNum: seqNum}
			sect.Parameters = append(sect.Parameters, param)
		}
	}

	if err := scanner.Err(); err != nil {
		return err
	}

	if len(sect.Parameters) > 0 || sect.Filter != "" {
		newCfg.Sections = append(newCfg.Sections, sect)
	}

	*cfg = newCfg

	return nil
}

// ParseError describes a parser error.
type ParseError struct {
	LineNo int
	Err    error
}

// Error implements error.
func (pe *ParseError) Error() string {
	return fmt.Sprintf("line %d: %s", pe.LineNo, pe.Err)
}
