// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package picfg_test

import (
	"bytes"
	"fmt"
	"io/ioutil"
	"os"
	"path/filepath"

	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/picfg"
)

type decoderSuite struct{}

var _ = Suite(&decoderSuite{})

func (s *decoderSuite) TestDecode(c *C) {
	var cfg picfg.ConfigTxt

	// comments are ignored
	dec := picfg.NewDecoder(bytes.NewBufferString("# comment"))
	c.Assert(dec.Decode(&cfg), IsNil)
	c.Check(cfg, DeepEquals, picfg.ConfigTxt{})

	// lines with only white-space are ignored
	dec = picfg.NewDecoder(bytes.NewBufferString("  \t \n"))
	c.Assert(dec.Decode(&cfg), IsNil)
	c.Check(cfg, DeepEquals, picfg.ConfigTxt{})

	// more complex example
	buf := bytes.NewBufferString(`
# comments are ignored and start with #

# empty lines are ignored as well

# sections are places in [square brackets]
[all]

# everything else is a key=value pair
key=value
other=stuff
goes=here
`)
	dec = picfg.NewDecoder(buf)

	c.Assert(dec.Decode(&cfg), IsNil)
	c.Check(cfg, DeepEquals, picfg.ConfigTxt{
		Sections: []picfg.Section{
			{
				Filter: "[all]",
				Parameters: []picfg.Parameter{
					{Name: "key", Value: "value", Location: picfg.Location{LineNo: 10, SeqNum: 2}},
					{Name: "other", Value: "stuff", Location: picfg.Location{LineNo: 11, SeqNum: 3}},
					{Name: "goes", Value: "here", Location: picfg.Location{LineNo: 12, SeqNum: 4}},
				},
				Location: picfg.Location{LineNo: 7, SeqNum: 1},
			},
		},
	})
}

func (s *decoderSuite) TestLocationString(c *C) {
	loc := picfg.Location{FileName: "config.txt", LineNo: 5}
	c.Check(loc.String(), Equals, "config.txt:5")

	// Nothing fancy happens for an empty location.
	loc = picfg.Location{}
	c.Check(loc.String(), Equals, ":0")
}

func (s *decoderSuite) TestDecoderRemembersFileName(c *C) {
	d := c.MkDir()
	name := filepath.Join(d, "config.txt")
	c.Assert(ioutil.WriteFile(name, []byte("kernel=linux.img"), 0o600), IsNil)

	f, err := os.Open(name)
	c.Assert(err, IsNil)

	defer func() {
		_ = f.Close()
	}()

	dec := picfg.NewDecoder(f)

	var cfg picfg.ConfigTxt

	c.Assert(dec.Decode(&cfg), IsNil)

	c.Check(cfg.Sections, HasLen, 1)
	c.Check(cfg.Sections[0].Parameters, HasLen, 1)
	c.Check(cfg.Sections[0].Parameters[0].Location.FileName, Equals, name)
}

func (s *decoderSuite) TestDecodeErrors(c *C) {
	var cfg picfg.ConfigTxt

	dec := picfg.NewDecoder(bytes.NewBufferString("key"))
	c.Assert(dec.Decode(&cfg), ErrorMatches, `line 1: cannot parse parameter out of "key"`)

	dec = picfg.NewDecoder(bytes.NewBufferString(fmt.Sprintf("key=%80s", "value")))
	c.Assert(dec.Decode(&cfg), ErrorMatches, `line 1: cannot parse line longer than 80 bytes`)

	dec = picfg.NewDecoder(bytes.NewBufferString("[all"))
	c.Assert(dec.Decode(&cfg), ErrorMatches, `line 1: cannot parse parameter out of "\[all"`)

	dec = picfg.NewDecoder(bytes.NewBufferString("\nkey"))
	err := dec.Decode(&cfg)
	c.Assert(err, ErrorMatches, `line 2: cannot parse parameter out of "key"`)
	c.Assert(err, DeepEquals, &picfg.ParseError{LineNo: 2, Err: &picfg.ParameterSyntaxError{Text: []byte(`key`)}})
}
