// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

// Package picfg models configuration files used by the Raspberry Pi bootloader
//
// Implementation is based on the official documentation
// https://www.raspberrypi.org/documentation/configuration/config-txt/
//
// A tl;dr; variant of the documentation is:
//      INI-style [section] key=value file with one exception when the key is
//      "initramfs" then the value separator is " " instead of "=". Sections
//      are stateful but this does not affect parsing.
package picfg

import (
	"bytes"
	"fmt"
)

// ConfigTxt represents the contents of a config.txt file.
//
// The file is represented as a slice of sections. Each section starts with an
// optional conditional filter and is followed by a slice of parameters.
//
// The representation does not support perfect round-trips and is intended for
// working with machine generated configuration files. Order of sections,
// conditional filters and properties is preserved, comments are lost.
type ConfigTxt struct {
	Sections []Section
}

// UnmarshalText implements TextUnmarshaler.
func (cfg *ConfigTxt) UnmarshalText(text []byte) error {
	return NewDecoder(bytes.NewBuffer(text)).Decode(cfg)
}

// MarshalText implements TextMarshaler.
func (cfg ConfigTxt) MarshalText() ([]byte, error) {
	var buf bytes.Buffer
	if err := NewEncoder(&buf).Encode(cfg); err != nil {
		return nil, err
	}

	return buf.Bytes(), nil
}

// Location pairs file name with a line number.
//
// The decoder provides location for each section and parameter.
// The encoder ignores those values.
type Location struct {
	FileName string
	LineNo   int

	// SeqNum is a sequence number, greater than zero and monotonically
	// increasing as the decoder reads files. Unlike LineNo, when an include
	// directive is read, the sequence number keeps increasing.
	//
	// Sequence number can be used to determine which of a pair of
	// parameters was read earlier.
	SeqNum uint
}

// String returns the location as a file-name:line-number string.
func (loc *Location) String() string {
	return fmt.Sprintf("%s:%d", loc.FileName, loc.LineNo)
}

// Section scopes parameters to an optional conditional filter.
//
// The way a conditional filter impacts boot selection is stateful and depends
// on the filters in prior sections. To understand the effective set of filters
// that apply, use the StateTracker from the condfilter package. The parser
// does not offer this functionality.
type Section struct {
	// Filter is either empty or matches /\[^]+\]/.
	Filter     string
	Parameters []Parameter
	Location   Location
}

// Parameter represents a single configurable entry.
//
// Most parameters are represented as a key=value pair. Some keys can be assigned
// to multiple times for a stacking effect. This is not represented in the
// parser as the inner semantics of the firmware is not precisely documented.
//
// As an exception, the "initramfs" key is not using "key=value" syntax and
// instead is using "key value", with a space instead of the equals sign.
type Parameter struct {
	Name     string
	Value    string
	Location Location
}

// MarshalText implements TextMarshaler.
func (p Parameter) MarshalText() ([]byte, error) {
	return []byte(p.String()), nil
}

// UnmarshalText implements TextUnmarshaler.
func (p *Parameter) UnmarshalText(text []byte) error {
	var pair [][]byte
	if bytes.HasPrefix(text, []byte(`initramfs `)) {
		pair = bytes.SplitN(text, []byte(" "), 2)
	} else {
		pair = bytes.SplitN(text, []byte("="), 2)
	}

	if len(pair) != 2 {
		return &ParameterSyntaxError{Text: text}
	}

	p.Name = string(pair[0])
	p.Value = string(pair[1])

	return nil
}

func (p Parameter) String() string {
	return fmt.Sprintf("%s%c%s", p.Name, p.glue(), p.Value)
}

// glue returns '=' for all parameters except for `initramfs`, where it returns ' '.
func (p *Parameter) glue() rune {
	if p.Name == "initramfs" {
		return ' '
	}

	return '='
}

// ParameterSyntaxError reports inability to parse a key-value parameter.
type ParameterSyntaxError struct {
	Text []byte
}

// Error implements the error interface.
func (err *ParameterSyntaxError) Error() string {
	return fmt.Sprintf("cannot parse parameter out of %q", string(err.Text))
}
