// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package pimodel_test

import (
	"io/ioutil"
	"path/filepath"

	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/picfg/pimodel"
)

type revCodeSuite struct{}

var _ = Suite(&revCodeSuite{})

func (s *revCodeSuite) TestString(c *C) {
	// Both old and new style revisions are represented the way people might expect.
	c.Check(pimodel.RevisionCode(0x0004).String(), Equals, "0004")
	c.Check(pimodel.RevisionCode(0xa03140).String(), Equals, "a03140")
}

func (s *revCodeSuite) TestMarshalText(c *C) {
	revCode := pimodel.RevisionCode(0x0004)
	data, err := revCode.MarshalText()
	c.Assert(err, IsNil)
	c.Check(data, DeepEquals, []byte("0004"))

	revCode = pimodel.RevisionCode(0xa03140)
	data, err = revCode.MarshalText()
	c.Assert(err, IsNil)
	c.Check(data, DeepEquals, []byte("a03140"))
}

func (s *revCodeSuite) TestUnmarshalText(c *C) {
	var revCode pimodel.RevisionCode

	err := revCode.UnmarshalText([]byte("0004"))
	c.Assert(err, IsNil)
	c.Check(revCode, Equals, pimodel.RevisionCode(0x0004))

	err = revCode.UnmarshalText([]byte("potato"))
	c.Assert(err, ErrorMatches, `strconv.ParseUint: parsing "potato": invalid syntax`)
}

func (s *revCodeSuite) TestKnownRevisionCodes(c *C) {
	for _, t := range []struct {
		code         pimodel.RevisionCode
		model        pimodel.Model
		revision     string
		ram          uint
		manufacturer string
		soc          pimodel.SoC
	}{
		// Data from the "New-style revision codes in use" table on
		// https://www.raspberrypi.org/documentation/hardware/raspberrypi/revision-codes/README.md

		// Old style codes.

		{
			code:         0x0002,
			model:        pimodel.Pi1B,
			revision:     "1.0",
			ram:          256,
			manufacturer: "Egoman",
			soc:          pimodel.BCM2835,
		},
		{
			code:         0x0003, /* same as 0x0002? */
			model:        pimodel.Pi1B,
			revision:     "1.0",
			ram:          256,
			manufacturer: "Egoman",
		},
		{
			code:         0x0004,
			model:        pimodel.Pi1B,
			revision:     "2.0",
			ram:          256,
			manufacturer: "Sony UK",
		},
		{
			code:         0x0005,
			model:        pimodel.Pi1B,
			revision:     "2.0",
			ram:          256,
			manufacturer: "Qisda",
		},
		{
			code:         0x0006,
			model:        pimodel.Pi1B,
			revision:     "2.0",
			ram:          256,
			manufacturer: "Egoman",
		},
		{
			code:         0x0007,
			model:        pimodel.Pi1A,
			revision:     "2.0",
			ram:          256,
			manufacturer: "Egoman",
		},
		{
			code:         0x0008,
			model:        pimodel.Pi1A,
			revision:     "2.0",
			ram:          256,
			manufacturer: "Sony UK",
		},
		{
			code:         0x0009,
			model:        pimodel.Pi1A,
			revision:     "2.0",
			ram:          256,
			manufacturer: "Qisda",
		},
		{
			code:         0x000d,
			model:        pimodel.Pi1B,
			revision:     "2.0",
			ram:          512,
			manufacturer: "Egoman",
		},
		{
			code:         0x000e,
			model:        pimodel.Pi1B,
			revision:     "2.0",
			ram:          512,
			manufacturer: "Sony UK",
		},
		{
			code:         0x000f,
			model:        pimodel.Pi1B,
			revision:     "2.0",
			ram:          512,
			manufacturer: "Egoman",
		},
		{
			code:         0x0010,
			model:        pimodel.Pi1BPlus,
			revision:     "1.2",
			ram:          512,
			manufacturer: "Sony UK",
		},
		{
			code:         0x0011,
			model:        pimodel.Cm1,
			revision:     "1.0",
			ram:          512,
			manufacturer: "Sony UK",
		},
		{
			code:         0x0012,
			model:        pimodel.Pi1APlus,
			revision:     "1.1",
			ram:          256,
			manufacturer: "Sony UK",
		},
		{
			code:         0x0013,
			model:        pimodel.Pi1BPlus,
			revision:     "1.2",
			ram:          512,
			manufacturer: "Embest",
		},
		{
			code:         0x0014,
			model:        pimodel.Cm1,
			revision:     "1.0",
			ram:          512,
			manufacturer: "Embest",
		},
		{
			code:         0x0015,
			model:        pimodel.Pi1APlus,
			revision:     "1.1",
			ram:          256,
			manufacturer: "Embest",
		},

		// New style codes.
		{
			code:         0x900021,
			model:        pimodel.Pi1APlus,
			revision:     "1.1",
			ram:          512,
			manufacturer: "Sony UK",
		},
		{
			code:         0x900032,
			model:        pimodel.Pi1BPlus,
			revision:     "1.2",
			ram:          512,
			manufacturer: "Sony UK",
		},
		{
			code:         0x900092,
			model:        pimodel.Pi0,
			revision:     "1.2",
			ram:          512,
			manufacturer: "Sony UK",
		},
		{
			code:         0x900092,
			model:        pimodel.Pi0,
			revision:     "1.2",
			ram:          512,
			manufacturer: "Sony UK",
		},
		{
			code:         0x900093,
			model:        pimodel.Pi0,
			revision:     "1.3",
			ram:          512,
			manufacturer: "Sony UK",
		},
		{
			code:         0x9000c1,
			model:        pimodel.Pi0W,
			revision:     "1.1",
			ram:          512,
			manufacturer: "Sony UK",
		},
		{
			code:         0x9020e0,
			model:        pimodel.Pi3APlus,
			revision:     "1.0",
			ram:          512,
			manufacturer: "Sony UK",
		},
		{
			code:         0x920092,
			model:        pimodel.Pi0,
			revision:     "1.2",
			ram:          512,
			manufacturer: "Embest",
		},
		{
			code:         0x920093,
			model:        pimodel.Pi0,
			revision:     "1.3",
			ram:          512,
			manufacturer: "Embest",
		},
		{
			code:         0x900061,
			model:        pimodel.Cm1,
			revision:     "1.1",
			ram:          512,
			manufacturer: "Sony UK",
		},
		{
			code:         0xa01040,
			model:        pimodel.Pi2B,
			revision:     "1.0",
			ram:          1024,
			manufacturer: "Sony UK",
		},
		{
			code:         0xa01041,
			model:        pimodel.Pi2B,
			revision:     "1.1",
			ram:          1024,
			manufacturer: "Sony UK",
		},
		{
			code:         0xa02082,
			model:        pimodel.Pi3B,
			revision:     "1.2",
			ram:          1024,
			manufacturer: "Sony UK",
		},
		{
			code:         0xa020a0,
			model:        pimodel.Cm3,
			revision:     "1.0",
			ram:          1024,
			manufacturer: "Sony UK",
		},
		{
			code:         0xa020d3,
			model:        pimodel.Pi3BPlus,
			revision:     "1.3",
			ram:          1024,
			manufacturer: "Sony UK",
		},
		{
			code:         0xa02042,
			model:        pimodel.Pi2B,
			revision:     "1.2",
			ram:          1024,
			manufacturer: "Sony UK",
			soc:          pimodel.BCM2837,
		},
		{
			code:         0xa21041,
			model:        pimodel.Pi2B,
			revision:     "1.1",
			ram:          1024,
			manufacturer: "Embest",
		},
		{
			code:         0xa22042,
			model:        pimodel.Pi2B,
			revision:     "1.2",
			ram:          1024,
			manufacturer: "Embest",
			soc:          pimodel.BCM2837,
		},
		{
			code:         0xa22082,
			model:        pimodel.Pi3B,
			revision:     "1.2",
			ram:          1024,
			manufacturer: "Embest",
		},
		{
			code:         0xa220a0,
			model:        pimodel.Cm3,
			revision:     "1.0",
			ram:          1024,
			manufacturer: "Embest",
		},
		{
			code:         0xa32082,
			model:        pimodel.Pi3B,
			revision:     "1.2",
			ram:          1024,
			manufacturer: "Sony Japan",
		},
		{
			code:         0xa52082,
			model:        pimodel.Pi3B,
			revision:     "1.2",
			ram:          1024,
			manufacturer: "Stadium",
		},
		{
			code:         0xa22083,
			model:        pimodel.Pi3B,
			revision:     "1.3",
			ram:          1024,
			manufacturer: "Embest",
		},
		{
			code:         0xa02100,
			model:        pimodel.Cm3Plus,
			revision:     "1.0",
			ram:          1024,
			manufacturer: "Sony UK",

			// All the 3+ models have the B0 stepping.
			soc: pimodel.BCM2837B0,
		},
		{
			code:         0xa03111,
			model:        pimodel.Pi4B,
			revision:     "1.1",
			ram:          1024,
			manufacturer: "Sony UK",
		},
		{
			code:         0xb03111,
			model:        pimodel.Pi4B,
			revision:     "1.1",
			ram:          2048,
			manufacturer: "Sony UK",
		},
		{
			code:         0xb03112,
			model:        pimodel.Pi4B,
			revision:     "1.2",
			ram:          2048,
			manufacturer: "Sony UK",
		},
		{
			code:         0xb03114,
			model:        pimodel.Pi4B,
			revision:     "1.4",
			ram:          2048,
			manufacturer: "Sony UK",
		},
		{
			code:         0xc03111,
			model:        pimodel.Pi4B,
			revision:     "1.1",
			ram:          4096,
			manufacturer: "Sony UK",
		},
		{
			code:         0xc03112,
			model:        pimodel.Pi4B,
			revision:     "1.2",
			ram:          4096,
			manufacturer: "Sony UK",
		},
		{
			code:         0xc03114,
			model:        pimodel.Pi4B,
			revision:     "1.4",
			ram:          4096,
			manufacturer: "Sony UK",
		},
		{
			code:         0xd03114,
			model:        pimodel.Pi4B,
			revision:     "1.4",
			ram:          8192,
			manufacturer: "Sony UK",
		},
		{
			code:         0xc03130,
			model:        pimodel.Pi400,
			revision:     "1.0",
			ram:          4096,
			manufacturer: "Sony UK",
		},
		{
			code:         0xa03140,
			model:        pimodel.Cm4,
			revision:     "1.0",
			ram:          1024,
			manufacturer: "Sony UK",
		},
		{
			code:         0xb03140,
			model:        pimodel.Cm4,
			revision:     "1.0",
			ram:          2048,
			manufacturer: "Sony UK",
		},
		{
			code:         0xc03140,
			model:        pimodel.Cm4,
			revision:     "1.0",
			ram:          4096,
			manufacturer: "Sony UK",
		},
		{
			code:         0xd03140,
			model:        pimodel.Cm4,
			revision:     "1.0",
			ram:          8192,
			manufacturer: "Sony UK",
		},
	} {
		comment := Commentf("revision code: %s", t.code)

		model, err := t.code.BoardType()
		c.Assert(err, IsNil, comment)
		c.Check(model, Equals, t.model, comment)

		revision, err := t.code.BoardRevision()
		c.Assert(err, IsNil, comment)
		c.Check(revision, Equals, t.revision, comment)

		ram, err := t.code.MemorySize()
		c.Assert(err, IsNil, comment)
		c.Check(ram, Equals, t.ram, comment)

		manufacturer, err := t.code.Manufacturer()
		c.Assert(err, IsNil, comment)
		c.Check(manufacturer, Equals, t.manufacturer, comment)

		soc, err := t.code.Processor()
		c.Assert(err, IsNil, comment)
		// Check the SoC type if one is given in the test data.
		if t.soc != pimodel.InvalidSoC {
			c.Check(soc, Equals, t.soc, comment)
		}
	}
}

func (s *revCodeSuite) TestWonkyMemorySize(c *C) {
	_, err := pimodel.RevisionCode(0x0).MemorySize()
	c.Assert(err, ErrorMatches, `cannot decode revision code 0000, unknown memory size`)
}

func (s *revCodeSuite) TestWonkyManufacturer(c *C) {
	_, err := pimodel.RevisionCode(0x0).Manufacturer()
	c.Assert(err, ErrorMatches, `cannot decode revision code 0000, unknown manufacturer`)
}

func (s *revCodeSuite) TestWonkyProcessor(c *C) {
	_, err := pimodel.RevisionCode((5 << 12) | (1 << 23)).Processor() // 5th SoC | new flag
	c.Assert(err, ErrorMatches, `cannot decode revision code 805000, unknown processor`)
}

func (s *revCodeSuite) TestWonkyRevision(c *C) {
	_, err := pimodel.RevisionCode(0x0001).BoardRevision()
	c.Assert(err, ErrorMatches, `cannot decode revision code 0001, unknown board revision`)
}

func (s *revCodeSuite) TestWonkyBoardType(c *C) {
	_, err := pimodel.RevisionCode(0x0001).BoardType()
	c.Assert(err, ErrorMatches, `cannot decode revision code 0001, unknown board type`)
}

type probeCpuinfoSuite struct{}

var _ = Suite(&probeCpuinfoSuite{})

func (s *probeCpuinfoSuite) TestProcCpuInfoFile(c *C) {
	c.Assert(pimodel.ProcCpuInfoFile, Equals, "/proc/cpuinfo")
}

const piZeroCpuInfo = `processor	: 0
model name	: ARMv6-compatible processor rev 7 (v6l)
BogoMIPS	: 697.95
Features	: half thumb fastmult vfp edsp java tls
CPU implementer	: 0x41
CPU architecture: 7
CPU variant	: 0x0
CPU part	: 0xb76
CPU revision	: 7

Hardware	: BCM2835
Revision	: 9000c1
Serial		: 00000000a3896d54
Model		: Raspberry Pi Zero W Rev 1.1
`

const Pi1BCpuInfo = `processor	: 0
model name	: ARMv6-compatible processor rev 7 (v6l)
BogoMIPS	: 697.95
Features	: half thumb fastmult vfp edsp java tls
CPU implementer	: 0x41
CPU architecture: 7
CPU variant	: 0x0
CPU part	: 0xb76
CPU revision	: 7

Hardware	: BCM2835
Revision	: 000f
Serial		: 0000000054281887
Model		: Raspberry Pi Model B Rev 2
`

func (s *probeCpuinfoSuite) TestProbeCpuInfo(c *C) {
	d := c.MkDir()
	name := filepath.Join(d, "cpuinfo")

	_, _, err := pimodel.ProbeCpuInfo(name)
	c.Assert(err, ErrorMatches, `cannot detect Raspberry Pi model: open .*/cpuinfo: no such file or directory`)

	c.Assert(ioutil.WriteFile(name, []byte(""), 0o600), IsNil)
	_, _, err = pimodel.ProbeCpuInfo(name)
	c.Assert(err, ErrorMatches, `cannot detect Raspberry Pi model: not a Raspberry Pi`)

	c.Assert(ioutil.WriteFile(name, []byte("Revision\t: potato"), 0o600), IsNil)
	_, _, err = pimodel.ProbeCpuInfo(name)
	c.Assert(err, ErrorMatches, `cannot detect Raspberry Pi model: strconv.ParseUint: parsing "potato": invalid syntax`)

	c.Assert(ioutil.WriteFile(name, []byte("Serial\t: tomato"), 0o600), IsNil)
	_, _, err = pimodel.ProbeCpuInfo(name)
	c.Assert(err, ErrorMatches, `cannot detect Raspberry Pi model: strconv.ParseUint: parsing "tomato": invalid syntax`)

	c.Assert(ioutil.WriteFile(name, []byte(piZeroCpuInfo), 0o600), IsNil)
	code, serial, err := pimodel.ProbeCpuInfo(name)
	c.Assert(err, IsNil)
	c.Check(code, Equals, pimodel.RevisionCode(0x9000c1))
	c.Check(serial, Equals, pimodel.SerialNumber(0x00000000a3896d54))

	c.Assert(ioutil.WriteFile(name, []byte(Pi1BCpuInfo), 0o600), IsNil)
	code, serial, err = pimodel.ProbeCpuInfo(name)
	c.Assert(err, IsNil)
	c.Check(code, Equals, pimodel.RevisionCode(0x000f))
	c.Check(serial, Equals, pimodel.SerialNumber(0x0000000054281887))
}

type serialNumSuite struct{}

var _ = Suite(&serialNumSuite{})

func (s *serialNumSuite) TestString(c *C) {
	num := pimodel.SerialNumber(0x00000000a3896d54)
	c.Check(num.String(), Equals, "00000000a3896d54")
}

func (s *serialNumSuite) TestMarshalText(c *C) {
	num := pimodel.SerialNumber(0x00000000a3896d54)
	data, err := num.MarshalText()
	c.Assert(err, IsNil)
	c.Check(data, DeepEquals, []byte("00000000a3896d54"))
}

func (s *serialNumSuite) TestUnmarshalText(c *C) {
	var num pimodel.SerialNumber

	err := num.UnmarshalText([]byte("00000000a3896d54"))
	c.Assert(err, IsNil)
	c.Check(num, Equals, pimodel.SerialNumber(0x00000000a3896d54))

	err = num.UnmarshalText([]byte("potato"))
	c.Assert(err, ErrorMatches, `strconv.ParseUint: parsing "potato": invalid syntax`)
}
