// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package boothandler

import (
	"errors"
	"fmt"
	"io"
	"os"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/boot"
)

// Handler is an interface between boot loader and RAUC.
//
// Custom boot backend is documented in detail at
// https://rauc.readthedocs.io/en/latest/integration.html#custom
//
// PrimarySlot and SetPrimarySlot respectively return and set the slot used for
// booting. SlotState and SetSlotState respectively return and set the boot
// state of a given slot.
type Handler interface {
	PrimarySlot() (boot.Slot, error)
	SetPrimarySlot(boot.Slot) error

	SlotState(boot.Slot) (boot.SlotState, error)
	SetSlotState(boot.Slot, boot.SlotState) error
}

const (
	// getPrimaryCmd is the sub-command invoked by RAUC to get the name of the primary slot.
	getPrimaryCmd = "get-primary"
	// setPrimaryCmd is the sub-command invoked by RAUC to set the name of the primary slot.
	setPrimaryCmd = "set-primary"
	// getBootStateCmd is the sub-command invoked by RAUC to get the state of a given slot.
	getBootStateCmd = "get-state"
	// getBootStateCmd is the sub-command invoked by RAUC to set the state of a given slot.
	setBootStateCmd = "set-state"
)

// Command runs the specific command of a custom boot handler.
//
// This function implements the protocol between RAUC and a custom, third party
// boot loader handler. The get-* commands print the retrieved value to stdout.
//
// See the description of CustomBootBackend interface for details.
func Command(handler Handler, cmd string, args ...string) (err error) {
	switch cmd {
	case getPrimaryCmd:
		return runGetPrimarySlotCmd(handler, args)
	case setPrimaryCmd:
		return runSetPrimarySlotCmd(handler, args)
	case getBootStateCmd:
		return runGetSlotStateCmd(handler, args)
	case setBootStateCmd:
		return runSetSlotStateCmd(handler, args)
	default:
		return ErrUnknownCmd
	}
}

var (
	// ErrExpectedPrimarySlot records unexpected number of arguments to the "set-primary" command.
	ErrExpectedPrimarySlot = errors.New("expected name of the primary slot")
	// ErrExpectedSlotName records unexpected number of arguments to the "get-state" or "set-state" command.
	ErrExpectedSlotName = errors.New("expected slot name")
	// ErrExpectedSlotState records lack of slot state name argument in "set-state" command.
	ErrExpectedSlotState = errors.New("expected slot state")
	// ErrUnexpectedArgument records extraneous unexpected argument to any of the comman.
	ErrUnexpectedArgument = errors.New("unexpected argument")
	// ErrUnknownCmd records usage of unknown command.
	ErrUnknownCmd = errors.New("unknown command")
)

var (
	// stdout is os.Stdout but can changed for testing
	stdout io.Writer = os.Stdout
	// stderr is os.Stderr but can changed for testing
	stderr io.Writer = os.Stderr
)

func runGetPrimarySlotCmd(handler Handler, args []string) error {
	switch len(args) {
	case 0:
		slot, err := handler.PrimarySlot()
		if err != nil {
			return err
		}

		_, err = fmt.Fprintln(stdout, slot)

		return err
	default:
		return ErrUnexpectedArgument
	}
}

func runSetPrimarySlotCmd(handler Handler, args []string) error {
	var slot boot.Slot

	switch len(args) {
	case 0:
		return ErrExpectedSlotName
	case 1:
		if err := slot.UnmarshalText([]byte(args[0])); err != nil {
			return err
		}

		return handler.SetPrimarySlot(slot)
	default:
		return ErrUnexpectedArgument
	}
}

func runGetSlotStateCmd(handler Handler, args []string) error {
	var slot boot.Slot

	switch len(args) {
	case 0:
		return ErrExpectedSlotName
	case 1:
		if err := slot.UnmarshalText([]byte(args[0])); err != nil {
			return err
		}

		state, err := handler.SlotState(slot)
		if err != nil {
			return err
		}

		_, err = fmt.Fprintln(stdout, state)

		return err
	default:
		return ErrUnexpectedArgument
	}
}

func runSetSlotStateCmd(handler Handler, args []string) error {
	var slot boot.Slot

	var state boot.SlotState

	switch len(args) {
	case 0:
		return ErrExpectedSlotName
	case 1:
		return ErrExpectedSlotState
	case 2:
		if err := slot.UnmarshalText([]byte(args[0])); err != nil {
			return err
		}

		if err := state.UnmarshalText([]byte(args[1])); err != nil {
			return err
		}

		return handler.SetSlotState(slot, state)
	default:
		return ErrUnexpectedArgument
	}
}
