// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package boothandler_test

import (
	"errors"
	"testing"

	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/boot"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/rauc/boothandler"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/rauc/rauctest"
)

func Test(t *testing.T) { TestingT(t) }

var errBoom = errors.New("boom")

type suiteBase struct {
	backend      rauctest.BootHandler
	stdio        *boothandler.TestStdio
	restoreStdio func()
}

func (s *suiteBase) SetUpTest(c *C) {
	s.backend.ResetCallbacks()
	s.stdio, s.restoreStdio = boothandler.MockStdio()
}

func (s *suiteBase) TearDownTest(c *C) {
	s.restoreStdio()
}

type dispatchSuite struct {
	suiteBase
}

var _ = Suite(&dispatchSuite{})

func (s *dispatchSuite) TestUnknownCommand(c *C) {
	err := boothandler.Command(&s.backend, "potato")
	c.Assert(err, ErrorMatches, `unknown command`)
}

// get-primary tests

func (s *dispatchSuite) TestGetPrimarySuccess(c *C) {
	s.backend.PrimarySlotFn = func() (boot.Slot, error) { return boot.SlotA, nil }
	err := boothandler.Command(&s.backend, "get-primary")
	c.Assert(err, IsNil)
	c.Check(s.stdio.Stdout.String(), Equals, "A\n")
	c.Check(s.stdio.Stderr.String(), Equals, "")
}

func (s *dispatchSuite) TestGetPrimaryFailure(c *C) {
	s.backend.PrimarySlotFn = func() (boot.Slot, error) { return boot.InvalidSlot, errBoom }
	err := boothandler.Command(&s.backend, "get-primary")
	c.Assert(err, ErrorMatches, `boom`)
}

func (s *dispatchSuite) TestGetPrimaryExtraArgs(c *C) {
	err := boothandler.Command(&s.backend, "get-primary", "foo")
	c.Assert(err, ErrorMatches, `unexpected argument`)
}

// set-primary tests

func (s *dispatchSuite) TestSetPrimarySuccess(c *C) {
	called := false
	s.backend.SetPrimarySlotFn = func(slot boot.Slot) error {
		called = true

		c.Check(slot, Equals, boot.SlotA)

		return nil
	}

	err := boothandler.Command(&s.backend, "set-primary", "A")
	c.Assert(err, IsNil)
	c.Check(called, Equals, true)
}

func (s *dispatchSuite) TestSetPrimaryFailure(c *C) {
	s.backend.SetPrimarySlotFn = func(slot boot.Slot) error {
		return errBoom
	}
	err := boothandler.Command(&s.backend, "set-primary", "A")
	c.Assert(err, ErrorMatches, `boom`)
}

func (s *dispatchSuite) TestSetPrimaryInvalidSlotName(c *C) {
	err := boothandler.Command(&s.backend, "set-primary", "potato")
	c.Assert(err, ErrorMatches, `invalid boot slot`)
}

func (s *dispatchSuite) TestSetPrimaryMissingArgs(c *C) {
	err := boothandler.Command(&s.backend, "set-primary")
	c.Assert(err, ErrorMatches, `expected slot name`)
}

func (s *dispatchSuite) TestSetPrimaryExtraArgs(c *C) {
	err := boothandler.Command(&s.backend, "set-primary", "A", "foo")
	c.Assert(err, ErrorMatches, `unexpected argument`)
}

// get-state tests

func (s *dispatchSuite) TestGetStateSuccess(c *C) {
	called := false
	s.backend.SlotStateFn = func(slot boot.Slot) (boot.SlotState, error) {
		called = true

		c.Assert(slot, Equals, boot.SlotA)

		return boot.GoodSlot, nil
	}
	err := boothandler.Command(&s.backend, "get-state", "A")
	c.Assert(err, IsNil)
	c.Check(called, Equals, true)
	c.Check(s.stdio.Stdout.String(), Equals, "good\n")
	c.Check(s.stdio.Stderr.String(), Equals, "")
}

func (s *dispatchSuite) TestGetStateFailure(c *C) {
	s.backend.SlotStateFn = func(slot boot.Slot) (boot.SlotState, error) {
		c.Assert(slot, Equals, boot.SlotA)
		return boot.InvalidSlotState, errBoom
	}
	err := boothandler.Command(&s.backend, "get-state", "A")
	c.Assert(err, ErrorMatches, `boom`)
}

func (s *dispatchSuite) TestGetStateInvalidState(c *C) {
	s.backend.SlotStateFn = func(slot boot.Slot) (boot.SlotState, error) {
		c.Assert(slot, Equals, boot.SlotB)
		return boot.InvalidSlotState, nil
	}
	err := boothandler.Command(&s.backend, "get-state", "B")
	c.Assert(err, IsNil)
	// XXX(zyga): this is probably a violation of the protocol
	c.Check(s.stdio.Stdout.String(), Equals, "invalid\n")
	c.Check(s.stdio.Stderr.String(), Equals, "")
}

func (s *dispatchSuite) TestGetStateMissingArgs(c *C) {
	err := boothandler.Command(&s.backend, "get-state")
	c.Assert(err, ErrorMatches, `expected slot name`)
}

func (s *dispatchSuite) TestGetStateInvalidSlotName(c *C) {
	err := boothandler.Command(&s.backend, "get-state", "potato")
	c.Assert(err, ErrorMatches, `invalid boot slot`)
}

func (s *dispatchSuite) TestGetStateExtraArgs(c *C) {
	err := boothandler.Command(&s.backend, "get-state", "A", "foo")
	c.Assert(err, ErrorMatches, `unexpected argument`)
}

// set-state tests

func (s *dispatchSuite) TestSetStateSuccess(c *C) {
	called := false
	s.backend.SetSlotStateFn = func(slot boot.Slot, state boot.SlotState) error {
		called = true

		c.Assert(slot, Equals, boot.SlotA)
		c.Assert(state, Equals, boot.GoodSlot)

		return nil
	}
	err := boothandler.Command(&s.backend, "set-state", "A", "good")
	c.Assert(err, IsNil)
	c.Check(called, Equals, true)
}

func (s *dispatchSuite) TestSetStateFailure(c *C) {
	s.backend.SetSlotStateFn = func(slot boot.Slot, state boot.SlotState) error {
		c.Assert(slot, Equals, boot.SlotA)
		c.Assert(state, Equals, boot.GoodSlot)

		return errBoom
	}
	err := boothandler.Command(&s.backend, "set-state", "A", "good")
	c.Assert(err, ErrorMatches, `boom`)
}

func (s *dispatchSuite) TestSetStateMissingSlotName(c *C) {
	err := boothandler.Command(&s.backend, "set-state")
	c.Assert(err, ErrorMatches, `expected slot name`)
}

func (s *dispatchSuite) TestSetStateMissingSlotState(c *C) {
	err := boothandler.Command(&s.backend, "set-state", "A")
	c.Assert(err, ErrorMatches, `expected slot state`)
}

func (s *dispatchSuite) TestSetStateInvalidSlotName(c *C) {
	err := boothandler.Command(&s.backend, "set-state", "potato", "good")
	c.Assert(err, ErrorMatches, `invalid boot slot`)
}

func (s *dispatchSuite) TestSetStateInvalidSlotState(c *C) {
	err := boothandler.Command(&s.backend, "set-state", "B", "potato")
	c.Assert(err, ErrorMatches, `invalid slot state`)
}

func (s *dispatchSuite) TestSetStateExtraArgs(c *C) {
	err := boothandler.Command(&s.backend, "set-state", "A", "good", "foo")
	c.Assert(err, ErrorMatches, `unexpected argument`)
}
