// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

// Package installhandler contains logic needed to by RAUC install handlers.
//
// RAUC install handlers are helper programs which can execute before and after
// each install operation. RAUC provides information about the operation with
// the help of environment variables.
package installhandler

import (
	"fmt"
	"sort"
	"strings"
)

// Participant is an interface to RAUC install handlers.
//
// The PreInstallHandler method is called before a RAUC bundle is fully
// installed. An error returned from this method blocks the installation process
// from completing.
//
// The PostInstallHandler method is called after a bundle was completely installed.
// Any error it returns is observed by RAUC but has no effect on the process.
type Participant interface {
	PreInstallHandler(ctx *HandlerContext) error
	PostInstallHandler(ctx *HandlerContext) error
}

// HandlerContext is the context provided by RAUC to an install handler.
type HandlerContext struct {
	// MountPrefix corresponds to RAUC_MOUNT_PREFIX.
	MountPrefix string
	// BundleMountPoint corresponds to RAUC_BUNDLE_MOUNT_POINT.
	BundleMountPoint string
	// CurrentBootName corresponds to RAUC_CURRENT_BOOTNAME.
	CurrentBootName string
	// UpdateSource corresponds to RAUC_UPDATE_SOURCE.
	UpdateSource string
	// SystemConfig corresponds to RAUC_SYSTEM_CONFIG.
	SystemConfig string
	// TargetSlotIDs corresponds to RAUC_TARGET_SLOTS.
	TargetSlotIDs []string
	// SlotIDs corresponds to RAUC_SLOTS.
	SlotIDs []string
	// Slots correspond to data collected from a number of variables. See the
	// documentation of the Slot type for details.
	Slots map[string]*Slot
	// ImageIDs is synthesized and is not provided by RAUC explicitly.
	ImageIDs []string
	// Images correspond to data collected from a number of variables. See the
	// documentation of the Image type for details.
	Images map[string]*Image
}

// Slot contains information about RAUC slots provided to install hooks.
type Slot struct {
	// ID is the ephemeral ID assigned to this slot by RAUC, for the purpose of
	// conveying information to the install handler. Slot IDs should not be assumed
	// to be stable across invocation of the handler.
	ID string
	// Name corresponds to RAUC_SLOT_NAME_$id, where $id is Slot.ID.
	Name string
	// Class corresponds to RAUC_SLOT_CLASS_$id, where $id is Slot.ID.
	Class string
	// Type corresponds to RAUC_SLOT_TYPE_$id, where $id is Slot.ID.
	Type string
	// Type corresponds to RAUC_SLOT_BOOTNAME_$id, where $id is Slot.ID.
	BootName string
	// Type corresponds to RAUC_SLOT_DEVICE_$id, where $id is Slot.ID.
	DevicePath string
	// ParentID corresponds to RAUC_SLOT_PARENT_$id, where $id is Slot.ID.
	ParentID string
}

// Image contains information about RAUC image provided to install hooks.
type Image struct {
	// ID is the ephemeral ID assigned to this image by RAUC, for the purpose of
	// conveying information to the install handler. Image IDs should not be assumed
	// to be stable across invocation of the handler.
	ID string
	// Name corresponds to RAUC_IMAGE_NAME_$id, where $id is Image.ID.
	Name string
	// Class corresponds to RAUC_IMAGE_CLASS_$id, where $id is Image.ID.
	Class string
	// Class corresponds to RAUC_IMAGE_DIGEST_$id, where $id is Image.ID.
	Digest string
}

// UnknownEnvError reports unknown environment variables in RAUC namespace.
type UnknownEnvError struct {
	Key   string
	Value string
}

// Error implements the error interface.
func (e *UnknownEnvError) Error() string {
	return fmt.Sprintf("unknown environment variable %s=%s", e.Key, e.Value)
}

// SlotReferenceError reports references to undefined slots.
type SlotReferenceError struct {
	SlotID string
	// EnvVar is the environment variable containing the reference.
	EnvVar string
}

// Error implements the error interface.
func (e *SlotReferenceError) Error() string {
	return fmt.Sprintf("variable %s refers to undefined slot %s", e.EnvVar, e.SlotID)
}

// HookContextError reports problems with install handler context.
type HookContextError struct {
	Err error
}

// Error implements the error interface.
func (e *HookContextError) Error() string {
	return fmt.Sprintf("cannot create handler context: %v", e.Err)
}

// Unwrap returns the error wrapped inside the handler context error.
func (e *HookContextError) Unwrap() error {
	return e.Err
}

// NewHandlerContext returns a new HookContext described by the given environment block.
func NewHandlerContext(env []string) (*HandlerContext, error) {
	ctx := HandlerContext{}

	if err := ctx.parseInstallEnv(env); err != nil {
		return nil, &HookContextError{Err: err}
	}

	ctx.synthesizeImageIDs()

	if err := ctx.validateSlotReferences(); err != nil {
		return nil, &HookContextError{Err: err}
	}

	return &ctx, nil
}

// Environ returns an environment block representing the handler context.
//
// A handler context can be converted to and from the environment representation.
func (ctx *HandlerContext) Environ() []string {
	// Initial capacity tuned to avoid realloc. The capacity matches
	// the number of append calls done below. Seven for the fixed fields,
	// then six for each slot and three for each image.
	env := make([]string, 0, 7+len(ctx.Slots)*6+len(ctx.Images)*3)

	const keyValue = "%s=%s"

	env = append(env, fmt.Sprintf(keyValue, raucMountPrefixEnvKey, ctx.MountPrefix))
	env = append(env, fmt.Sprintf(keyValue, raucBundleMountPointEnvKey, ctx.BundleMountPoint))
	env = append(env, fmt.Sprintf(keyValue, raucCurrentBootNameEnvKey, ctx.CurrentBootName))
	env = append(env, fmt.Sprintf(keyValue, raucUpdateSourceEnvKey, ctx.UpdateSource))
	env = append(env, fmt.Sprintf(keyValue, raucSystemConfigEnvKey, ctx.SystemConfig))
	// Note that some versions of RAUC add an extra trailing space here.
	// This behavior is not reproduced.
	env = append(env, fmt.Sprintf(keyValue, raucTargetSlotsEnvKey, strings.Join(ctx.TargetSlotIDs, " ")))
	env = append(env, fmt.Sprintf(keyValue, raucSlotsEnvKey, strings.Join(ctx.SlotIDs, " ")))

	const keyIDValue = "%s%s=%s"

	for id, slot := range ctx.Slots {
		env = append(env, fmt.Sprintf(keyIDValue, raucSlotBootNameEnvKeyPrefix, id, slot.BootName))
		env = append(env, fmt.Sprintf(keyIDValue, raucSlotTypeEnvKeyPrefix, id, slot.Type))
		env = append(env, fmt.Sprintf(keyIDValue, raucSlotClassEnvKeyPrefix, id, slot.Class))
		env = append(env, fmt.Sprintf(keyIDValue, raucSlotDeviceEnvKeyPrefix, id, slot.DevicePath))
		env = append(env, fmt.Sprintf(keyIDValue, raucSlotNameEnvKeyPrefix, id, slot.Name))
		env = append(env, fmt.Sprintf(keyIDValue, raucSlotParentEnvKeyPrefix, id, slot.ParentID))
	}

	for id, image := range ctx.Images {
		env = append(env, fmt.Sprintf(keyIDValue, raucImageNameEnvKeyPrefix, id, image.Name))
		env = append(env, fmt.Sprintf(keyIDValue, raucImageClassEnvKeyPrefix, id, image.Class))
		env = append(env, fmt.Sprintf(keyIDValue, raucImageDigestEnvKeyPrefix, id, image.Digest))
	}

	// Sort for nice determinism.
	sort.Strings(env)

	return env
}

// parseInstallEnv parses RAUC handler context from environment block.
//
// The resulting context is partial and needs to be passed to
// synthesizeImageIDs and validate before being used.
func (ctx *HandlerContext) parseInstallEnv(env []string) error {
	for _, envItem := range env {
		idx := strings.IndexRune(envItem, '=')
		if idx == -1 {
			continue
		}

		envKey, envValue := envItem[:idx], envItem[idx+1:]
		if !strings.HasPrefix(envKey, "RAUC_") {
			continue
		}

		if err := ctx.parseEnvItem(envKey, envValue); err != nil {
			return err
		}
	}

	return nil
}

const (
	// Variables describing the rest of the handler context.

	raucMountPrefixEnvKey      = "RAUC_MOUNT_PREFIX"
	raucBundleMountPointEnvKey = "RAUC_BUNDLE_MOUNT_POINT"
	raucUpdateSourceEnvKey     = "RAUC_UPDATE_SOURCE"
	raucCurrentBootNameEnvKey  = "RAUC_CURRENT_BOOTNAME"
	raucSystemConfigEnvKey     = "RAUC_SYSTEM_CONFIG"
	raucTargetSlotsEnvKey      = "RAUC_TARGET_SLOTS"
	raucSlotsEnvKey            = "RAUC_SLOTS"

	// variables describing RAUC slots.

	raucSlotBootNameEnvKeyPrefix = "RAUC_SLOT_BOOTNAME_"
	raucSlotTypeEnvKeyPrefix     = "RAUC_SLOT_TYPE_"
	raucSlotClassEnvKeyPrefix    = "RAUC_SLOT_CLASS_"
	raucSlotDeviceEnvKeyPrefix   = "RAUC_SLOT_DEVICE_"
	raucSlotNameEnvKeyPrefix     = "RAUC_SLOT_NAME_"
	raucSlotParentEnvKeyPrefix   = "RAUC_SLOT_PARENT_"

	// variables describing RAUC images.

	raucImageNameEnvKeyPrefix   = "RAUC_IMAGE_NAME_"
	raucImageClassEnvKeyPrefix  = "RAUC_IMAGE_CLASS_"
	raucImageDigestEnvKeyPrefix = "RAUC_IMAGE_DIGEST_"
)

func (ctx *HandlerContext) parseEnvItem(envKey, envValue string) error {
	switch envKey {
	case raucMountPrefixEnvKey:
		ctx.MountPrefix = envValue
	case raucBundleMountPointEnvKey:
		ctx.BundleMountPoint = envValue
	case raucUpdateSourceEnvKey:
		ctx.UpdateSource = envValue
	case raucCurrentBootNameEnvKey:
		ctx.CurrentBootName = envValue
	case raucSystemConfigEnvKey:
		ctx.SystemConfig = envValue
	case raucTargetSlotsEnvKey:
		// Trim trailing spaces that are possible around this value in some versions of RAUC.
		envValue = strings.TrimSpace(envValue)
		if envValue != "" {
			IDs := strings.Split(envValue, " ")
			ctx.TargetSlotIDs = IDs
		}
	case raucSlotsEnvKey:
		// Trim trailing spaces that are possible around this value in some versions of RAUC.
		envValue = strings.TrimSpace(envValue)
		if envValue != "" {
			IDs := strings.Split(envValue, " ")
			ctx.SlotIDs = IDs
		}
	default:
		return ctx.parseSlotImageEnvItem(envKey, envValue)
	}

	return nil
}

func (ctx *HandlerContext) parseSlotImageEnvItem(envKey, envValue string) error {
	switch {
	// Slot attributes.
	case strings.HasPrefix(envKey, raucSlotNameEnvKeyPrefix):
		slotID := envKey[len(raucSlotNameEnvKeyPrefix):]
		slot := ctx.ensureSlot(slotID)
		slot.Name = envValue
	case strings.HasPrefix(envKey, raucSlotClassEnvKeyPrefix):
		slotID := envKey[len(raucSlotClassEnvKeyPrefix):]
		slot := ctx.ensureSlot(slotID)
		slot.Class = envValue
	case strings.HasPrefix(envKey, raucSlotTypeEnvKeyPrefix):
		slotID := envKey[len(raucSlotTypeEnvKeyPrefix):]
		slot := ctx.ensureSlot(slotID)
		slot.Type = envValue
	case strings.HasPrefix(envKey, raucSlotBootNameEnvKeyPrefix):
		slotID := envKey[len(raucSlotBootNameEnvKeyPrefix):]
		slot := ctx.ensureSlot(slotID)
		slot.BootName = envValue
	case strings.HasPrefix(envKey, raucSlotDeviceEnvKeyPrefix):
		slotID := envKey[len(raucSlotDeviceEnvKeyPrefix):]
		slot := ctx.ensureSlot(slotID)
		slot.DevicePath = envValue
	case strings.HasPrefix(envKey, raucSlotParentEnvKeyPrefix):
		slotID := envKey[len(raucSlotParentEnvKeyPrefix):]
		slot := ctx.ensureSlot(slotID)
		slot.ParentID = envValue
		// Image attributes.
	case strings.HasPrefix(envKey, raucImageNameEnvKeyPrefix):
		imageID := envKey[len(raucImageNameEnvKeyPrefix):]
		image := ctx.ensureImage(imageID)
		image.Name = envValue
	case strings.HasPrefix(envKey, raucImageClassEnvKeyPrefix):
		imageID := envKey[len(raucImageClassEnvKeyPrefix):]
		image := ctx.ensureImage(imageID)
		image.Class = envValue
	case strings.HasPrefix(envKey, raucImageDigestEnvKeyPrefix):
		imageID := envKey[len(raucImageDigestEnvKeyPrefix):]
		image := ctx.ensureImage(imageID)
		image.Digest = envValue
	default:
		return &UnknownEnvError{Key: envKey, Value: envValue}
	}

	return nil
}

func (ctx *HandlerContext) ensureSlot(slotID string) (slot *Slot) {
	var ok bool

	if ctx.Slots == nil {
		ctx.Slots = make(map[string]*Slot)
	}

	if slot, ok = ctx.Slots[slotID]; !ok {
		slot = &Slot{ID: slotID}
		ctx.Slots[slotID] = slot
	}

	return slot
}

func (ctx *HandlerContext) ensureImage(imageID string) (image *Image) {
	var ok bool

	if ctx.Images == nil {
		ctx.Images = make(map[string]*Image)
	}

	if image, ok = ctx.Images[imageID]; !ok {
		image = &Image{ID: imageID}
		ctx.Images[imageID] = image
	}

	return image
}

func (ctx *HandlerContext) synthesizeImageIDs() {
	// Synthesize ImageIDs, similar to SlotIDs.
	// RAUC does not provide those explicitly.
	for id := range ctx.Images {
		ctx.ImageIDs = append(ctx.ImageIDs, id)
	}

	sort.Strings(ctx.ImageIDs)
}

func (ctx *HandlerContext) validateSlotReferences() error {
	for _, slotID := range ctx.SlotIDs {
		if _, ok := ctx.Slots[slotID]; !ok {
			return &SlotReferenceError{SlotID: slotID, EnvVar: raucSlotsEnvKey}
		}
	}

	for _, slotID := range ctx.TargetSlotIDs {
		if _, ok := ctx.Slots[slotID]; !ok {
			return &SlotReferenceError{SlotID: slotID, EnvVar: raucTargetSlotsEnvKey}
		}
	}

	for _, slot := range ctx.Slots {
		if slot.ParentID != "" {
			if _, ok := ctx.Slots[slot.ParentID]; !ok {
				return &SlotReferenceError{
					SlotID: slot.ParentID,
					EnvVar: raucSlotParentEnvKeyPrefix + slot.ID,
				}
			}
		}
	}

	return nil
}

// ImageWithClass returns an image with the given class.
func (ctx *HandlerContext) ImageWithClass(cls string) *Image {
	for _, imageID := range ctx.ImageIDs {
		if image := ctx.Images[imageID]; image.Class == cls {
			return image
		}
	}

	return nil
}

// TargetSlotWith class returns a target slot with the given class.
func (ctx *HandlerContext) TargetSlotWithClass(cls string) *Slot {
	for _, slotID := range ctx.TargetSlotIDs {
		if slot := ctx.Slots[slotID]; slot.Class == cls {
			return slot
		}
	}

	return nil
}
