// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

// Package rauctest contains test utilities for the rauc package.
package rauctest

import (
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/boot"
)

// BootHandler implements rauc.CustomBootBackend and is useful for writing tests.
type BootHandler struct {
	PrimarySlotFn    func() (boot.Slot, error)
	SetPrimarySlotFn func(boot.Slot) error

	SlotStateFn    func(boot.Slot) (boot.SlotState, error)
	SetSlotStateFn func(boot.Slot, boot.SlotState) error
}

// ResetCallbacks resets all callback functions to nil.
func (handler *BootHandler) ResetCallbacks() {
	*handler = BootHandler{}
}

// PrimarySlot calls the PrimarySlotFn callback function.
func (handler *BootHandler) PrimarySlot() (boot.Slot, error) {
	if handler.PrimarySlotFn == nil {
		panic("please provide PrimarySlotFn callback function")
	}

	return handler.PrimarySlotFn()
}

// SetPrimarySlot calls the SetPrimarySlotFn callback function.
func (handler *BootHandler) SetPrimarySlot(slot boot.Slot) error {
	if handler.SetPrimarySlotFn == nil {
		panic("please provide SetPrimarySlotFn callback function")
	}

	return handler.SetPrimarySlotFn(slot)
}

// SlotState calls the SlotStateFn callback function.
func (handler *BootHandler) SlotState(slot boot.Slot) (boot.SlotState, error) {
	if handler.SlotStateFn == nil {
		panic("please provide SlotStateFn callback function")
	}

	return handler.SlotStateFn(slot)
}

// SetSlotState calls the SetSlotStateFn callback function.
func (handler *BootHandler) SetSlotState(slot boot.Slot, state boot.SlotState) error {
	if handler.SetSlotStateFn == nil {
		panic("please provide SetSlotStateFn callback function")
	}

	return handler.SetSlotStateFn(slot, state)
}

// BootState implements rauc.BootModeHolder for the BootProtocolAdapter
type BootState struct {
	bootMode boot.Mode

	// BootModeChanged can be used to observe or verify changes to boot mode.
	BootModeChanged func(boot.Mode) error
}

// BootMode returns the current boot mode.
func (state *BootState) BootMode() boot.Mode {
	return state.bootMode
}

// SetBootMode sets the new boot mode.
//
// Changes to boot mode are propagated to BootModeChanged callback before
// becoming effective, enabling both notifications and validation use-cases.
func (state *BootState) SetBootMode(bootMode boot.Mode) error {
	if state.bootMode == bootMode {
		return nil
	}

	if state.BootModeChanged != nil {
		if err := state.BootModeChanged(bootMode); err != nil {
			return err
		}
	}

	state.bootMode = bootMode

	return nil
}
