// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package rauctest_test

import (
	"testing"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/boot"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/rauc/rauctest"
	. "gopkg.in/check.v1"
)

func Test(t *testing.T) { TestingT(t) }

type bootBackendSuite struct{}

var _ = Suite(&bootBackendSuite{})

func (s *bootBackendSuite) TestPrimarySlot(c *C) {
	var backend rauctest.BootHandler

	c.Check(func() { _, _ = backend.PrimarySlot() }, PanicMatches, `please provide PrimarySlotFn callback function`)

	backend.PrimarySlotFn = func() (boot.Slot, error) {
		return boot.SlotA, nil
	}

	slot, err := backend.PrimarySlot()
	c.Assert(err, IsNil)
	c.Check(slot, Equals, boot.SlotA)
}

func (s *bootBackendSuite) TestSetPrimarySlot(c *C) {
	var backend rauctest.BootHandler

	c.Check(func() { _ = backend.SetPrimarySlot(boot.SlotA) }, PanicMatches, `please provide SetPrimarySlotFn callback function`)

	called := false
	backend.SetPrimarySlotFn = func(slot boot.Slot) error {
		c.Assert(slot, Equals, boot.SlotA)

		called = true

		return nil
	}

	err := backend.SetPrimarySlot(boot.SlotA)
	c.Assert(err, IsNil)
	c.Check(called, Equals, true)
}

func (s *bootBackendSuite) TestSlotState(c *C) {
	var backend rauctest.BootHandler

	c.Check(func() { _, _ = backend.SlotState(boot.SlotA) }, PanicMatches, `please provide SlotStateFn callback function`)

	called := false
	backend.SlotStateFn = func(slot boot.Slot) (boot.SlotState, error) {
		c.Assert(slot, Equals, boot.SlotA)

		called = true

		return boot.GoodSlot, nil
	}

	state, err := backend.SlotState(boot.SlotA)
	c.Assert(err, IsNil)
	c.Check(state, Equals, boot.GoodSlot)
	c.Check(called, Equals, true)
}

func (s *bootBackendSuite) TestSetSlotState(c *C) {
	var backend rauctest.BootHandler

	c.Check(func() { _ = backend.SetSlotState(boot.SlotA, boot.GoodSlot) }, PanicMatches, `please provide SetSlotStateFn callback function`)

	called := false
	backend.SetSlotStateFn = func(slot boot.Slot, state boot.SlotState) error {
		c.Assert(slot, Equals, boot.SlotA)
		c.Assert(state, Equals, boot.GoodSlot)

		called = true

		return nil
	}

	err := backend.SetSlotState(boot.SlotA, boot.GoodSlot)
	c.Assert(err, IsNil)
	c.Check(called, Equals, true)
}

func (s *bootBackendSuite) TestResetCallbacks(c *C) {
	backend := rauctest.BootHandler{
		PrimarySlotFn:    func() (boot.Slot, error) { return boot.InvalidSlot, nil },
		SetPrimarySlotFn: func(boot.Slot) error { return nil },
		SlotStateFn:      func(slot boot.Slot) (boot.SlotState, error) { return boot.InvalidSlotState, nil },
		SetSlotStateFn:   func(slot boot.Slot, state boot.SlotState) error { return nil },
	}

	backend.ResetCallbacks()

	c.Check(backend, DeepEquals, rauctest.BootHandler{})
}

type bootStateSuite struct{}

var _ = Suite(&bootStateSuite{})

func (s *bootStateSuite) TestBootModeReadWrite(c *C) {
	bootState := rauctest.BootState{}

	c.Check(bootState.BootMode(), Equals, boot.Normal)

	err := bootState.SetBootMode(boot.Try)
	c.Assert(err, IsNil)
	c.Check(bootState.BootMode(), Equals, boot.Try)
}

func (s *bootStateSuite) TestBootModeChangeObserver(c *C) {
	changes := make([]boot.Mode, 0, 3)
	bootState := rauctest.BootState{
		BootModeChanged: func(newBootMode boot.Mode) error {
			changes = append(changes, newBootMode)
			return nil
		},
	}

	c.Assert(bootState.SetBootMode(boot.Normal), IsNil)
	c.Check(bootState.BootMode(), Equals, boot.Normal)
	c.Assert(bootState.SetBootMode(boot.Try), IsNil)
	c.Check(bootState.BootMode(), Equals, boot.Try)
	c.Assert(bootState.SetBootMode(boot.Normal), IsNil)
	c.Check(bootState.BootMode(), Equals, boot.Normal)

	c.Check(changes, DeepEquals, []boot.Mode{boot.Try, boot.Normal})
}

func (s *bootStateSuite) TestBootModeChangeGuard(c *C) {
	bootState := rauctest.BootState{
		BootModeChanged: func(newBootMode boot.Mode) error {
			if newBootMode != boot.Normal && newBootMode != boot.Try {
				return boot.ErrInvalidBootMode
			}
			return nil
		},
	}

	c.Assert(bootState.SetBootMode(boot.Normal), IsNil)
	c.Check(bootState.BootMode(), Equals, boot.Normal)
	c.Assert(bootState.SetBootMode(boot.Try), IsNil)
	c.Check(bootState.BootMode(), Equals, boot.Try)
	c.Assert(bootState.SetBootMode(boot.InvalidBootMode), ErrorMatches, `invalid boot mode`)
	c.Check(bootState.BootMode(), Equals, boot.Try)
}
