// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

// Package service implements the sysotad D-Bus service.
package service

import (
	"errors"

	dbus "github.com/godbus/dbus/v5"
	"github.com/godbus/dbus/v5/introspect"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/boot"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/dbusutil"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/ota"
)

const (
	// BootLoaderInterfaceName is the name of the SystemOTA BootLoader interface.
	BootLoaderInterfaceName = "dev.ostc.sysota1.BootLoader"
)

// BootLoaderIntrospectData is the D-Bus introspection data for the BootLoader interface of the service object.
var BootLoaderIntrospectData = introspect.Interface{
	Name: BootLoaderInterfaceName,
	Methods: []introspect.Method{
		{
			Name: "QueryActive",
			Args: []introspect.Arg{
				{
					Name:      "slot",
					Type:      "s",
					Direction: dbusutil.Out,
				},
			},
		},
		{
			Name: "QueryInactive",
			Args: []introspect.Arg{
				{
					Name:      "slot",
					Type:      "s",
					Direction: dbusutil.Out,
				},
			},
		},
		{
			Name: "TrySwitch",
			Args: []introspect.Arg{
				{
					Name:      "slot",
					Type:      "s",
					Direction: dbusutil.In,
				},
			},
		},
		{
			Name: "Reboot",
			Args: []introspect.Arg{
				{
					Name:      "flags",
					Type:      "u",
					Direction: dbusutil.In,
				},
			},
		},
		{
			Name: "CommitSwitch",
		},
		{
			Name: "CancelSwitch",
		},
	},
}

var errNoBootProtocol = errors.New("boot protocol is not set")

// BootProtocolService is a dbusutil.HostedService exporting boot.Protocol over D-BUs.
type BootProtocolService struct {
	bootProto boot.Protocol
}

// JoinServiceHost integrates with dbusutil.ServiceHost.
func (svc *BootProtocolService) JoinServiceHost(reg dbusutil.ServiceRegistration) error {
	methods := map[string]interface{}{
		"QueryActive":   svc.QueryActive,
		"QueryInactive": svc.QueryInactive,
		"TrySwitch":     svc.TrySwitch,
		"Reboot":        svc.Reboot,
		"CommitSwitch":  svc.CommitSwitch,
		"CancelSwitch":  svc.CancelSwitch,
	}

	return reg.RegisterObject(ota.ServiceObjectPath, BootLoaderInterfaceName, methods, BootLoaderIntrospectData)
}

// QueryActive exposes the boot.Protocol.QueryActive method over D-Bus.
func (svc *BootProtocolService) QueryActive() (string, *dbus.Error) {
	if svc.bootProto == nil {
		return "", dbus.MakeFailedError(errNoBootProtocol)
	}

	slot, err := svc.bootProto.QueryActive()
	if err != nil {
		return "", dbus.MakeFailedError(err)
	}

	return slot.String(), nil
}

// QueryInactive exposes the boot.Protocol.QueryInactive method over D-Bus.
func (svc *BootProtocolService) QueryInactive() (string, *dbus.Error) {
	if svc.bootProto == nil {
		return "", dbus.MakeFailedError(errNoBootProtocol)
	}

	slot, err := svc.bootProto.QueryInactive()
	if err != nil {
		return "", dbus.MakeFailedError(err)
	}

	return slot.String(), nil
}

// TrySwitch exposes the boot.Protocol.TrySwitch method over D-Bus.
func (svc *BootProtocolService) TrySwitch(slotName string) *dbus.Error {
	if svc.bootProto == nil {
		return dbus.MakeFailedError(errNoBootProtocol)
	}

	var slot boot.Slot
	if err := slot.UnmarshalText([]byte(slotName)); err != nil {
		return dbus.MakeFailedError(err)
	}

	if err := svc.bootProto.TrySwitch(slot); err != nil {
		return dbus.MakeFailedError(err)
	}

	return nil
}

// Reboot exposes the boot.Protocol.Reboot method over D-Bus.
//
// The flags argument is not translated in any way.
func (svc *BootProtocolService) Reboot(flags uint) *dbus.Error {
	if svc.bootProto == nil {
		return dbus.MakeFailedError(errNoBootProtocol)
	}

	if err := svc.bootProto.Reboot(boot.RebootFlags(flags)); err != nil {
		return dbus.MakeFailedError(err)
	}

	return nil
}

// CommitSwitch exposes the boot.Protocol.CommitSwitch method over D-Bus.
func (svc *BootProtocolService) CommitSwitch() *dbus.Error {
	if svc.bootProto == nil {
		return dbus.MakeFailedError(errNoBootProtocol)
	}

	if err := svc.bootProto.CommitSwitch(); err != nil {
		return dbus.MakeFailedError(err)
	}

	return nil
}

// CancelSwitch exposes the boot.Protocol.CancelSwitch method over D-Bus.
func (svc *BootProtocolService) CancelSwitch() *dbus.Error {
	if svc.bootProto == nil {
		return dbus.MakeFailedError(errNoBootProtocol)
	}

	if err := svc.bootProto.CancelSwitch(); err != nil {
		return dbus.MakeFailedError(err)
	}

	return nil
}
