// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

// Package oper implements SystemOTA Operation object
package oper

import (
	"fmt"
	"sync"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/dbusutil"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/dbusutil/objmgr"
	dbus "github.com/godbus/dbus/v5"
	"github.com/godbus/dbus/v5/introspect"
	"github.com/godbus/dbus/v5/prop"
)

const (
	statusProperty = "Status"
	idProperty     = "ID"
)

const (
	// InterfaceName is the name of the SystemOTA Operation interface.
	InterfaceName = "dev.ostc.sysota1.Operation"
)

// IntrospectData is the D-Bus introspection data for the operation object.
var IntrospectData = introspect.Interface{
	Name: InterfaceName,
	Methods: []introspect.Method{
		{
			Name: "Cancel",
		},
		{
			Name: "Resume",
		},
		{
			Name: "Pause",
		},
	},
	Properties: []introspect.Property{
		{
			Name:   idProperty,
			Type:   "i",
			Access: "read",
			Annotations: []introspect.Annotation{
				{
					Name:  dbusutil.PropertyAnnotationEmitsChangedSignal,
					Value: "const",
				},
			},
		},
		{
			Name:   statusProperty,
			Type:   "s",
			Access: "read",
			Annotations: []introspect.Annotation{
				{
					Name:  dbusutil.PropertyAnnotationEmitsChangedSignal,
					Value: "true",
				},
			},
		},
	},
}

// Status encodes one of possible states an operation can be in.
type Status int

const (
	// StatusWaiting indicates that an operation is waiting to start.
	StatusWaiting Status = iota
	// StatusRunning indicates that an operation is in progress.
	StatusRunning
	// StatusPaused indicates than an operation has been paused.
	StatusPaused
	// StatusCancelled indicates that an operation has been cancelled.
	StatusCancelled
	// StatusFinished indicates that an operation has finished.
	StatusFinished
)

// String returns the textual representation of the status.
func (st Status) String() string {
	switch st {
	case StatusWaiting:
		return "waiting"
	case StatusRunning:
		return "running"
	case StatusPaused:
		return "paused"
	case StatusCancelled:
		return "cancelled"
	case StatusFinished:
		return "finished"
	default:
		return fmt.Sprintf("status/%d", st)
	}
}

// ObjectPath returns the D-Bus object path for an operation with the given ID.
func ObjectPath(id int) dbus.ObjectPath {
	return dbus.ObjectPath(fmt.Sprintf("/dev/ostc/sysota1/Service/operation/%d", id))
}

// Operation represents a background activity that takes a non-trivial amount of time.
type Operation struct {
	m      sync.RWMutex
	conn   *dbus.Conn
	id     int
	status Status
}

// New returns a new operation.
func New(conn *dbus.Conn, ID int) *Operation {
	return &Operation{conn: conn, id: ID}
}

// Export exports the operation object on the bus.
func (op *Operation) Export() (err error) {
	op.m.Lock()
	defer op.m.Unlock()

	defer func() {
		if err != nil {
			_ = op.unexportUnlocked()
		}
	}()

	path := ObjectPath(op.id)

	opMethods := map[string]interface{}{
		"Pause":  op.Pause,
		"Resume": op.Resume,
		"Cancel": op.Cancel,
	}
	if err := op.conn.ExportMethodTable(opMethods, path, InterfaceName); err != nil {
		return err
	}

	propsMethods := map[string]interface{}{
		"Get":    op.Get,
		"GetAll": op.GetAll,
		"Set":    op.Set,
	}
	if err := op.conn.ExportMethodTable(propsMethods, path, dbusutil.PropertiesInterfaceName); err != nil {
		return err
	}

	node := &introspect.Node{
		Name: string(path),
		Interfaces: []introspect.Interface{
			introspect.IntrospectData,
			prop.IntrospectData,
			objmgr.IntrospectData,
			IntrospectData,
		},
	}
	if err := op.conn.Export(introspect.NewIntrospectable(node), path, dbusutil.IntrospectableInterfaceName); err != nil {
		return err
	}

	return nil
}

// Unexport removes the object from the set of published bus objects.
func (op *Operation) Unexport() error {
	op.m.Lock()
	defer op.m.Unlock()

	return op.unexportUnlocked()
}

func (op *Operation) unexportUnlocked() error {
	path := ObjectPath(op.id)
	if err := op.conn.Export(nil, path, InterfaceName); err != nil {
		return err
	}

	if err := op.conn.Export(nil, path, dbusutil.PropertiesInterfaceName); err != nil {
		return err
	}

	if err := op.conn.Export(nil, path, dbusutil.IntrospectableInterfaceName); err != nil {
		return err
	}

	return nil
}

// ID returns the unique identifier of the operation.
func (op *Operation) ID() int {
	op.m.RLock()
	defer op.m.RUnlock()

	return op.id
}

// Status returns the current status of the operation.
func (op *Operation) Status() Status {
	op.m.RLock()
	defer op.m.RUnlock()

	return op.status
}

// Pause pauses the operation, if possible.
func (op *Operation) Pause() *dbus.Error {
	op.m.Lock()
	defer op.m.Unlock()

	switch op.status {
	case StatusWaiting:
		op.status = StatusPaused
		return nil
	default:
		return dbus.NewError("dev.ostc.sysota1.Operation.CannotPause", nil)
	}
}

// Resume resumes a paused operation.
func (op *Operation) Resume() *dbus.Error {
	op.m.Lock()
	defer op.m.Unlock()

	switch op.status {
	case StatusPaused:
		op.status = StatusWaiting
		return nil
	default:
		return dbus.NewError("dev.ostc.sysota1.Operation.CannotResume", nil)
	}
}

// Cancel cancels an operation.
func (op *Operation) Cancel() *dbus.Error {
	op.m.Lock()
	defer op.m.Unlock()

	switch op.status {
	case StatusWaiting:
		op.status = StatusCancelled
		return nil
	default:
		return dbus.NewError("dev.ostc.sysota1.Operation.CannotCancel", nil)
	}
}

// XXX: Properties are implemented without github.com/godbus/v5/prop.Properties
// more precise control. We may drop this later, depending on how it gets used.

// Get implements org.freedesktop.DBus.Properties.Get.
func (op *Operation) Get(iface, property string) (dbus.Variant, *dbus.Error) {
	op.m.RLock()
	defer op.m.RUnlock()

	switch iface {
	case InterfaceName:
		switch property {
		case idProperty:
			return dbus.MakeVariant(op.id), nil
		case statusProperty:
			return dbus.MakeVariant(op.status.String()), nil
		default:
			return dbus.Variant{}, prop.ErrPropNotFound
		}
	default:
		return dbus.Variant{}, prop.ErrIfaceNotFound
	}
}

// GetAll implements org.freedesktop.DBus.Properties.GetAll.
func (op *Operation) GetAll(iface string) (map[string]dbus.Variant, *dbus.Error) {
	op.m.RLock()
	defer op.m.RUnlock()

	switch iface {
	case InterfaceName:
		return map[string]dbus.Variant{
			idProperty:     dbus.MakeVariant(op.id),
			statusProperty: dbus.MakeVariant(op.status.String()),
		}, nil
	default:
		return nil, prop.ErrIfaceNotFound
	}
}

// Set implements org.freedesktop.Properties.Set.
func (op *Operation) Set(iface, property string, newValue dbus.Variant) *dbus.Error {
	op.m.Lock()
	defer op.m.Unlock()

	switch iface {
	case InterfaceName:
		switch property {
		case idProperty, statusProperty:
			return prop.ErrReadOnly
		default:
			return prop.ErrPropNotFound
		}
	default:
		return prop.ErrIfaceNotFound
	}
}
