// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

// Package service implements the sysotad D-Bus service.
package service

import (
	"fmt"
	"sync"

	dbus "github.com/godbus/dbus/v5"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/boot"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/boot/piboot"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/dbusutil"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/ota"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/ota/raucadapter"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/ota/raucinterface"
)

// RequestBusName requests a well-known name on the bus
//
// Bus name must be claimed after creating and exporting the service
// object, as it is used as an indicator that a systemd-service is ready
// for accepting connections.
func RequestBusName(conn *dbus.Conn) error {
	reply, err := conn.RequestName(ota.ServiceBusName, dbus.NameFlagDoNotQueue)
	if err != nil {
		return err
	}

	if reply != dbus.RequestNameReplyPrimaryOwner {
		return fmt.Errorf("cannot claim bus name %s", ota.ServiceBusName)
	}

	return nil
}

// Service implements the D-Bus service capable of downloading system updates.
type Service struct {
	m sync.RWMutex

	state  *ota.SystemState
	config *ota.Config

	sh *dbusutil.ServiceHost

	bootService   BootProtocolService
	updateService UpdateService

	raucAdapter *raucadapter.Adapter
	raucService *raucinterface.Service
}

var bootProviders = map[ota.BootLoaderType]func(config *ota.Config) (boot.Protocol, error){
	ota.InertBootLoader: func(config *ota.Config) (boot.Protocol, error) { return nil, nil },
	ota.PiBoot:          func(config *ota.Config) (boot.Protocol, error) { return piboot.New(config.BootPartitionMountDir) },
}

type unsupportedBootLoaderError ota.BootLoaderType

func (e unsupportedBootLoaderError) Error() string {
	return fmt.Sprintf("unsupported boot loader type: %s", ota.BootLoaderType(e))
}

// configuredBootProtocol returns the boot protocol described by the configuration.
func configuredBootProtocol(config *ota.Config) (boot.Protocol, error) {
	if loader, ok := bootProviders[config.BootLoaderType]; ok {
		return loader(config)
	}

	return nil, unsupportedBootLoaderError(config.BootLoaderType)
}

// New returns a new service.
func New(config *ota.Config, state *ota.SystemState, conn *dbus.Conn) (*Service, error) {
	if config == nil {
		panic("config cannot be nil")
	}

	if state == nil {
		panic("state cannot be nil")
	}

	if conn == nil {
		panic("conn cannot be nil")
	}

	bootProto, err := configuredBootProtocol(config)
	if err != nil {
		return nil, err
	}

	svc := &Service{state: state, config: config}

	svc.raucAdapter = raucadapter.New(bootProto, svc)
	svc.sh = dbusutil.NewServiceHost(conn)

	// Configure and add the boot service
	// TODO(zyga): disable this by default and only enable it
	// when [debug] BootProtocol=true is set in the config file.
	svc.bootService.bootProto = bootProto
	svc.sh.AddHostedService(&svc.bootService)

	// Configure and add the update service.
	svc.updateService.maker = "dummy-maker"
	svc.updateService.model = "dummy-model"
	svc.updateService.stream = "dummy-stream"
	svc.sh.AddHostedService(&svc.updateService)

	// Configure and add the RAUC service
	svc.raucService = raucinterface.NewService(svc.raucAdapter, svc.raucAdapter)
	svc.sh.AddHostedService(svc.raucService)

	return svc, nil
}

// BootMode implements raucadapter.BootModeHolder and returns the current boot mode.
func (svc *Service) BootMode() boot.Mode {
	return svc.state.BootMode
}

// SetBootMode implements raucadapter.BootModeHolder and sets the new boot mode.
func (svc *Service) SetBootMode(bootMode boot.Mode) error {
	if svc.state.BootMode == bootMode {
		return nil
	}

	svc.state.BootMode = bootMode

	// TODO(zyga): emit the PropertyChanged signal here.
	// Any errors should be logged and ignored.

	return nil
}

// Export exports the service object on the bus.
func (svc *Service) Export() (err error) {
	svc.m.Lock()
	defer svc.m.Unlock()

	return svc.sh.Export()
}

// Unexport removes the object from the set of published bus objects.
func (svc *Service) Unexport() error {
	svc.m.Lock()
	defer svc.m.Unlock()

	return svc.sh.Unexport()
}

// IsIdle returns true if all the hosted service are idle.
func (svc *Service) IsIdle() bool {
	return svc.updateService.IsIdle()
}

// Maker returns the maker name used by the update service.
func (svc *Service) Maker() string {
	return svc.updateService.Maker()
}

// Model returns the model name used by the update service.
func (svc *Service) Model() string {
	return svc.updateService.Model()
}

// Stream returns the stream used by the update service.
func (svc *Service) Stream() string {
	return svc.updateService.Stream()
}

// SetStream sets the stream used by the update service.
func (svc *Service) SetStream(stream string) error {
	return svc.updateService.SetStream(stream)
}
