// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

// Package service implements the sysotad D-Bus service.
package service

import (
	"fmt"
	"sync"

	dbus "github.com/godbus/dbus/v5"
	"github.com/godbus/dbus/v5/introspect"
	"github.com/godbus/dbus/v5/prop"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/dbusutil"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/dbusutil/objmgr"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/ota"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/service/oper"
)

const (
	// UpdateServiceInterfaceName is the name of the SystemOTA Service interface.
	// TODO(zyga): re-think the name.
	UpdateServiceInterfaceName = "dev.ostc.sysota1.Service"

	makerProperty  = "Maker"
	modelProperty  = "Model"
	streamProperty = "Stream"
)

// ServiceIntrospectData is the D-Bus introspection data for Service interface of the service object.
//
// It is provided explicitly without using reflection for a bit more control
// over the provided data, given that we implement the properties interface
// ourselves.
var ServiceIntrospectData = introspect.Interface{
	Name: UpdateServiceInterfaceName,
	Methods: []introspect.Method{
		{
			Name: "UpdateStreams",
		},
		{
			Name: "UpdateDevice",
			Args: []introspect.Arg{
				{
					Name:      "hints",
					Type:      "a{ss}",
					Direction: dbusutil.In,
				},
				{
					Name:      "operation",
					Type:      "o",
					Direction: dbusutil.Out,
				},
			},
		},
	},
	Properties: []introspect.Property{
		{
			Name:   makerProperty,
			Type:   "s",
			Access: "read",
			Annotations: []introspect.Annotation{
				{
					Name:  dbusutil.PropertyAnnotationEmitsChangedSignal,
					Value: "const",
				},
			},
		},
		{
			Name:   modelProperty,
			Type:   "s",
			Access: "read",
			Annotations: []introspect.Annotation{
				{
					Name:  dbusutil.PropertyAnnotationEmitsChangedSignal,
					Value: "const",
				},
			},
		},
		{
			Name:   streamProperty,
			Type:   "s",
			Access: "readwrite",
			Annotations: []introspect.Annotation{
				{
					Name:  dbusutil.PropertyAnnotationEmitsChangedSignal,
					Value: "true",
				},
			},
		},
	},
}

// UpdateService provides functions for downloading and applying updates.
//
// Maker is a near-arbitrary strign that identifier the device maker.
// Model is a near-arbitrary string that is selected by the device maker.
// Stream is a path-like string that identifiers which of perhaps several versions,
// branches or stability levels to use.
//
// Available updates are specific to a given (maker,model,stream) tuple.
type UpdateService struct {
	m    sync.RWMutex
	conn *dbus.Conn

	maker  string
	model  string
	stream string

	ops []*oper.Operation

	lastOperationID int
}

// JoinServiceHost integrates with dbusutil.ServiceHost.
func (svc *UpdateService) JoinServiceHost(reg dbusutil.ServiceRegistration) error {
	// Remember connection to send signals.
	svc.conn = reg.DBusConn()

	const path = ota.ServiceObjectPath

	svcMethods := map[string]interface{}{
		"UpdateStreams": svc.UpdateStreams,
		"UpdateDevice":  svc.UpdateDevice,
	}
	if err := reg.RegisterObject(path, UpdateServiceInterfaceName, svcMethods, ServiceIntrospectData); err != nil {
		return err
	}

	propsMethods := map[string]interface{}{
		"Get":    svc.Get,
		"GetAll": svc.GetAll,
		"Set":    svc.Set,
	}
	if err := reg.RegisterObject(path, dbusutil.PropertiesInterfaceName, propsMethods, prop.IntrospectData); err != nil {
		return err
	}

	objmgrMethods := map[string]interface{}{
		"GetManagedObjects": svc.GetManagedObjects,
	}
	if err := reg.RegisterObject(path, objmgr.InterfaceName, objmgrMethods, objmgr.IntrospectData); err != nil {
		return err
	}

	return nil
}

// Maker returns the name of the maker of the device.
func (svc *UpdateService) Maker() string {
	svc.m.RLock()
	defer svc.m.RUnlock()

	return svc.maker
}

// Model returns the model name of the device.
func (svc *UpdateService) Model() string {
	svc.m.RLock()
	defer svc.m.RUnlock()

	return svc.model
}

// Stream returns the name of the update stream the device is subscribed to.
func (svc *UpdateService) Stream() string {
	svc.m.RLock()
	defer svc.m.RUnlock()

	return svc.stream
}

// SetStream sets the new stream name to the given value.
func (svc *UpdateService) SetStream(stream string) error {
	svc.m.Lock()
	defer svc.m.Unlock()

	return svc.setStreamUnlocked(stream)
}

// setStreamUnlocked sets the new stream name to the given value.
//
// It must be called with the write lock held.
func (svc *UpdateService) setStreamUnlocked(stream string) error {
	// TODO check that stream is valid/known.
	svc.stream = stream
	// TODO: persist the configuration
	updated := map[string]dbus.Variant{
		streamProperty: dbus.MakeVariant(stream),
	}
	invalidated := []string{}

	return svc.conn.Emit(ota.ServiceObjectPath, dbusutil.PropertiesChangedSignal, UpdateServiceInterfaceName, updated, invalidated)
}

// IsIdle returns true if the OTA service is idle and can be shut down.
func (svc *UpdateService) IsIdle() bool {
	svc.m.RLock()
	defer svc.m.RUnlock()

	return len(svc.ops) == 0
}

// UpdateStreams asks the update server available streams.
//
// Updating the stream map gives the device awareness of the available streams
// and their meta-data.
//
// TODO: Implement this for real by calling thing from the OTA backend.
// XXX: The return type is probably wrong. It should return "ao"
// representing the known streams.
func (svc *UpdateService) UpdateStreams() *dbus.Error {
	svc.m.Lock()
	defer svc.m.Unlock()

	fmt.Printf("called UpdateStreams\n")

	return nil
}

// UpdateDevice starts and asynchronous update operation.
//
// The OTA service checks the configured update server for available updates
// and returns an object which can be used to track the progress and result.
//
// TODO: Implement this for real by calling thing from the OTA backend.
func (svc *UpdateService) UpdateDevice(hints map[string]string) (dbus.ObjectPath, *dbus.Error) {
	svc.m.Lock()
	defer svc.m.Unlock()

	// TODO: implement this for real
	fmt.Printf("called UpdateDevice with hints %v\n", hints)

	op, err := svc.makeOp()
	if err != nil {
		return "", dbus.MakeFailedError(err)
	}

	svc.ops = append(svc.ops, op)

	return oper.ObjectPath(op.ID()), nil
}

func (svc *UpdateService) makeOp() (*oper.Operation, error) {
	svc.lastOperationID++
	op := oper.New(svc.conn, svc.lastOperationID)

	// NOTE: dbusErr is of type *dbus.Error. A nil value converts to non-nil
	// error, because error is an internal interface. Avoid undesired conversion
	// yielding non-nil errors pointing to nil values of specific pointer type.
	operProps, dbusErr := op.GetAll(oper.InterfaceName)
	if dbusErr != nil {
		// *dbus.Error converted to error
		return nil, dbusErr
	}

	path := oper.ObjectPath(op.ID())
	ifacesProps := map[string]map[string]dbus.Variant{oper.InterfaceName: operProps}

	if err := svc.conn.Emit(path, objmgr.InterfacesAddedSignal, path, ifacesProps); err != nil {
		return nil, err
	}

	if err := op.Export(); err != nil {
		return nil, err
	}

	return op, nil
}

// XXX: Properties are implemented without github.com/godbus/v5/prop.Properties
// more precise control. We may drop this later, depending on how it gets used.

// Get implements org.freedesktop.DBus.Properties.Get.
func (svc *UpdateService) Get(iface, property string) (dbus.Variant, *dbus.Error) {
	svc.m.RLock()
	defer svc.m.RUnlock()

	switch iface {
	case UpdateServiceInterfaceName:
		switch property {
		case makerProperty:
			return dbus.MakeVariant(svc.maker), nil
		case modelProperty:
			return dbus.MakeVariant(svc.model), nil
		case streamProperty:
			return dbus.MakeVariant(svc.stream), nil
		default:
			return dbus.Variant{}, prop.ErrPropNotFound
		}
	default:
		return dbus.Variant{}, prop.ErrIfaceNotFound
	}
}

// GetAll implements org.freedesktop.DBus.Properties.GetAll.
func (svc *UpdateService) GetAll(iface string) (map[string]dbus.Variant, *dbus.Error) {
	svc.m.RLock()
	defer svc.m.RUnlock()

	switch iface {
	case UpdateServiceInterfaceName:
		return map[string]dbus.Variant{
			makerProperty:  dbus.MakeVariant(svc.maker),
			modelProperty:  dbus.MakeVariant(svc.model),
			streamProperty: dbus.MakeVariant(svc.stream),
		}, nil
	default:
		return nil, prop.ErrIfaceNotFound
	}
}

// Set implements org.freedesktop.Properties.Set.
func (svc *UpdateService) Set(iface, property string, newValue dbus.Variant) (err *dbus.Error) {
	svc.m.Lock()
	defer svc.m.Unlock()

	switch iface {
	case UpdateServiceInterfaceName:
		switch property {
		case makerProperty, modelProperty:
			return prop.ErrReadOnly
		case streamProperty:
			if newValue.Signature() != dbus.SignatureOf(svc.stream) {
				return prop.ErrInvalidArg
			}

			if err := svc.setStreamUnlocked(newValue.Value().(string)); err != nil {
				return dbus.MakeFailedError(err)
			}
		default:
			return prop.ErrPropNotFound
		}
	default:
		return prop.ErrIfaceNotFound
	}

	return nil
}

// GetManagedObjects implements org.freedesktop.DBus.ObjectManager.GetManagedObjects.
func (svc *UpdateService) GetManagedObjects() (map[dbus.ObjectPath]map[string]map[string]dbus.Variant, *dbus.Error) {
	svc.m.RLock()
	defer svc.m.RUnlock()

	objsIfacesProps := make(map[dbus.ObjectPath]map[string]map[string]dbus.Variant, len(svc.ops))

	for _, op := range svc.ops {
		props, dbusErr := op.GetAll(oper.InterfaceName)
		if dbusErr != nil {
			return nil, dbusErr
		}

		objsIfacesProps[oper.ObjectPath(op.ID())] = map[string]map[string]dbus.Variant{
			oper.InterfaceName: props,
		}
	}

	return objsIfacesProps, nil
}
