// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package mksquashfs_test

import (
	"fmt"
	"io/ioutil"
	"os"
	"os/exec"
	"path/filepath"
	"testing"

	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/squashfstools/mksquashfs"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/squashfstools/mksquashfs/squashgzip"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/squashfstools/mksquashfs/squashlz4"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/squashfstools/mksquashfs/squashlzma"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/squashfstools/mksquashfs/squashlzo"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/squashfstools/mksquashfs/squashxz"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/squashfstools/mksquashfs/squashzstd"
)

func Test(t *testing.T) { TestingT(t) }

type mksquashfsSuite struct{}

var _ = Suite(&mksquashfsSuite{})

func (s *mksquashfsSuite) TestWithoutOptions(c *C) {
	cmd := mksquashfs.Command("output.img", nil, "input-file", "input-dir")
	c.Check(cmd.Args, DeepEquals, []string{
		"mksquashfs", "input-file", "input-dir", "output.img",
		"-comp", "gzip", "-noappend", "-no-progress", "-no-recovery",
		"-no-exports", "-no-sparse", "-no-xattrs", "-noI", "-noD", "-noF", "-noX",
		"-no-fragments", "-nopad"})
}

func (s *mksquashfsSuite) TestInputFilesAndOutputFiles(c *C) {
	cmd := mksquashfs.Command("output.img", nil, "input-file", "input-dir")
	c.Check(cmd.Args[:4], DeepEquals, []string{"mksquashfs", "input-file", "input-dir", "output.img"})
}

// TODO(zyga): add checkers like HasArgument and ArgumentValue

func argIndex(cmd *exec.Cmd, wantedArg string) int {
	for idx, arg := range cmd.Args {
		if arg == wantedArg {
			return idx
		}
	}

	return -1
}

func argValue(cmd *exec.Cmd, wantedArg string) string {
	idx := argIndex(cmd, wantedArg)
	if idx == -1 {
		panic(fmt.Sprintf("cannot find argument %s in %v\n", wantedArg, cmd.Args))
	}

	if idx+1 < len(cmd.Args) {
		return cmd.Args[idx+1]
	}

	panic(fmt.Sprintf("argument %s has no value: %v", wantedArg, cmd.Args))
}

var compressConfigs = []mksquashfs.CompressionConfig{
	&squashgzip.Config{},
	&squashlz4.Config{},
	&squashlzma.Config{},
	&squashlzo.Config{},
	&squashxz.Config{},
	&squashzstd.Config{},
}

func (s *mksquashfsSuite) TestCreateWithCompression(c *C) {
	for _, config := range compressConfigs {
		cmd := mksquashfs.Command("output.img", &mksquashfs.Options{Compression: config}, "input-dir")
		c.Check(argValue(cmd, "-comp"), Equals, config.CompressionName())
	}
}

func (s *mksquashfsSuite) TestCreateWithBlockSize(c *C) {
	cmd := mksquashfs.Command("output.img", &mksquashfs.Options{BlockSize: 4096}, "input-dir")
	c.Check(argValue(cmd, "-b"), Equals, "4096")

	cmd = mksquashfs.Command("output.img", &mksquashfs.Options{BlockSize: 0}, "input-dir")
	c.Check(argIndex(cmd, "-b"), Equals, -1)
}

func (s *mksquashfsSuite) TestCreateWithNfsExportable(c *C) {
	cmd := mksquashfs.Command("output.img", &mksquashfs.Options{NfsExportable: true}, "input-dir")
	c.Check(argIndex(cmd, "-no-exports"), Equals, -1)

	cmd = mksquashfs.Command("output.img", &mksquashfs.Options{NfsExportable: false}, "input-dir")
	c.Check(argIndex(cmd, "-no-exports"), Not(Equals), -1)
}

func (s *mksquashfsSuite) TestCreateWithSparseFiles(c *C) {
	cmd := mksquashfs.Command("output.img", &mksquashfs.Options{Sparse: true}, "input-dir")
	c.Check(argIndex(cmd, "-no-sparse"), Equals, -1)

	cmd = mksquashfs.Command("output.img", &mksquashfs.Options{Sparse: false}, "input-dir")
	c.Check(argIndex(cmd, "-no-sparse"), Not(Equals), -1)
}

func (s *mksquashfsSuite) TestCreateWithXattrs(c *C) {
	// There's an affirmative option for xattrs, unlike most of the other options.
	cmd := mksquashfs.Command("output.img", &mksquashfs.Options{Xattrs: true}, "input-dir")
	c.Check(argIndex(cmd, "-xattrs"), Not(Equals), -1)
	c.Check(argIndex(cmd, "-no-xattrs"), Equals, -1)

	cmd = mksquashfs.Command("output.img", &mksquashfs.Options{Xattrs: false}, "input-dir")
	c.Check(argIndex(cmd, "-xattrs"), Equals, -1)
	c.Check(argIndex(cmd, "-no-xattrs"), Not(Equals), -1)
}

func (s *mksquashfsSuite) TestCreateWithInodeCompression(c *C) {
	cmd := mksquashfs.Command("output.img", &mksquashfs.Options{CompressInodes: true}, "input-dir")
	c.Check(argIndex(cmd, "-noI"), Equals, -1)

	cmd = mksquashfs.Command("output.img", &mksquashfs.Options{CompressInodes: false}, "input-dir")
	c.Check(argIndex(cmd, "-noI"), Not(Equals), -1)
}

func (s *mksquashfsSuite) TestCreateWithDataCompression(c *C) {
	cmd := mksquashfs.Command("output.img", &mksquashfs.Options{CompressData: true}, "input-dir")
	c.Check(argIndex(cmd, "-noD"), Equals, -1)

	cmd = mksquashfs.Command("output.img", &mksquashfs.Options{CompressData: false}, "input-dir")
	c.Check(argIndex(cmd, "-noD"), Not(Equals), -1)
}

func (s *mksquashfsSuite) TestCreateWithFragmentCompression(c *C) {
	cmd := mksquashfs.Command("output.img", &mksquashfs.Options{CompressFragments: true}, "input-dir")
	c.Check(argIndex(cmd, "-noF"), Equals, -1)

	cmd = mksquashfs.Command("output.img", &mksquashfs.Options{CompressFragments: false}, "input-dir")
	c.Check(argIndex(cmd, "-noF"), Not(Equals), -1)
}

func (s *mksquashfsSuite) TestCreateWithXattrCompression(c *C) {
	cmd := mksquashfs.Command("output.img", &mksquashfs.Options{CompressXattrs: true}, "input-dir")
	c.Check(argIndex(cmd, "-noX"), Equals, -1)

	cmd = mksquashfs.Command("output.img", &mksquashfs.Options{CompressXattrs: false}, "input-dir")
	c.Check(argIndex(cmd, "-noX"), Not(Equals), -1)
}

func (s *mksquashfsSuite) TestCreateWithFragments(c *C) {
	cmd := mksquashfs.Command("output.img", &mksquashfs.Options{Fragments: true}, "input-dir")
	c.Check(argIndex(cmd, "-no-fragments"), Equals, -1)

	cmd = mksquashfs.Command("output.img", &mksquashfs.Options{Fragments: false}, "input-dir")
	c.Check(argIndex(cmd, "-no-fragments"), Not(Equals), -1)
}

func (s *mksquashfsSuite) TestCreateWithPadding(c *C) {
	cmd := mksquashfs.Command("output.img", &mksquashfs.Options{Pad: true}, "input-dir")
	c.Check(argIndex(cmd, "-nopad"), Equals, -1)

	cmd = mksquashfs.Command("output.img", &mksquashfs.Options{Pad: false}, "input-dir")
	c.Check(argIndex(cmd, "-nopad"), Not(Equals), -1)
}

func (s *mksquashfsSuite) TestCreateWithNumProcessors(c *C) {
	cmd := mksquashfs.Command("output.img", &mksquashfs.Options{NumProcessors: 2}, "input-dir")
	c.Check(argIndex(cmd, "-processors"), Not(Equals), -1)
	c.Check(argValue(cmd, "-processors"), Equals, "2")

	cmd = mksquashfs.Command("output.img", &mksquashfs.Options{NumProcessors: 0}, "input-dir")
	c.Check(argIndex(cmd, "-processors"), Equals, -1)
}

func (s *mksquashfsSuite) TestCreateWithReadQueueSize(c *C) {
	cmd := mksquashfs.Command("output.img", &mksquashfs.Options{ReadQueueSizeMB: 16}, "input-dir")
	c.Check(argIndex(cmd, "-read-queue"), Not(Equals), -1)
	c.Check(argValue(cmd, "-read-queue"), Equals, "16")

	cmd = mksquashfs.Command("output.img", &mksquashfs.Options{ReadQueueSizeMB: 0}, "input-dir")
	c.Check(argIndex(cmd, "-read-queue"), Equals, -1)
}

func (s *mksquashfsSuite) TestCreateWithWriteQueueSize(c *C) {
	cmd := mksquashfs.Command("output.img", &mksquashfs.Options{WriteQueueSizeMB: 16}, "input-dir")
	c.Check(argIndex(cmd, "-write-queue"), Not(Equals), -1)
	c.Check(argValue(cmd, "-write-queue"), Equals, "16")

	cmd = mksquashfs.Command("output.img", &mksquashfs.Options{WriteQueueSizeMB: 0}, "input-dir")
	c.Check(argIndex(cmd, "-write-queue"), Equals, -1)
}

func (s *mksquashfsSuite) TestCreateWithFragmentQueueSize(c *C) {
	cmd := mksquashfs.Command("output.img", &mksquashfs.Options{FragmentQueueSizeMB: 16}, "input-dir")
	c.Check(argIndex(cmd, "-frag-queue"), Not(Equals), -1)
	c.Check(argValue(cmd, "-frag-queue"), Equals, "16")

	cmd = mksquashfs.Command("output.img", &mksquashfs.Options{FragmentQueueSizeMB: 0}, "input-dir")
	c.Check(argIndex(cmd, "-frag-queue"), Equals, -1)
}

// systemSuite checks the interaction of squashfs-tools on the running system.
type systemSuite struct{}

var _ = Suite(&systemSuite{})

func (s *systemSuite) SetUpSuite(c *C) {
	if _, err := exec.LookPath("mksquashfs"); err != nil {
		c.Skip(err.Error())
	}
}

func (s *systemSuite) TestReallyMksquashfs(c *C) {
	d := c.MkDir()

	inDir := filepath.Join(d, "in-dir")
	err := os.Mkdir(inDir, 0o700)
	c.Assert(err, IsNil)

	err = ioutil.WriteFile(filepath.Join(inDir, "canary.txt"), []byte("Hello"), 0o600)
	c.Assert(err, IsNil)

	imgFile := filepath.Join(d, "image.img")
	opts := mksquashfs.Options{
		Xattrs:        true,
		NfsExportable: true,
		Sparse:        true,

		CompressInodes:    true,
		CompressData:      true,
		CompressXattrs:    true,
		CompressFragments: true,

		Pad:       true,
		Fragments: true,
	}

	for _, config := range compressConfigs {
		comment := Commentf("compression format %s", config.CompressionName())
		opts.Compression = config
		err = mksquashfs.Run(imgFile, &opts, inDir)
		c.Assert(err, IsNil, comment)

		fi, err := os.Stat(imgFile)
		c.Assert(err, IsNil, comment)

		c.Check(fi.Size(), Not(Equals), 0)
	}
}
