// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package squashfstools_test

import (
	"errors"
	"fmt"
	"io/ioutil"
	"os"
	"os/exec"
	"path/filepath"
	"testing"

	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/squashfstools"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/squashfstools/mksquashfs"
	"git.ostc-eu.org/OSTC/OHOS/components/sysota/squashfstools/unsquashfs"
)

func Test(t *testing.T) { TestingT(t) }

// systemSuite checks the interaction of squashfs-tools on the running system.
type systemSuite struct{}

var _ = Suite(&systemSuite{})

func (s *systemSuite) SetUpSuite(c *C) {
	if _, err := exec.LookPath("mksquashfs"); err != nil {
		c.Skip(err.Error())
	}

	if _, err := exec.LookPath("unsquashfs"); err != nil {
		c.Skip(err.Error())
	}
}

func (s *systemSuite) TestSmoke(c *C) {
	d := c.MkDir()

	inDir := filepath.Join(d, "in-dir")
	err := os.Mkdir(inDir, 0o700)
	c.Assert(err, IsNil)

	err = ioutil.WriteFile(filepath.Join(inDir, "canary.txt"), []byte("Hello"), 0o600)
	c.Assert(err, IsNil)

	imgFile := filepath.Join(d, "image.img")
	err = mksquashfs.Run(imgFile, nil, inDir)

	if err != nil {
		fmt.Print(err)
	}

	c.Assert(err, IsNil)

	_, err = os.Stat(imgFile)
	c.Assert(err, IsNil)

	outDir := filepath.Join(d, "out-dir")
	err = unsquashfs.Run(imgFile, outDir, nil, "canary.txt")
	c.Assert(err, IsNil)

	data, err := ioutil.ReadFile(filepath.Join(outDir, "canary.txt"))
	c.Assert(err, IsNil)
	c.Check(data, DeepEquals, []byte("Hello"))
}

type cmdErrSuite struct{}

var _ = Suite(&cmdErrSuite{})

func (s *cmdErrSuite) TestSmoke(c *C) {
	cmd := exec.Command("false")
	_, err := cmd.Output()
	e := squashfstools.NewCommandError(cmd, err.(*exec.ExitError))

	c.Assert(e, ErrorMatches, `command error: \[false\] exited with code 1`+"\n")
	c.Assert(errors.Unwrap(e), Equals, err)

	cmd = exec.Command("sh", "-c", "echo boom >&2 && exit 2")
	_, err = cmd.Output()
	e = squashfstools.NewCommandError(cmd, err.(*exec.ExitError))

	c.Assert(e, ErrorMatches, `command error: \[sh -c echo boom >&2 && exit 2\] exited with code 2`+"\n"+`boom`+"\n")
	c.Assert(errors.Unwrap(e), Equals, err)
}
