// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: Huawei Inc.

package unsquashfs_test

import (
	"fmt"
	"os/exec"
	"testing"

	. "gopkg.in/check.v1"

	"git.ostc-eu.org/OSTC/OHOS/components/sysota/squashfstools/unsquashfs"
)

func Test(t *testing.T) { TestingT(t) }

type unsquashfsSuite struct{}

var _ = Suite(&unsquashfsSuite{})

// TODO(zyga): add checkers like HasArgument and ArgumentValue

func argIndex(cmd *exec.Cmd, wantedArg string) int {
	for idx, arg := range cmd.Args {
		if arg == wantedArg {
			return idx
		}
	}

	return -1
}

func argValue(cmd *exec.Cmd, wantedArg string) string {
	idx := argIndex(cmd, wantedArg)
	if idx == -1 {
		panic(fmt.Sprintf("cannot find argument %s in %v\n", wantedArg, cmd.Args))
	}

	if idx+1 < len(cmd.Args) {
		return cmd.Args[idx+1]
	}

	panic(fmt.Sprintf("argument %s has no value: %v", wantedArg, cmd.Args))
}

func (s *unsquashfsSuite) TestExtractWithoutOptions(c *C) {
	cmd := unsquashfs.Command("input.img", "output-dir", nil, "a-file", "a-dir")
	c.Check(cmd.Args, DeepEquals, []string{
		"unsquashfs", "-no-progress", "-no-xattrs", "-dest", "output-dir", "input.img", "a-file", "a-dir",
	})
}

func (s *unsquashfsSuite) TestExtractWithXattrs(c *C) {
	// There's an affirmative option for xattrs, unlike most of the other options.
	cmd := unsquashfs.Command("input.img", "output-dir", &unsquashfs.Options{Xattrs: true}, "a-file")
	c.Check(argIndex(cmd, "-xattrs"), Not(Equals), -1)
	c.Check(argIndex(cmd, "-no-xattrs"), Equals, -1)

	cmd = unsquashfs.Command("input.img", "output-dir", &unsquashfs.Options{Xattrs: false}, "a-file")
	c.Check(argIndex(cmd, "-xattrs"), Equals, -1)
	c.Check(argIndex(cmd, "-no-xattrs"), Not(Equals), -1)
}

func (s *unsquashfsSuite) TestExtractWithNumProcessors(c *C) {
	cmd := unsquashfs.Command("input.img", "output-dir", &unsquashfs.Options{NumProcessors: 2}, "a-file")
	c.Check(argIndex(cmd, "-processors"), Not(Equals), -1)
	c.Check(argValue(cmd, "-processors"), Equals, "2")

	cmd = unsquashfs.Command("input.img", "dir", &unsquashfs.Options{NumProcessors: 0}, "output-file", "output-dir")
	c.Check(argIndex(cmd, "-processors"), Equals, -1)
}

func (s *unsquashfsSuite) TestExtractWithDataQueueSize(c *C) {
	cmd := unsquashfs.Command("input.img", "output-dir", &unsquashfs.Options{DataQueueSizeMB: 16}, "a-file")
	c.Check(argIndex(cmd, "-data-queue"), Not(Equals), -1)
	c.Check(argValue(cmd, "-data-queue"), Equals, "16")

	cmd = unsquashfs.Command("input.img", "output-dir", &unsquashfs.Options{DataQueueSizeMB: 0}, "a-file")
	c.Check(argIndex(cmd, "-data-queue"), Equals, -1)
}

func (s *unsquashfsSuite) TestExtractWithFragmentQueueSize(c *C) {
	cmd := unsquashfs.Command("input.img", "output-dir", &unsquashfs.Options{FragmentQueueSizeMB: 16}, "a-file")
	c.Check(argIndex(cmd, "-frag-queue"), Not(Equals), -1)
	c.Check(argValue(cmd, "-frag-queue"), Equals, "16")

	cmd = unsquashfs.Command("input.img", "output-dir", &unsquashfs.Options{FragmentQueueSizeMB: 0}, "a-file")
	c.Check(argIndex(cmd, "-frag-queue"), Equals, -1)
}

func (s *unsquashfsSuite) TestExtractWithForce(c *C) {
	cmd := unsquashfs.Command("input.img", "output-dir", &unsquashfs.Options{Force: true}, "a-file")
	c.Check(argIndex(cmd, "-force"), Not(Equals), -1)

	cmd = unsquashfs.Command("input.img", "output-dir", &unsquashfs.Options{Force: false}, "a-file")
	c.Check(argIndex(cmd, "-force"), Equals, -1)
}
